#include "formula.h"
#include "model.h"
#include "../utils/utils.h"
#include <sstream>

Formula::~Formula() {}

// Constructors:
Formula::Formula() {
	m_operation = CONSTANT;
	initCommon();
}
Formula::Formula(const Formula& f) {
	m_owner = f.m_owner;
	m_operation = f.m_operation;
	m_children = f.m_children;
	m_parents = f.m_parents;

	m_value = f.m_value;
	//m_valueComputed = f.m_valueComputed;
	m_valueComputedSyncValue = f.m_valueComputedSyncValue;
	m_gradient = f.m_gradient;
	//m_gradientComputed = f.m_gradientComputed;
	m_gradientComputedSyncValue = f.m_gradientComputedSyncValue;
	m_index = f.m_index;
}
Formula::Formula(Op o) {
	m_operation = o;
	initCommon();
}
Formula::Formula(double v) {
	m_operation = CONSTANT;
	initCommon();
	//m_valueComputed = true;
	m_valueComputedSyncValue = 1;
	m_value = v;
}
Formula::Formula(PARAM_ID index) {
	initCommon();
	m_operation = PARAMETER;
	m_gradient = 0.0;
	m_index = index;
}
void Formula::initCommon() {
	//m_valueComputed = false;
	m_valueComputedSyncValue = 0;
	m_value = 0.0;
	//m_gradientComputed = false;
	m_gradientComputedSyncValue = 0;
	if (m_operation == LOGPLUS || m_operation == LOGTIMES || m_operation == LOGDIVIDE || m_operation == LOGPOWER) {
		m_gradient = neg_inf;
	} else {
		m_gradient = 0.0;
	}
	m_index = -1;
}

Formula& Formula::operator=(const Formula& f) {
	m_owner = f.m_owner;
	m_operation = f.m_operation;
	m_children = f.m_children;
	m_parents = f.m_parents;
	m_value = f.m_value;
	//m_valueComputed = f.m_valueComputed;
	m_valueComputedSyncValue = f.m_valueComputedSyncValue;
	m_gradient = f.m_gradient;
	//m_gradientComputed = f.m_gradientComputed;
	m_gradientComputedSyncValue = f.m_gradientComputedSyncValue;
	m_index = f.m_index;
	return *this;
}

// Reset functions:
void Formula::reset(Op o, Model* model, int numChildren) {
	m_operation = o;
	//m_valueComputed = false;
	m_valueComputedSyncValue = 0;
	m_value = 0.0;
	m_index = -1;
	resetCommon(model, numChildren);
}
void Formula::reset(double v, Model* model) {
	m_operation = CONSTANT;
	//m_valueComputed = true;
	m_valueComputedSyncValue = 1;
	m_value = v;	// reset to given value
	m_index = -1;
	resetCommon(model);
}
void Formula::reset(PARAM_ID index, Model* model) {
	m_operation = PARAMETER;
	//m_valueComputed = false;
	m_valueComputedSyncValue = 0;
	m_value = 0.0;
	m_index = index;
	resetCommon(model);
}
void Formula::resetCommon(Model* model, int numChildren) {
	//m_gradientComputed = false;
	m_gradientComputedSyncValue = 0;
	if (m_operation == LOGPLUS || m_operation == LOGTIMES || m_operation == LOGDIVIDE || m_operation == LOGPOWER) {
		m_gradient = neg_inf;
	} else {
		m_gradient = 0.0;
	}
	m_children.clear();
	m_parents.clear();
	if (m_operation == EXP || m_operation == LOG || m_operation == NEG || m_operation == SQRT || m_operation == TANH
			|| m_operation == RELU || m_operation == TRUNC_RELU || m_operation == LEAKY_RELU || m_operation == LOGISTIC
			|| m_operation == GD || m_operation == ERF) {
		m_children.reserve(1);
	} else if (m_operation == POWER || m_operation == LOGPOWER || m_operation == MINUS || m_operation == DIVIDE || m_operation == LOGDIVIDE) {
		m_children.reserve(2);
	} else if (numChildren > 0) {
		m_children.reserve(numChildren);
	}
	m_owner = model;
}

void Formula::setOwner(Model* m) { m_owner = m; }

int Formula::getIndex() const { return m_index; }
double Formula::getGradient() const { return m_gradient; }
int Formula::getNumParents() const { return m_parents.size(); }
int Formula::getNumChildren() const { return m_children.size(); }

void Formula::addChild(Formula* f) {
	m_children.push_back(f);
	f->m_parents.push_back(make_pair(this, m_children.size() - 1));
}

// Calls reserve(numChildren) on the children vector of this Formula node.
void Formula::reserveChildren(int numChildren) {
	m_children.reserve(numChildren);
}


void Formula::changedParamValues(bool recurOnChildren) {
	// Reset ``computed'' flags.
	//m_valueComputed = false;
	//m_gradientComputed = false;
	m_valueComputedSyncValue = 0;
	m_gradientComputedSyncValue = 0;
	if (recurOnChildren) {
		// Recur on children.
		for (unsigned int i = 0; i < m_children.size(); ++i) {
			m_children[i]->changedParamValues(recurOnChildren);
		}
	}
}

double Formula::evaluate(int syncValue) {
	//if (m_valueComputed) return m_value;
	if (m_valueComputedSyncValue == syncValue) return m_value;
	switch (m_operation) {
	case POWER:
		m_value = pow(m_children[0]->evaluate(syncValue), m_children[1]->evaluate(syncValue));
		break;
	case TIMES:
	{
		// Initialize m_value to the multiplicative identity element.
		m_value = 1.0;
		int n = m_children.size();
		// Evaluate each child and multiply onto m_value, placing the result into m_value.
		for (int i = 0; i < n; ++i) m_value *= m_children[i]->evaluate(syncValue);
		break;
	}
	case LOGPLUS:
	{
		m_value = neg_inf;	// Defined in utils.cpp.
		int n = m_children.size();
		// Evaluate each child and log-add to m_value.
		for (int i = 0; i < n; ++i) m_value = logAdd(m_value, m_children[i]->evaluate(syncValue));
		break;
	}
	case PLUS:
	case LOGTIMES:
	{
		m_value = 0.0;
		int n = m_children.size();
		// Evaluate each child and add to m_value
		for (int i = 0; i < n; ++i) m_value += m_children[i]->evaluate(syncValue);
		break;
	}
	case MINUS:
	case LOGDIVIDE:
		// subtract the second operand from the first, placing result into m_value
		// if m_args.size() > 2, ignores remaining operands
		m_value = m_children[0]->evaluate(syncValue) - m_children[1]->evaluate(syncValue);
		break;
	case DIVIDE:
		// divide the first operand by the second, placing result into m_value
		// if m_args.size() > 2, ignores remaining operands
		m_value = m_children[0]->evaluate(syncValue) / m_children[1]->evaluate(syncValue);
		break;
	case EXP:
		// exponentiate the argument (assumed to be only one), placing result into m_value
		m_value = exp(m_children[0]->evaluate(syncValue));
		break;
	case LOG:
	{
		// take the log of the argument (assumed to be only one, but not checked)
		// if the argument is negative, prints a warning and changes it to positive.. does this make sense?
		double childValue = m_children[0]->evaluate(syncValue);
		if (childValue < 0.0) cerr << "Error: Tried to take the log of a negative value: " << childValue << endl;
		m_value = log(childValue);
		break;
	}
	case NEG:
		// negate the single child
		m_value = -1.0 * m_children[0]->evaluate(syncValue);
		break;
	case LOGPOWER:
	{
		int n = m_children.size();
		if (n > 2) {
			cerr << "Warning: more than two children of a LOGPOWER node" << endl;
		}
		if (n == 2) {
			// the first child is the base
			double base = neg_inf, exponent = neg_inf;
			//if (m_args[0]->m_operation == LOGLOOKUP || m_args[0]->m_operation == CONSTANT) {
				base = m_children[0]->evaluate(syncValue);
			//} else {
			//	cerr << "Warning: base in a LOGPOWER node must be a LOGLOOKUP node or a CONSTANT node" << endl;
			//}
			// the second child is the exponent
			if (m_children[1]->m_operation == CONSTANT) {
				exponent = m_children[1]->evaluate(syncValue);
			} else {
				cerr << "Warning: exponent in a LOGPOWER node must be a CONSTANT node" << endl;
			}
			m_value = exp(exponent) * base;	// in the log domain, we multiply the base by exp(exponent)
		} else {
			cerr << "Warning: unexpected number of children of a LOGPOWER node: " << n << endl;
		}
		break;
	}
	case SQRT:
		m_value = sqrt(m_children[0]->evaluate(syncValue));
		break;
	case TANH:
		// Compute the hyperbolic tangent of the single child.
		m_value = tanh(m_children[0]->evaluate(syncValue));
		break;
	case RELU:
		// Compute the rectified linear unit response of the single child.
		m_value = m_children[0]->evaluate(syncValue);
		if (m_value <= 0) m_value = 0.0;
		break;
	case TRUNC_RELU:
		// Compute the truncated rectified linear unit response of the single child.
		m_value = m_children[0]->evaluate(syncValue);
		if (m_value <= 0.0) m_value = 0.0;
		if (m_value > 1.0) m_value = 1.0;
		break;
	case LEAKY_RELU:
		// Compute the leaky rectified linear unit response of the single child.
		m_value = m_children[0]->evaluate(syncValue);
		if (m_value <= 0) m_value = 0.01 * m_value;
		break;
	case LOGISTIC:
		// Compute the logistic function (1/(1+exp(-x))) of the single child x.
		m_value = 1.0 / (1.0 + exp(-1.0 * m_children[0]->evaluate(syncValue)));
		break;
	case GD:
		// Compute the Gudermannian function (arcsin(tanh(x))) of the single child x.
		m_value = asin(tanh(m_children[0]->evaluate(syncValue)));
		break;
	case ERF:
		// Compute the error function of the single child x.
		m_value = erf(m_children[0]->evaluate(syncValue));
		break;
	case MAX:
	{
		// Compute the max over all child nodes.
		m_index = -1;
		m_value = neg_inf;
		int n = m_children.size();
		for (int i = 0; i < n; ++i) {
			double currValue = m_children[i]->evaluate(syncValue);
			if (currValue > m_value) {
				m_value = currValue;
				m_index = i;
			}
		}
		if (m_index == -1) {
			cerr << "Error: in evaluate() of a MAX node, no max found!" << endl;
			m_index = 0;
			if (n >= 1) m_value = m_children[0]->evaluate(syncValue);
		}
		break;
	}
	case SAMPLE_MULTINOMIAL:
	{
		m_value = neg_inf;
		// Go through all events in the multinomial and add their scores.
		int n = m_children.size();
		double sumScores = 0.0;
		for (int i = 0; i < n; ++i) {
			sumScores += m_children[i]->evaluate(syncValue);
		}

		if (sumScores <= 0.0) {
			//m_maxChildIndex = 0;
			m_index = 0;
			if (n >= 1) {
				m_value = m_children[0]->evaluate(syncValue);
			} else {
				m_value = 0.0;
			}
		} else {
			// Go back through and divide by the sum to get the probability for each event.
			// At the same time, fill in the cumulative probability and sample.
			//m_maxChildIndex = -1;
			m_index = -1;
			double randDouble = getRandomDouble();
			double cumProb = 0.0;
			int ctr = 0;
			for (int i = 0; i < n; ++i) {
				// Divide by the sum to get the probability of this event.
				double currProb = m_children[i]->evaluate(syncValue);
				currProb /= sumScores;
				// Update the cumulative probability for this multinomial by adding the probability.
				cumProb += currProb;
				// Check if the random draw chose this event.
				if (randDouble <= cumProb) {
					//	cerr << "setting sampledEvent to ctr = " << ctr << endl;
					//m_maxChildIndex = ctr;
					m_index = ctr;
					m_value = m_children[i]->evaluate(syncValue);
					break;
				}
				++ctr;
			}
			if (m_index == -1) {
				cerr << "Warning: in evaluate() of a SAMPLE_MULT node: sampling unsuccessful, no valid events were found. random # = " << randDouble
					 << ", total log-probability mass = " << cumProb << ". Using event 0." << endl;
				m_index = 0;
				if (n >= 1) {
					m_value = m_children[0]->evaluate(syncValue);
				} else {
					cerr << "Error: 0 children of SAMPLE_MULT node." << endl;
				}
			}
		}
		break;
	}
	case PROBABILISTIC_AVERAGING:
	{
		// The probabilistic averaging node type is used at test time for stochastic pooling nodes.
		// It sums the children, first scaling them by their probabilities.

		// Go through all events in the multinomial and add their scores.
		int n = m_children.size();
		double sumScores = 0.0;
		for (int i = 0; i < n; ++i) {
			sumScores += m_children[i]->evaluate(syncValue);
		}

		if (sumScores <= 0.0) {
			// TODO What should we do here?
			if (n >= 1) {
				m_value = m_children[0]->evaluate(syncValue);
			} else {
				m_value = 0.0;
			}
		} else {
			// Go back through and divide by the sum to get the probability for each event,
			// then use it to weight the current value and add the product to the running sum.
			m_value = 0.0;
			for (int i = 0; i < n; ++i) {
				// Divide by the sum to get the probability of this event.
				double currScore = m_children[i]->evaluate(syncValue);
				double currProb = currScore / sumScores;
				// Add the product of the current score times the current probability.
				m_value += (currProb * currScore);
			}
		}
		break;
	}
	case PARAMETER:
		// Set m_value to the parameter value.
		m_value = m_owner->getValue(m_index);
		break;
	case CONSTANT:
		// Do nothing; i.e., simply return m_value below.
		break;
	}
	//m_valueComputed = true;
	m_valueComputedSyncValue = syncValue;
	return m_value;
}

double Formula::computeDerivative(int syncValue) {
	if (m_operation != PARAMETER) {
		cerr << "Warning: Called computeDerivative() on a non-parameter node." << endl;
		return -1.0;
	} else {
		//if (m_gradientComputed) return m_gradient;
		if (m_gradientComputedSyncValue == syncValue) return m_gradient;

		m_gradient = 0.0;
		//cerr << "Initial gradient for param " << m_name << " (" << m_index << ") = " << currentGradient << endl;
		// Pull down the derivative from each parent.
		if (m_parents.size() > 0) {
			for (vector<pair<Formula*, int> >::iterator parentIterator = m_parents.begin(); parentIterator != m_parents.end(); ++parentIterator) {
				m_gradient += (parentIterator->first)->pull(parentIterator->second, syncValue);
			}
		}
		// Add the current Formula's gradient to the parameter's gradient component in the Model.
		m_owner->addGradient(m_index, m_gradient);
		//cerr << "Final gradient for param " << m_name << " (" << m_index << ") = " << currentGradient << endl;
		//m_gradientComputed = true;
		m_gradientComputedSyncValue = syncValue;
		return m_gradient;
	}
}


/**
 * Pulls gradient from parent nodes, then computes gradient at current node,
 * and returns gradient update for given argument (child node).
 * @param m
 * @param childIndex
 * @return
 */
double Formula::pull(int childIndex, int syncValue) {
	double valueToBackprop;

	// If we haven't yet computed the gradient at the current node, pull from each parent.
	//if (!m_gradientComputed) {
	if (m_gradientComputedSyncValue != syncValue) {
		m_gradient = 0.0;
		if (m_parents.size() == 0) {
			m_gradient = 1.0;
		//} else {
		//	pair<Formula*, int>* enditer = &(m_parents[m_parents.size() - 1]);
		//	for (pair<Formula*, int>* myiter = &(m_parents[0]); myiter <= enditer; ++myiter) {
		//		m_gradient += (myiter->first)->pull(myiter->second, syncValue);
		//	}
		}

		// Go through all parents of the current node and pull gradient to the current node.
		if (m_parents.size() > 0) {
			for (vector<pair<Formula*, int> >::iterator parentIterator = m_parents.begin(); parentIterator != m_parents.end(); ++parentIterator) {
			//for (unsigned int i = 0; i < m_parents.size(); ++i) {
				// Pull from the current parent, passing the index of the current node
				// in the current parent's list of arguments
				//m_gradient += m_parents[i]->pull(m_indicesInParentLists[i]);
				//pair<Formula*, int>& currParent = m_parents[i];
				//pair<Formula*, int>& currParent = *parentIterator;
				//m_gradient += currParent.first->pull(currParent.second, syncValue);

				m_gradient += (parentIterator->first)->pull(parentIterator->second, syncValue);
			}
		}
		// Set the flag to true so that next time we pull from this node we don't
		// have to redo this computation.
		//m_gradientComputed = true;
		m_gradientComputedSyncValue = syncValue;
	}
	// Get the value to backpropagate to the desired child.
	valueToBackprop = getValueToBackprop(childIndex, syncValue);
	return valueToBackprop * m_gradient;
}

double Formula::getValueToBackprop(int childIndex, int syncValue) {
	switch (m_operation) {
	case POWER:
	{
		// to do this operation, the sign of t1 MUST be positive
		double t1 = m_children[0]->evaluate(syncValue);
		double t2 = m_children[1]->evaluate(syncValue);
		if (t1 < 0.0) {
			cerr << "Error: Sign of base x in x^y is negative!" << endl;
		}
		if (childIndex == 0) {
			// the derivative of x^y wrt x is y * x^(y-1)
			return t2 * pow(t1, t2 - 1);// * inc_val;
		} else {
			// the derivative of x^y wrt y is log(x) * x^y
			//if (m_valueComputed) {
			if (m_valueComputedSyncValue == syncValue) {
				return log(t1) * m_value; //* inc_val;
			} else {
				return log(t1) * pow(t1, t2); //* inc_val;
			}
		}
	}
	case LOGTIMES:
		// Compute/get the product of all children.
		evaluate(syncValue); // The product will now be in m_value.
		// If the result is nonzero, we can just log-divide out the given child's value.
		if (m_value != neg_inf) {
			// Log-divide total log-product by the given child's value.
			return (m_value - m_children[childIndex]->evaluate(syncValue));
		} else {
			// If the result is zero, we need to compute the log-product of the other children.
			// Initialize m_value to the multiplicative identity element in the log semiring (0).
			double valueToBackprop = 0.0;
			int n = m_children.size();
			// Log-multiply each child onto valueToBackprop except for childIndex.
			for (int i = 0; i < n; ++i) {
				if (i != childIndex) valueToBackprop += m_children[i]->evaluate(syncValue);
			}
			return valueToBackprop;
		}
	case TIMES:
		// Compute/get the product of all children.
		evaluate(syncValue); // The product will now be in m_value.
		// If the result is nonzero, we can just divide out the given child's value.
		if (m_value != 0.0) {
			// Divide the total product by the given child's value.
			return m_value / m_children[childIndex]->evaluate(syncValue);
		} else {
			// If the result is zero, we need to compute the product of the other children.
			// Initialize m_value to the multiplicative identity element (1).
			double valueToBackprop = 1.0;
			int n = m_children.size();
			// Multiply each child onto valueToBackprop except for childIndex.
			for (int i = 0; i < n; ++i) {
				if (i != childIndex) valueToBackprop *= m_children[i]->evaluate(syncValue);
			}
			return valueToBackprop;
		}
	case LOGPLUS:
		// For log-plus nodes, simply return 0.0.
		return 0.0;
	case PLUS:
		// For plus nodes, simply return 1.0.
		return 1.0;
	case MINUS:
		if (childIndex == 0) {
			// The derivative of the parent with respect to the left child is 1.
			return 1.0;
		} else if (childIndex == 1) {
			// The derivative of the parent with respect to the right child is -1.
			return -1.0;
		}
		// If the node has more than 2 children, return 0 for them.
		return 0.0;
	case LOGDIVIDE:
		if (childIndex == 0) {
			// The derivative of the parent with respect to the left child is 0.
			return 0.0;
		} else if (childIndex == 1) {
			// The derivative of the parent with respect to the right child is -1.
			return -1.0;
		}
		// If the node has more than 2 children, return 0 for them.
		return 0.0;
	case DIVIDE:
	{
		double t1 = m_children[0]->evaluate(syncValue);
		double t2 = m_children[1]->evaluate(syncValue);
		if (childIndex == 0) {
			// The derivative of the parent with respect to the left child is (1/t2).
			return (1.0 / t2);
		} else if (childIndex == 1) {
			// The derivative of the parent with respect to the right child is -t1/(t2^2).
			return (-t1) / (t2 * t2);
		}
		// If the node has more than 2 children, return 0 for them.
		return 0.0;
	}
	case EXP:
		// The derivative of e^x is e^x.
		return m_value;
	case LOG:
	{
		// The derivative of log(x) is 1/x.
		// Note: we do not do any zero checking
		double child0Value = m_children.at(0)->evaluate(syncValue);
		if (child0Value == 0.0) {
			cerr << "Warning: attempted to divide by zero when computing derivative of LOG node. Using infinity instead." << endl;
			return pos_inf;
		} else {
			return (1.0 / child0Value);
		}
	}
	case LOGPOWER:
		if (childIndex == 0) {
			// Get the value of the first child (the base).
			double baseValue = m_children.at(0)->evaluate(syncValue);
			if (m_children.size() == 2) {
				// Get the value of the other child (the exponent).
				double exponentValue = m_children[1]->evaluate(syncValue);
				if (baseValue == neg_inf) cerr << "Warning: baseValue == neg_inf!  Dividing by zero!" << endl;
				//if (exponentValue == neg_inf) cerr << "Warning: exponentValue == neg_inf!  Dividing by zero!" << endl;
				// The derivative of a^b wrt a is b*a^(b-1).
				// So, in the log domain, we want to return log(b*a^(b-1)) = log(b) + log(a^(b-1)) = log(b) + (b-1)*log(a).
				// We have log(a) and log(b), in baseValue and exponentValue, respectively.
				// So, the solution is: exponentValue + (exp(exponentValue) - 1)*baseValue.
				double ret;
				if (exponentValue == neg_inf) {
					ret = neg_inf;
				} else {
					ret = exponentValue + ((exp(exponentValue) - 1) * baseValue);
				}
				return ret;
			} else {
				cerr << "Warning: unexpected number of arguments of LOGPOWER node: " << m_children.size() << endl;
				return neg_inf;
			}
		} else {
			// do nothing
			return neg_inf;
		}
		break;
	case SQRT:
		// The derivative of sqrt(x) is 1/(2*sqrt(x)).
		evaluate(syncValue);	// sqrt(x) will now be in m_value.
		return (1.0 / (2.0 * m_value));
	case NEG:
		// The derivative of -x is -1.
		return -1.0;
	case TANH:
		// The derivative of tanh(x) = 1 - tanh^2 (x).
		evaluate(syncValue); // tanh(x) will now be in m_value.
		return (1.0 - (m_value * m_value));
	case RELU:
		// The derivative of relu(x) is 0 if x <=0 and 1 otherwise.
		evaluate(syncValue);	// relu(x) will now be in m_value.
		if (m_value <= 0.0) return 0.0;
		return 1.0;
	case TRUNC_RELU:
		// The derivative of the truncated relu(x) is 0 if x <=0 or x>1, and 1 otherwise.
		evaluate(syncValue);	// truncated relu(x) will now be in m_value.
		if (m_value <= 0.0) return 0.0;
		if (m_value > 1.0) return 0.0;
		return 1.0;
	case LEAKY_RELU:
		// The derivative of leaky relu(x) is 0.01 if x <=0 and 1 otherwise.
		evaluate(syncValue);	// relu(x) will now be in m_value.
		if (m_value <= 0.0) return 0.01;
		return 1.0;
	case LOGISTIC:
		// The derivative of logistic(x) = logistic(x) * (1 - logistic(x)).
		evaluate(syncValue);
		return (m_value * (1.0 - m_value));
		break;
	case GD:
		// The derivative of the Gudermannian function of x is 1/(cosh(x)).
		evaluate(syncValue);
		return (1.0 / (cosh(m_value)));
	case ERF:
		// The derivative of the error function of x is (2/sqrt(pi))exp{-x^2}.
		evaluate(syncValue);
		return (2.0 * exp(-1.0 * m_value * m_value)) / sqrt(3.14159265358979323846);
	case MAX:
	{
		evaluate(syncValue);
		// If we are backpropagating to the max-value child, return 1.
		if (childIndex == m_index) return 1.0;
		// Else return 0.
		return 0.0;
	}
	case SAMPLE_MULTINOMIAL:
		evaluate(syncValue);
		// If we are backpropagating to the sampled child, return 1.
		if (childIndex == m_index) return 1.0;
		// Else return 0.
		return 0.0;
	case PARAMETER:
		cerr << "Error: called getValueToBackprop() on a PARAMETER node. Should never occur." << endl;
		return -1.0;
	case CONSTANT:
		return -1.0;
	}
	return -1.0;
}

// Sets m_value to given newValue. Useful for reusing constant Formula nodes.
void Formula::setValue(double newValue) {
	m_value = newValue;
}

void Formula::printFormula(int numtabs) const {
	for (int i = 0; i < numtabs; i++) {
		cout << "\t";
	}
	stringstream s;
	s << getStringForOp(m_operation);
	if (m_operation == PARAMETER) s << "_" << m_index;
	s << "\t" << m_value << "\t" << m_gradient;
	cout << s.str() << endl;
	for (unsigned int i = 0; i < m_children.size(); ++i) {
		m_children[i]->printFormula(numtabs + 1);
	}
}

string Formula::toString() const {
	stringstream s;
	s << "\tval=" << m_value << "\t\tgrad" << m_gradient;
	return s.str();
}
