//
// Class to represent a model. See details below.
//
// author: Kevin Gimpel
// date: 2008-2015
//
#ifndef MODEL_H_
#define MODEL_H_

#include <iostream>
#include <iomanip>
#include <fstream>
#include <vector>
#include <cstdlib>
#include <map>
#include <math.h>
#include "../utils/utils.h"
#include "formula_pool.h"
#include "formula.h"

typedef unsigned int PARAM_ID;
enum Algorithm { SGD, SGD_MOMENTUM, ADAGRAD, ADADELTA };

/**
 * The Model class contains data structures and functions for training and decoding with
 * models.
 * The key components are:
 * (1) m_V, the parameter vector
 * (2) m_G, the gradient vector
 * (3) FormulaPools to track Formula node objects
 *
 * The model provides functions for accessing and changing the parameter and gradient vectors.
 *
 * For optimal performance and correctness, the user of this class should never have to create or destroy
 * Formula objects. The Model class uses FormulaPools which own Formula objects.
 * To get an operator Formula object with operator op, call getFormulaObject(op).
 * To get a constant Formula object with value v, call getConstantFormulaObject(v).
 * To get a parameter Formula object for parameter id, call getCachedParameterFormula(id).
 * This uses the parameter Formula node cache by default.
 * If you instead wish to bypass the cache (e.g., if you are using your own parameter Formula
 * cache or you want to keep separate parameter Formula nodes in your FDAGs each time a parameter
 * is used), use getParameterFormulaObject(id) instead of getCachedParameterFormula(id).
 *
 * When calling getFormulaObject for a PLUS or TIMES node, pass the number of expected children
 * as the second argument; this will improve both memory and runtime efficiency. If you end up using
 * a different number of children than that passed, it will not affect correctness, only time/memory usage.
 * You don't have to do this for operators that are necessarily unary or binary, just for those like
 * PLUS and TIMES that can have arbitrarily-many operands.
 *
 * After creating the FDAG for an example, the evalAndComputeDerivatives() function is provided for
 * evaluating the FDAG and computing derivatives of its root with respect to the parameters in
 * the parameter Formulas cache. For debugging, you can pass true as the final argument (doFiniteDifferenceChecking)
 * to do simple finite-difference checking of the computed derivatives. The computed derivatives are
 * stored in the gradient vector (m_G). By default, the derivatives are accumulated in m_G, i.e.,
 * the new derivative computed by evalAndComputeDerivatives() is added to the existing value in m_G.
 *
 * Functions are provided for parameter updating after the gradient vector is filled in:
 * doSGDUpdate(), doSGDMomentumUpdate(), doAdaGradUpdate(), and doAdaDeltaUpdate().
 * After calling them, call resetGradientVector() to reset the gradient vector to all zeroes.
 *
 * When you are finished processing an example and no longer need any of the Formula objects you
 * created, call resetFormulaObjects(). If you are using the parameter Formulas cache (m_parameterFormulas),
 * pass true as the argument.
 *
 * Mini-batch optimization is easily supported: when computing the gradient on an FDAG, the
 * new derivative is added to the old entry in the gradient vector. Call evalAndComputeDerivatives()
 * on a mini-batch of examples, one by one (interspersed with calls to resetFormulaObjects()),
 * and then call do*Update() and resetGradientVector() at the end of the mini-batch. See examples in
 * src/samples/.
 *
 * author: Kevin Gimpel
 * date: 2008-2015
 */
class Model {
	public:
		Model();
		Model(int formulaCapacity);
		// Initialize a model with the given number of Formula nodes in the FormulaPool.
		Model(int unaryFormulaCapacity, int binaryFormulaCapacity, int fertileFormulaCapacity, int barrenFormulaCapacity);
		virtual ~Model();

		// Creates a new parameter in the model, initializes its value to the given value, and returns its PARAM_ID.
		// useCache should be true if you are expecting to use the parameter Formula cache.
		PARAM_ID initializeNewParameter(double value, bool useCache = true);
		void initializeParameterMatrix(vector<vector<PARAM_ID> >& m, unsigned int numRows, unsigned int numCols, double scalingFactor);
		void initializeParameterMatrixGivenRowLengths(vector<vector<PARAM_ID> >& m, vector<int>& numRows, double scalingFactor);
		void initializeParameterMatrices(vector<vector<vector<PARAM_ID> > >& m, const vector<int>& numRows, const vector<int>& numCols, double scalingFactor);
		void initializeParameterVector(vector<PARAM_ID>& v, unsigned int numEntries, double scalingFactor);

		// Functions for getting and setting parameter and gradient values.
		double getValue(PARAM_ID index);
		void setValue(PARAM_ID index, double newValue);
		double getGradient(PARAM_ID index);
		void setGradient(PARAM_ID index, double newGradient);
		void addGradient(PARAM_ID index, double gradientIncrement);

		// Returns number of entries in parameter and gradient vectors.
		int getNumParams();

		// Functions for getting Formula objects for use when building FDAGs:
		// For operator Formula nodes:
		// (Optionally provide the number of children that will be added to this Formula and that number will be pre-allocated
		// and the appropriate FormulaPool will be used to optimize speed and memory usage.)
		Formula* getFormulaObject(Op o, int numChildren = 0);
		// For constant Formula nodes:
		Formula* getConstantFormulaObject(double v);
		// If you want to compute gradients, use the following function whenever you need a
		// parameter Formula node. It will either return the cached Formula stored in
		// m_parameterFormulas or create a new one, store its pointer in m_parameterFormulas,
		// and return the pointer.
		Formula* getCachedParameterFormula(PARAM_ID id);
		// Like above function, but uses given cache instead of Model-internal m_parameterFormulas cache.
		Formula* getCachedParameterFormula(PARAM_ID id, vector<Formula*>& parameterFormulas);
		// For parameter Formula nodes (bypasses cache of parameter Formula nodes):
		Formula* getParameterFormulaObject(PARAM_ID id);

		// Resets formula pool, allowing reuse of all previously-created Formula nodes.
		// Call this when you are starting the generation of the FDAG for a new example.
		void resetFormulaObjects(bool resetParameterFormulaCache);
		// Resets cached parameter formulas in m_parameterFormulas, setting them all to NULL.
		// This is called by resetFormulas if resetParameterFormulaCache = true.
		void resetCachedParameterFormulas(vector<Formula*>& parameterFormulas);

		void resetFormulaComputedFlags();

		//int getNumberOfFormulasUsedInPool() const;
		void getNumberOfFormulasUsedInPool(int& unary, int& binary, int& fertile, int& barren) const;

		// Resets gradient vector to all 0. Call after making a gradient update.
		void resetGradientVector();
		// Computes and return norm of parameter vector.
		double computeParameterVectorNorm();
		// Computes and return norm of gradient vector.
		double computeGradientVectorNorm();
		// Multiplies each entry in the gradient vector by the given constant c.
		void scaleGradientVectorByConstant(double c);

		// Evaluates f then computes derivatives for all non-NULL entries in m_parameterFormulas.
		double evalAndComputeDerivatives(Formula* f, int exampleNumber, bool doFiniteDifferenceChecking, bool alwaysPrint = false);
		double evalAndComputeDerivatives(Formula* f, vector<Formula*>& parameterFormulas, int syncValue, int exampleNumber, bool doFiniteDifferenceChecking, bool alwaysPrint = false);
		// Performs stochastic gradient descent update on parameters using current gradient vector and given stepsize.
		void doSGDUpdate(double stepsize);
		// Performs stochastic gradient descent w/ momentum update on parameters using given previous updates, current
		// gradient vector, and given stepsize and momentum coefficient.
		double doSGDMomentumUpdate(vector<double>& previousUpdates, double stepsize, double momentumCoefficient);
		// Performs (diagonal) AdaGrad update on parameters using current gradient vector given sum of squared
		// gradients and stepsize.
		double doAdaGradUpdate(vector<double>& sumOfSquaredGradients, double stepsize);
		// Performs AdaDelta update on parameters using current gradient vector given sum of squared
		// gradients, sum of squared updates, and the two hyperparameters epsilon and rho.
		double doAdaDeltaUpdate(vector<double>& sumOfSquaredGradients, vector<double>& sumOfSquaredUpdates, double epsilon, double rho);

		// Converts string into Algorithm enum.
		static Algorithm getAlgorithm(const string& algString) {
			Algorithm alg;
			string lowercasedAlgString = lowercase(algString);
			if (lowercasedAlgString == "sgd") {
				alg = SGD;
			} else if (lowercasedAlgString == "sgd_momentum") {
				alg = SGD_MOMENTUM;
			} else if (lowercasedAlgString == "adagrad") {
				alg = ADAGRAD;
			} else if (lowercasedAlgString == "adadelta") {
				alg = ADADELTA;
			} else {
				cerr << "Warning: Unexpected optimization algorithm: " << algString << ". Using AdaDelta." << endl;
				alg = ADADELTA;
			}
			return alg;
		}

	protected:
		// Initializes Model using initial Formula capacity.
		void initCommon(int unaryFormulaCapacity, int binaryFormulaCapacity, int fertileFormulaCapacity, int barrenFormulaCapacity);

		// Performs finite difference checking.
		void finiteDifferenceCheck(Formula* f, PARAM_ID paramInd, double result, double deriv, double delta, int exampleNumber, bool alwaysPrint = false);

		vector<double> m_V;  // Parameter vector.
		vector<double> m_G;  // Gradient vector.

		// When building the FDAG for a single example, the m_parameterFormulas vector
		// holds pointers to previously-created Formula nodes for parameter nodes, which are reused
		// whenever possible.
		// We use a vector for fast lookup by parameter index, but for large sparse models it would be much more
		// space-efficient to use a map or hash due to the sparsity of parameter use in a single example.
		// It would also become more time-efficient due to the need to reset these Formula*'s to NULL after
		// processing each example.
		vector< Formula* > m_parameterFormulas;
		//map< PARAM_ID, Formula* > m_parameterFormulas;

		// Number of parameters in the model.
		int m_numParams;

	private:
		// We store Formula objects here using FormulaPools. They should be called upon to retrieve Formula
		// objects whenever they are needed. Once ALL existing Formula's are no longer needed, they should all
		// be freed. FormulaPool does NOT support freeing a single Formula at a time. All must be
		// freed at the same time. This is done for efficiency and since it matches the expected
		// use case for Models, namely that Formula's are needed when building the FDAG for a single
		// example, but when moving on to the next example, all can be thrown out.

		// We use three different pools because they have different characteristics in the sizes of their
		// parent and child lists.
		FormulaPool m_fertileFormulaPool;
		FormulaPool m_unaryFormulaPool;
		FormulaPool m_binaryFormulaPool;
		FormulaPool m_barrenFormulaPool;
};

// Gets parameter value corresponding to given index.
inline double Model::getValue(PARAM_ID index) { return m_V.at(index); }


#endif /*MODEL_H_*/
