//#include <stdlib.h>
#include "../classifier_utils.h"
#include "cnn.h"

void ConvolutionalNeuralNetwork::initializeParameters(Model& model, int dataInputDimension, int numLabels, const string& filters,
		bool initializeSoftmaxParameters) {
	m_inputDimension = dataInputDimension;
	m_numLabels = numLabels;

	double scalingFactor = 0.2;

	if (filters != "" && filters != "0") {
		vector<string> filterStrings;
		tokenize(filters, ',', filterStrings);
		for (unsigned int i = 0; i < filterStrings.size(); ++i) {
			vector<string> filterTokens;
			tokenize(filterStrings[i], ':', filterTokens);
			if (filterTokens.size() == 2) {
				int filterWidth = atoi(filterTokens[0].c_str());
				int numberOfFilters = atoi(filterTokens[1].c_str());

				string hiddenVectorSizeString = "";
				int hiddenVectorSizeInt = -1;
				bool initSoftmaxFalse = false;

				for (unsigned int j = 0; j < numberOfFilters; ++j) {
					// Initialize a filter with the given width.
					Filter f;
					f.m_width = filterWidth;
					int numWeights = filterWidth * m_inputDimension;
					for (unsigned int k = 0; k < numWeights; ++k) {
						f.m_weights.push_back(model.initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
					}
					f.m_bias = model.initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5));

					m_filters.push_back(f);
				}
				cerr << "Created " << numberOfFilters << " filters of width " << filterWidth << "." << endl;
			} else {
				cerr << "Error: expected each (comma-delimited) filter to have the format \"width:number\". "
						"Given: " << filterStrings[i] << endl;
			}
		}
	}
	m_numFilters = m_filters.size();
	cerr << "Using " << m_numFilters << " convolutional filters." << endl;

	if (initializeSoftmaxParameters) {
		m_softmaxWeights.clear();
		m_softmaxBias.clear();

		int outputDimension = m_inputDimension;
		if (m_numFilters > 0) outputDimension = m_numFilters;

		for (unsigned int i = 0; i < m_numLabels; ++i) {
			vector<PARAM_ID> rowi;
			for (unsigned int j = 0; j < outputDimension; ++j) {
				rowi.push_back(model.initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
			}
			m_softmaxWeights.push_back(rowi);
			m_softmaxBias.push_back(model.initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
		}
		// code from lstm.cpp:
		// Parameters for softmax weight matrices. Outermost vector ranges over hidden layers.
		// Dimensions of matrix i: m_layerWidths[i] x m_numLabels.
		//setSoftmaxMatrixDimensions(numLabels, m_softmaxMatrixInputDimensions, m_softmaxMatrixOutputDimensions, m_layerWidths);
		//model.initializeParameterMatrices(m_softmaxWeights, m_softmaxMatrixInputDimensions, m_softmaxMatrixOutputDimensions, scalingFactor);

		// Parameters for softmax bias vectors. Outermost vector ranges over hidden layers.
		// Dimensions of vector i: m_numLabels.
		//model.initializeParameterMatrix(m_softmaxBias, m_numLayers, numLabels, scalingFactor);
	}
}

int ConvolutionalNeuralNetwork::getNumberOfFilters() const {
	return m_numFilters;
}

// Resets data structures for processing the next example (whether in training or decoding).
void ConvolutionalNeuralNetwork::resetForNewExample() {
	// With no convolutional layer, the number of ci nodes is the same as the input dimension.
	unsigned int numCiNodes = m_inputDimension;
	if (m_numFilters > 0) numCiNodes = m_numFilters;

	// Clear all data structures containing cached formulas from the previous datum, and set them up for the new datum.
	if (m_ciNodes.size() == numCiNodes) {
		for (unsigned int i = 0; i < m_ciNodes.size(); ++i) {
			m_ciNodes[i] = NULL;
		}
	} else {
		m_ciNodes.clear();
		m_ciNodes.reserve(numCiNodes);
		for (unsigned int i = 0; i < numCiNodes; ++i) {
			m_ciNodes.push_back(NULL);
		}
	}
}

Formula* ConvolutionalNeuralNetwork::createFilterPoolingOutputNode(Model& model, SequenceDatum& datum, unsigned int i) {
	if (m_ciNodes.at(i) != NULL) return m_ciNodes[i];
	// Node c_i is a max-over-time node for filter i.
	// Get a reference to filter i.
	Filter& filter = m_filters.at(i);
	int numChildren = datum.m_sequenceLength + 1 - filter.m_width;
	if (numChildren == 0) {
		Formula* ret = model.getConstantFormulaObject(0.0);
		m_ciNodes[i] = ret;
		return ret;

	} else {
		// Create a max-pooling node.
		Formula* poolingNode = model.getFormulaObject(MAX, numChildren);
		// Go through all positions in the sequence.
		for (int wordIndex = 0; wordIndex < datum.m_sequenceLength; ++wordIndex) {
			// Only apply the filter if its width fits in the remainder of the sentence.
			if (wordIndex + filter.m_width <= datum.m_sequenceLength) {
				poolingNode->addChild(createFilterOutputNodeAtPosition(model, filter, datum, wordIndex));
			}
		}
		m_ciNodes[i] = poolingNode;
		return poolingNode;
	}
}

Formula* ConvolutionalNeuralNetwork::createFilterOutputNodeAtPosition(Model& model, Filter& filter, const SequenceDatum& datum, int position) {
	Formula* filterSum = model.getFormulaObject(PLUS, (filter.m_width * m_inputDimension) + 1);
	// Go through the number of words in the filter.
	for (int i = 0; i < filter.m_width; ++i) {
		// If the current desired word position in the sentence is a valid word position,
		if (position + i < datum.m_sequenceLength) {
			// Apply the filter to the current item in the sequence.
			for (int j = 0; j < m_inputDimension; ++j) {
				// Multiply the filter weight by the parameter in the current item.
				Formula* prod = model.getFormulaObject(TIMES, 2);
				// TODO speed up by avoiding multiplication.
				int filterPosition = (i * m_inputDimension) + j;
				Formula* filterWeightFormula = model.getCachedParameterFormula(filter.m_weights.at(filterPosition));
				prod->addChild(filterWeightFormula);
				Formula* ngramFormula = datum.m_vectFormulas.at(position + i).at(j);
				prod->addChild(ngramFormula);

				filterSum->addChild(prod);
			}
		}
	}
	// filterSum now contains the dot product of the filter and the n-gram.
	// Add the bias.
	filterSum->addChild(model.getCachedParameterFormula(filter.m_bias));
	Formula* activation = model.getFormulaObject(LEAKY_RELU, 1);
	activation->addChild(filterSum);
	return activation;
}
