#include <iostream>
#include <stdlib.h>
#include <sys/time.h>
#include "../../utils/cmdparams.h"
//#include "../../fdag/model.h"
//#include "classifier_utils.h"
#include "paraphrase_model.h"
//#include "rnn/lstm.h"
//#include "rnn/gru.h"


int main(int argc, char** argv) {
    ParamsArray params(argc, argv);
    params.setDescription("Trains a paraphrase model using recurrent neural networks.");
    params.addConstraint("train_file",
    		"file containing training data, where each line contains the sentence followed by a tab then the "
    		"label (0 or 1)", false); // not optional
    params.addConstraint("dev_file",
    		"file containing dev data to compute accuracy for, same format as train_file", false); // not optional
    params.addConstraint("test_file",
    		"file containing test data to compute accuracy for, same format as train_file", false); // not optional
    params.addConstraint("phrase_file",
    		"file containing phrases for demo mode, one phrase per line; if none given (or \"NONE\" is given), will not run demo mode", true); // optional
    params.addConstraint("cv",
    		"if > 0, will do CV on train_file instead of testing on dev/test; this is the number of folds to use with CV (default: 0)", true); // optional

    params.addConstraint("word_vector_file",
    		"file containing word vectors, where each line contains the word followed by a tab then the "
    		"space-separated vector entries", false); // not optional
    params.addConstraint("update_word_vectors",
    		"if true, will learn update given word vectors during learning (default: true)", true); // optional
    params.addConstraint("init_word_vectors_randomly",
    		"if true, will initialize word vectors randomly before learning (default: false)", true); // optional

    //params.addConstraint("directions", "directions of RNNs to use; comma-delimited list of directions like \"forward\" and \"backward\" (default: forward)", true); // optional
    //params.addConstraint("layer_widths", "widths of hidden layers, comma-delimited. e.g., \"10,10,5\", "
    //		"or use \"0\" for no hidden layers (default: 10,5)", true); // optional
    params.addConstraint("forward_lstms", "configuration of forward LSTMs to train; in format \"number:layer_widths.number:layer_widths...\", where \"number\" "
    		"is the number of forward LSTMs to train that have the following sequence of hidden layer widths. E.g., to train one forward LSTM with "
    		"one hidden layer of width 10 and 2 forward LSTMs with one hidden layer of width 5 and a second hidden layer of width 7, pass \"1:10.2:5,7\". "
    		"Pass \"0\" if no forward LSTMs are desired. (default: 1:10)", true); // optional
    params.addConstraint("backward_lstms", "configuration of backward LSTMs to train, same format as forward_lstms (default: 0)", true); // optional
    params.addConstraint("forward_grus", "configuration of forward GRUs to train, same format as forward_lstms (default: 0)", true); // optional
    params.addConstraint("backward_grus", "configuration of backward GRUs to train, same format as forward_lstms (default: 0)", true); // optional
    params.addConstraint("algorithm", "optimization algorithm; can be \"sgd_momentum\", \"adagrad\", or \"adadelta\" (case insensitive) (default: adadelta)", true); // optional
    params.addConstraint("stepsize", "stepsize for AdaGrad (default: 0.1)", true); // optional
    params.addConstraint("minibatch_size", "mini-batch size for AdaGrad (default: 10)", true); // optional
    params.addConstraint("num_epochs", "number of epochs of learning (default: 3)", true); // optional
    params.addConstraint("decode_every_k", "number of examples to decode after processing, if train data is bigger than 11k examples (default: 3000)", true); // optional
    params.addConstraint("dropout_rate", "probability of setting an output unit to zero; use a value <= 0 for no dropout (default: 0.5)", true); // optional
    params.addConstraint("gradient_norm_threshold", "max gradient norm (\"threshold\" from Pascanu et al paper; use a value <= 0 for no gradient rescaling (default: 1.0)", true); // optional
    params.addConstraint("margin", "value of delta in loss function (default: 1.0)", true); // optional
    params.addConstraint("lambda_w", "value of lambda_w, the regularization strength for the model parameters (default: 0.1)", true); // optional

   	if (!params.runConstraints("run_paraphrase_model")) return 0;

	// Read in command-line parameters.
	string trainFilename = params.asString("train_file");
	string devFilename = params.asString("dev_file");
	string testFilename = params.asString("test_file");
	string phraseFilename = params.asString("phrase_file", "NONE");
	int cvNumFolds = atoi(params.asString("cv", "0").c_str());
	string wordVectorFilename = params.asString("word_vector_file");
	bool updateWordVectors = (params.asString("update_word_vectors", "true") == "true");
	bool initWordVectorsRandomly = (params.asString("init_word_vectors_randomly", "false") == "true");
	string forwardLSTMs = params.asString("forward_lstms", "1:10");
	string backwardLSTMs = params.asString("backward_lstms", "0");
	string forwardGRUs = params.asString("forward_grus", "0");
	string backwardGRUs = params.asString("backward_grus", "0");
	string algorithm = params.asString("algorithm", "adadelta");
	double stepsize = atof(params.asString("stepsize", "0.1").c_str());
	int minibatchSize = atoi(params.asString("minibatch_size", "10").c_str());
	int numEpochs = atoi(params.asString("num_epochs", "3").c_str());
	int decodeEveryK = atoi(params.asString("decode_every_k", "3000").c_str());
	double dropoutRate = atof(params.asString("dropout_rate", "0.5").c_str());
	double gradientNormThreshold = atof(params.asString("gradient_norm_threshold", "1.0").c_str());
	double margin = atof(params.asString("margin", "1.0").c_str());
	double lambdaW = atof(params.asString("lambda_w", "0.1").c_str());

	// Change following to true to do finite difference checking of gradients.
	bool doFiniteDifferenceChecking = false;
	srand(1);

	struct timeval initialTime, endTime;
	gettimeofday(&initialTime, 0);

	if (cvNumFolds <= 0) {
		vector<string> trainLines, devLines, testLines;
		readLinesFromFile(trainFilename, trainLines);
		readLinesFromFile(devFilename, devLines);
		readLinesFromFile(testFilename, testLines);

		ParaphraseModel model(10000, 2000000, 10000, 60000);
		double accuracy = model.trainAndTest(trainLines, devLines, testLines, wordVectorFilename, forwardLSTMs, backwardLSTMs,
			forwardGRUs, backwardGRUs, algorithm, stepsize, minibatchSize, numEpochs, decodeEveryK, dropoutRate,
			updateWordVectors, initWordVectorsRandomly, gradientNormThreshold, margin, lambdaW, doFiniteDifferenceChecking, phraseFilename);
		cerr << "Final accuracy: " << accuracy << endl;
	} else {
		// Load data.
		vector<string> data;
		readLinesFromFile(trainFilename, data);
		//int inputDimension = loadData(dataFile, data, &numLabels);
		// Do n-fold cross-validation.
		double accuracySum = 0.0;
		unsigned int dataSize = data.size();
		int foldSize = dataSize / cvNumFolds;
		vector<double> accuracies;
		for (int i = 0; i < cvNumFolds; ++i) {
			cerr << "For test fold " << i << ", creating test set with examples " << (i*foldSize) << " to " << (((i+1)*foldSize) - 1) << endl;
			// Generate train/test sets for the current fold (approximately based on discretized index cutoffs).
			vector<string> trainLines, devLines, testLines;
			for (unsigned int j = 0; j < dataSize; ++j) {
				if (((int)j) >= (i*foldSize) && ((int)j) < (i+1)*foldSize) {
					// If the current datum falls within the indices of the current test fold, add it to testData.
					testLines.push_back(data[j]);
					//cerr << "test index: " << j << endl;
				} else {
					// Otherwise, add 90% of it to the training set and 10% to the dev set
					if (j % 10 == 0) {
						devLines.push_back(data[j]);
						//cerr << "dev index: " << j << endl;
					} else {
						trainLines.push_back(data[j]);
						//cerr << "train index: " << j << endl;
					}
				}
			}
			cerr << "Size of training set: " << trainLines.size() << endl;
			cerr << "Size of dev set: " << devLines.size() << endl;
			cerr << "Size of test set: " << testLines.size() << endl;
			// Create a model for the current split.
			ParaphraseModel model(10000, 2000000, 10000, 60000);
			//Model model(10000, 2000000, 10000, 60000);
			// Train and test.
			double currAccuracy = model.trainAndTest(trainLines, devLines, testLines, wordVectorFilename,
				forwardLSTMs, backwardLSTMs, forwardGRUs, backwardGRUs, algorithm, stepsize, minibatchSize,
				numEpochs, decodeEveryK, dropoutRate, updateWordVectors, initWordVectorsRandomly, gradientNormThreshold,
				margin, lambdaW, doFiniteDifferenceChecking, phraseFilename);
			accuracies.push_back(currAccuracy);
			accuracySum += currAccuracy;
		}
		double accuracyMean = accuracySum / ((double) cvNumFolds);
		cerr << "CV mean accuracy: \t" << accuracyMean << endl;
		cerr << "CV accuracy stddev: \t" << standardDeviation(accuracies) << endl;
	}
	gettimeofday(&endTime, 0);
	double cumulativeTime = computeTimeDifference(initialTime, endTime);
	cerr << "Total training time:\t" << cumulativeTime << endl;
    return 0;
}
