#ifndef BIDIR_LSTM_H_
#define BIDIR_LSTM_H_

#include "../classifier_utils.h"
#include "../../fdag/formula.h"
#include "../../fdag/model.h"

// This implementation permits multiple sequences to be modeled with the same parameters in the same datum.
// There is a single set of parameter matrices for all sequences, but the struct keeps distinct matrices
// of cached nodes for the different sequences. The functions to build Formulas take the index of the sequence
// of interest so that they know which cache to use.
class DeepBidirectionalLSTM {
public:
	void initializeParameters(Model& model, int dataInputDimension, int numLabels, const string& layerWidths,
		unsigned int numSequencesPerDatum, bool initializeSoftmaxParameters, bool useLookahead);
	int getTotalHiddenVectorSize();
	Formula* createHiddenNodeAtTimeAtDepth(Model& model, SequenceDatum& datum, unsigned int sequenceIndex, unsigned int i,
		int t, unsigned int depth, double inputDropoutRate, double hiddenDropoutRate, int terminationTime = 0);
	void getGateReport(Model& model, SequenceDatum& datum, unsigned int sequenceIndex, int t, unsigned int depth,
		double& inputGateSum, double& forgetGateSum, double& outputGateSum, int terminationTime = 0);
	void resetForNewExample(unsigned int sequenceIndex, int exampleLength);

	// The number of hidden layers.
	int m_numLayers;
	// Vector of hidden layer widths.
	vector<int> m_layerWidths;
	int m_cachedLayerWidthSum;
	// Parameters for softmax weight matrices. Outermost vector ranges over hidden layers.
	// Dimensions of matrix i: m_layerWidths[i] x m_numLabels.
	vector<vector<vector<PARAM_ID> > > m_softmaxWeights;
	// Parameters for softmax bias vectors. Outermost vector ranges over hidden layers.
	// Dimensions of vector i: m_numLabels.
	vector<vector<PARAM_ID> > m_softmaxBias;

private:
	Formula* createGateAtTimeAtDepth(Model& model, SequenceDatum& datum, unsigned int sequenceIndex, unsigned int i, int t,
		unsigned int depth,
		const vector<vector<vector<PARAM_ID> > >& weightsObsGate,
		const vector<vector<vector<PARAM_ID> > >& weightsHiddenGate,
		const vector<vector<PARAM_ID> >& weightsCellGate, const vector<vector<PARAM_ID> >& gateBias,
		const vector<vector<vector<PARAM_ID> > >& weightsNextObsGate,
		bool outputGate, double inputDropoutRate, double hiddenDropoutRate, int terminationTime = 0);
	void setPrevNextTime(bool& usePrevNext, int& prevNextTime, int t, int terminationTime, int numWords) const;
	void setLookaheadPrevNextTime(bool& useLookaheadPrevNext, int& lookaheadPrevNextTime, int t, int terminationTime, int numWords) const;

	Formula* createCellNodeAtTimeAtDepth(Model& model, SequenceDatum& datum, unsigned int sequenceIndex, unsigned int i, int t,
		unsigned int depth, double inputDropoutRate, double hiddenDropoutRate, int terminationTime = 0);

	vector<int> m_verticalMatrixInputDimensions;
	vector<int> m_verticalMatrixOutputDimensions;
	vector<int> m_horizontalMatrixInputDimensions;
	vector<int> m_horizontalMatrixOutputDimensions;
	vector<int> m_softmaxMatrixInputDimensions;
	vector<int> m_softmaxMatrixOutputDimensions;
	// Parameters for observation-gate matrices; more generally, these matrices connect variables vertically.
	// Outermost vector ranges over hidden layers.
	// Dimensions of matrix 0: m_inputDimension x m_layerWidths[0] (because they connect the input to the first hidden layer).
	// Dimensions of matrix i > 0: m_layerWidths[i-1] x m_layerWidths[i].
	vector<vector<vector<PARAM_ID> > > m_weightsObsInput, m_weightsObsForget, m_weightsObsCell, m_weightsObsOutput;
	vector<vector<vector<PARAM_ID> > > m_weightsNextObsInput, m_weightsNextObsForget, m_weightsNextObsOutput;
	// Parameters for hidden-gate matrices; more generally, these matrices connect variables horizontally.
	// Outermost vector ranges over hidden layers.
	// Dimensions of matrix i: m_layerWidths[i] x m_layerWidths[i].
	vector<vector<vector<PARAM_ID> > > m_weightsHiddenInput, m_weightsHiddenForget, m_weightsHiddenCell, m_weightsHiddenOutput;

	// Parameters for diagonal matrices connecting cells and gates.
	// Outermost vector ranges over hidden layers.
	// Length of vector i: m_layerWidths[i].
	vector<vector<PARAM_ID> > m_weightsCellInput, m_weightsCellForget, m_weightsCellOutput;

	// Parameters for gate and cell bias vectors.
	// Outermost vector ranges over hidden layers.
	// Length of vector i: m_layerWidths[i].
	vector<vector<PARAM_ID> > m_inputGateBias, m_forgetGateBias, m_cellBias, m_outputGateBias;

	// Pointers to Formulas for cached nodes for current set of sequences, which are NULL if unset (or dropped out).
	// Outermost vector ranges over sequences,
	// then next ranges over time (and so has size datum.m_words.size()),
	// then next ranges over hidden layer, and innermost ranges over units in hidden layer.
	// Dimensions of matrix i: m_numLayers x m_layerWidths[i].
	vector<vector<vector<vector<Formula*> > > > m_cachedCellNodes;
	vector<vector<vector<vector<Formula*> > > > m_cachedHiddenNodes;

	bool m_useLookahead;
};

#endif
