#include <iostream>
#include <stdlib.h>
#include <sys/time.h>
#include "../utils/cmdparams.h"
#include "../fdag/model.h"
#include "classifier_utils.h"

struct ModelParams {
	// The input dimension is the number of features.
	unsigned int m_inputDimension;
	// The number of labels.
	unsigned int m_numLabels;
	// The weights matrix has size numLabels x inputDimension.
	vector<vector<PARAM_ID> > m_weights;
	// The bias vector has size numLabels.
	vector<PARAM_ID> m_bias;
};

void initializeAllParameters(Model& model, ModelParams& params, int inputDimension, int numLabels) {
	params.m_inputDimension = inputDimension;
	params.m_numLabels = numLabels;
	params.m_weights.clear();
	params.m_bias.clear();

	// Initialize all weights to 0.
	// model.initializeNewParameter() returns the PARAM_ID (parameter index) of the new parameter,
	// which we store in the m_weights and m_bias data structures.
	for (unsigned int j = 0; j < params.m_inputDimension; ++j) {
		vector<PARAM_ID> rowj;
		for (unsigned int i = 0; i < params.m_numLabels; ++i) {
			rowj.push_back(model.initializeNewParameter(0.0));
		}
		params.m_weights.push_back(rowj);
	}
	for (unsigned int i = 0; i < params.m_numLabels; ++i) {
		params.m_bias.push_back(model.initializeNewParameter(0.0));
	}
	cerr << "Initialized model. # parameters = " << model.getNumParams() << endl;
}

// Resets data structures for processing the next example (whether in training or decoding).
void resetForNewExample(Model& model) {
	// We're done with all formula nodes we've previously created, so reset them and the cache.
	bool resetCachedParameterFormulasTrue = true;
	model.resetFormulaObjects(resetCachedParameterFormulasTrue);
}

Formula* createScoreFDAGForLabel(Model& model, ModelParams& params, unsigned int i, const Datum& datum) {
	Formula* sumOverFeatures = model.getFormulaObject(PLUS);
	// Add the bias for label i as a child parameter formula.
	sumOverFeatures->addChild(model.getCachedParameterFormula(params.m_bias[i]));
	// Go through all features (input dimensions).
	for (unsigned int j = 0; j < params.m_inputDimension; ++j) {
		Formula* weightTimesFeature = model.getFormulaObject(TIMES);
		// Add the weight for feature j and label i as a child node.
		weightTimesFeature->addChild(model.getCachedParameterFormula(params.m_weights[j][i]));
		// Add feature j as a child node.
		weightTimesFeature->addChild(model.getConstantFormulaObject(datum.m_features.at(j)));
		sumOverFeatures->addChild(weightTimesFeature);
	}
	return sumOverFeatures;
}

Formula* createLossFDAGForDatum(Model& model, ModelParams& params, const Datum& datum, int exampleNumber) {
	// Create a term for the correct label's score, which is negated in the loss.
	Formula* negCorrect = model.getFormulaObject(NEG);

	// Uncomment the two lines below to avoid reusing the same score FDAG twice in the loss (to compare runtime).
	//Formula* scoreCorrect = createScoreFDAGForLabel(model, params, datum.m_label, datum);
	//negCorrect->addChild(scoreCorrect);

	// Create a sum node to sum over labels.
	Formula* sum = model.getFormulaObject(PLUS);
	// Sum over labels.
	for (unsigned int i = 0; i < params.m_numLabels; ++i) {
		Formula* score = createScoreFDAGForLabel(model, params, i, datum);

		// If the current label is the true label, add the score to the negated correct term.
		// Comment out the line below to avoid reusing the same score FDAG twice in the loss (to compare runtime).
		if (i == datum.m_label) negCorrect->addChild(score);

		// Always add the exponentiated label score to the sum over labels.
		Formula* expScore = model.getFormulaObject(EXP);
		expScore->addChild(score);
		sum->addChild(expScore);
	}
	// Put a log in front of the sum of exponentiated label scores.
	Formula* logSum = model.getFormulaObject(LOG);
	logSum->addChild(sum);
	// Add the negated correct score to the log-sum over all exponentiated label scores.
	Formula* negCorrectPlusSum = model.getFormulaObject(PLUS);
	negCorrectPlusSum->addChild(negCorrect);
	negCorrectPlusSum->addChild(logSum);
	// Return the sum.
	return negCorrectPlusSum;
}

Formula* createL2RegularizationTerm(Model& model, ModelParams& modelParams, int numExamples, double C) {
	// The root node is a product node.
	Formula* ret = model.getFormulaObject(TIMES);
	// The first term is the constant multiplier.
	Formula* multiplier = model.getConstantFormulaObject(C / (2.0 * ((double)numExamples)));
	ret->addChild(multiplier);
	Formula* l2term = model.getFormulaObject(PLUS);
	// go through all the parameters
	for (unsigned int j = 0; j < modelParams.m_inputDimension; ++j) {
		for (unsigned int i = 0; i < modelParams.m_numLabels; ++i) {
			// Square the bias weight for feature j and label i.
			Formula* paramSquared = model.getFormulaObject(TIMES);
			paramSquared->addChild(model.getCachedParameterFormula(modelParams.m_weights.at(j).at(i)));
			paramSquared->addChild(model.getCachedParameterFormula(modelParams.m_weights.at(j).at(i)));
			l2term->addChild(paramSquared);
		}
	}
	for (unsigned int i = 0; i < modelParams.m_numLabels; ++i) {
		// Square the bias weight for label i.
		Formula* paramSquared = model.getFormulaObject(TIMES);
		paramSquared->addChild(model.getCachedParameterFormula(modelParams.m_bias.at(i)));
		paramSquared->addChild(model.getCachedParameterFormula(modelParams.m_bias.at(i)));
		l2term->addChild(paramSquared);
	}
	ret->addChild(l2term);
	return ret;
}

// Classifies the given datum by performing an argmax over all labels of the score.
bool classifyDatum(Model& model, ModelParams& params, const Datum& datum, unsigned int exampleNumber) {
	double maxScore = neg_inf;
	int maxLabel = -1;
	// Perform the argmax over all labels.
	for (unsigned int i = 0; i < params.m_numLabels; ++i) {
		Formula* score = createScoreFDAGForLabel(model, params, i, datum);
		double currScore = score->evaluate();
		if (currScore > maxScore) {
			maxScore = currScore;
			maxLabel = i;
		}
	}
	if (maxLabel == -1) {
		cerr << "Error: no label found with score > negative infinity! on line #" << exampleNumber << endl;
	} else {
		if (maxLabel == datum.m_label) return true;
	}
	return false;
}

// Iterates through given data and classifies each datum, then compares predicted label
// to the true label and returns the percent correct.
double classifyData(Model& model, ModelParams& params, const vector<Datum>& data) {
	int numCorrect = 0;
	for (unsigned int i = 0; i < data.size(); ++i) {
		resetForNewExample(model);
		bool correct = classifyDatum(model, params, data[i], i);
		if (correct) ++numCorrect;
	}
	return (((double)numCorrect) / ((double)(data.size())));
}

double trainAndTest(Model& model, ModelParams& params, vector<Datum>& trainData,
		vector<Datum>& testData, double C, bool doFiniteDifferenceChecking, int fold) {
	struct timeval initialTime, endTime;
	gettimeofday(&initialTime, 0);

	double stepsize = 0.1;
	int minibatchSize = 1;
	unsigned int numEpochs = 200;
	double trainAccuracy = 0.0;
	double testAccuracy = 0.0;

	// Initialize sum of squared gradients for AdaGrad.
	vector<double> sumOfSquaredGradients(model.getNumParams(), 0.0);
	// Scale feature values -- divide each by its max value in the training data.
	scaleFeatures(trainData, testData);

	// Train for numEpochs iterations of training.
	for (unsigned int iter = 1; iter <= numEpochs; ++iter) {
		double iterLoss = 0.0;
		double iterGradientNorm = 0.0;
		for (unsigned int exampleIndex = 0; exampleIndex < trainData.size(); ++exampleIndex) {
			// We are processing a new example, so reset Formula data structures in the model.
			resetForNewExample(model);
			// Create the formula containing the loss for the current example.
			Formula* datumLossFDAG = createLossFDAGForDatum(model, params, trainData[exampleIndex], exampleIndex);
			if (C > 0.0) {
				Formula* l2RegularizationTerm = createL2RegularizationTerm(model, params, trainData.size(), C);
				datumLossFDAG->addChild(l2RegularizationTerm);
			}
			// Compute gradient and loss for current example.
			double loss = model.evalAndComputeDerivatives(datumLossFDAG, exampleIndex, doFiniteDifferenceChecking);
			if (loss != loss) {
				cerr << "Error: nan loss, example " << exampleIndex << " with label " << trainData[exampleIndex].m_label << endl;
			} else {
				iterLoss += loss;
			}
			// If we're at the end of a mini-batch or at the end of the training data, make an update.
			bool doUpdate = false;
			if (exampleIndex % minibatchSize == 0 || exampleIndex == trainData.size() - 1) doUpdate = true;
			if (doUpdate) {
				double gradientSquaredNorm = model.doAdaGradUpdate(sumOfSquaredGradients, stepsize);
				iterGradientNorm += sqrt(gradientSquaredNorm);
				model.resetGradientVector();
			}
		}
		if (iter % 10 == 0) cerr << "it " << iter << "\tloss\t" << iterLoss <<
				"\tparam_norm\t" << model.computeParameterVectorNorm() <<
				"\tgrad_norm\t" << iterGradientNorm << endl;
	}
	// Compute final train/test accuracy.
	trainAccuracy = classifyData(model, params, trainData);
	cerr << "Fold " << fold << " train accuracy:\t" << trainAccuracy << endl;
	testAccuracy = classifyData(model, params, testData);
	cerr << "Fold " << fold << " test accuracy:\t" << testAccuracy << endl;
	gettimeofday(&endTime, 0);
	double cumulativeTime = computeTimeDifference(initialTime, endTime);
	cerr << "Fold " << fold << " time:\t" << cumulativeTime << endl;
	return testAccuracy;
}

int main(int argc, char** argv) {
    ParamsArray params(argc, argv);
    params.setDescription("Demonstrates logistic regression using fdag; uses AdaGrad to minimize log loss for a multiclass classification task.");
    params.addConstraint("data_file", "each line should contain comma-delimited real-valued features with label in final field", false); // not optional
    params.addConstraint("C", "L2 regularization coefficient (default: 0.0)", true); // optional
   	if (!params.runConstraints("run_logistic_regression")) return 0;
	// Read in command-line parameters.
	string dataFile = params.asString("data_file");
	double C = atof(params.asString("C", "0.0").c_str());

	// Change following to true to do finite difference checking of gradients.
	bool doFiniteDifferenceChecking = false;
	struct timeval initialTime, endTime;
	gettimeofday(&initialTime, 0);
	// Load data.
	vector<Datum> data;
	int numLabels;
	int inputDimension = loadData(dataFile, data, &numLabels);
	// Do 5-fold cross-validation.
	double accuracySum = 0.0;
	unsigned int dataSize = data.size();
	int foldSize = dataSize / 5;
	for (int i = 0; i < 5; ++i) {
		// Generate train/test sets for the current fold (approximately based on discretized index cutoffs).
		vector<Datum> trainData, testData;
		for (unsigned int j = 0; j < dataSize; ++j) {
			if (((int)j) >= (i*foldSize) && ((int)j) < (i+1)*foldSize) {
				// If the current datum falls within the indices of the current test fold, add it to testData.
				testData.push_back(data[j]);
			} else {
				// Otherwise, add it to the training set.
				trainData.push_back(data[j]);
			}
		}
		// Create a model for the current fold with initial Formula capacity 5000
		// (different values won't really affect runtime here since there are so few Formula's in use).
		Model model(5000);
		// Create and initialize its parameters.
		ModelParams modelParams;
		initializeAllParameters(model, modelParams, inputDimension, numLabels);
		// Train and test.
		accuracySum += trainAndTest(model, modelParams, trainData, testData, C, doFiniteDifferenceChecking, i+1);
	}
	gettimeofday(&endTime, 0);
	double cumulativeTime = computeTimeDifference(initialTime, endTime);
	cerr << "Total training time:\t" << cumulativeTime << endl;
	cerr << "5-fold CV mean accuracy: \t" << (accuracySum / 5.0) << endl;
    return 0;
}
