#include <iostream>
#include <stdlib.h>
#include <sys/time.h>
#include "../utils/cmdparams.h"
#include "../fdag/model.h"
#include "classifier_utils.h"

//static const string UNK_STRING = "UUUNKKK";

//struct ModelParams {
//	vector<PARAM_ID> m_nbowBias;
//	vector<vector<PARAM_ID> > m_softmaxWeights;
//	vector<PARAM_ID> m_softmaxBias;
//	vector<Formula*> m_ciNodes;	// Pointers to Formulas for c_i nodes, which are NULL if unset (or dropped out).
//};

struct ModelParams {
	// Parameters and flags for word vectors:
	bool m_updateWordVectors;				// If true, will include word vector entries as parameters to be learned.
	map<string, vector<double> > m_wordVectors;	// Word vector entries from word vector file.
	map<string, vector<PARAM_ID> > v;		// Word vector parameters, initialized to word vector values from file.
	int m_wordVectorDimensionality;			// Dimensionality of word vectors.
	vector<double>* m_unknownWordVector;	// Pointer to unknown word vector.

	// The input dimension is the number of features.
	unsigned int m_inputDimension;
	// We store separate vectors of the layer input and output dimensions for convenience.
	//vector<unsigned int> m_layerInputSizes;
	//vector<unsigned int> m_layerOutputSizes;
	// The number of labels.
	int m_numLabels;
	// The number of layers.
	//unsigned int m_numLayers;
	// The index to pass as the final layer index when recursively building the FDAGs for all layers.
	//unsigned int m_finalLayerIndex;
	// The width of the (single) hidden layer.
	int m_layerWidth;

	// The weight matrices for the layers have the following sizes:
	// inputDimension x m_layerWidths[0]
	// m_layerWidths[0] x m_layerWidths[1]
	// ...
	// m_layerWidths[m_numLayers - 2] x m_layerWidths[m_numLayers - 1]
	// m_layerWidths[m_numLayers - 1] x m_numLabels   <-- this is for the final softmax layer
	//vector<vector<vector<PARAM_ID> > > m_weights;
	vector<vector<PARAM_ID> > m_weightsObsHidden;
	vector<vector<PARAM_ID> > m_forwardWeightsHiddenHidden;
	//vector<vector<PARAM_ID> > m_forwardWeightsObsHidden;
	vector<vector<PARAM_ID> > m_backwardWeightsHiddenHidden;
	//vector<vector<PARAM_ID> > m_backwardWeightsObsHidden;

	// The bias vectors for the layers have the following sizes:
	// m_layerWidths[0]
	// m_layerWidths[1]
	// ...
	// m_layerWidths[m_numLayers - 1]
	// m_numLabels   <-- this is for the final softmax layer
	//vector<vector<PARAM_ID> > m_bias;
	vector<PARAM_ID> m_forwardBias;
	vector<PARAM_ID> m_backwardBias;

	// The softmax weight matrix has dimension layerWidth x numLabels.
	vector<vector<PARAM_ID> > m_forwardSoftmaxWeights;
	vector<vector<PARAM_ID> > m_backwardSoftmaxWeights;
	// The softmax bias vector has length numLabels.
	vector<PARAM_ID> m_softmaxBias;

	// Pointers to Formulas for cached nodes for current example, which are NULL if unset (or dropped out).
	// Dimensionality of m_cachedNodes should be datum.m_words.size() x params.m_layerWidth.
	vector<vector<Formula*> > m_cachedForwardNodes;
	vector<vector<Formula*> > m_cachedBackwardNodes;

	bool m_useForwardRNN;
	bool m_useBackwardRNN;
};

void initializeRNNDirectionParameters(Model& model, ModelParams& params, vector<vector<PARAM_ID> >& weightsHiddenHidden, vector<PARAM_ID>& bias,
		vector<vector<PARAM_ID> >& softmaxWeights, double scalingFactor) {
	// The weight matrix that connects consecutive hidden vectors has dimension params.m_layerWidth x params.m_layerWidth.
	weightsHiddenHidden.clear();
	weightsHiddenHidden.reserve(params.m_layerWidth);
	for (unsigned int i = 0; i < params.m_layerWidth; ++i) {
		vector<PARAM_ID> rowi;
		rowi.reserve(params.m_layerWidth);
		for (unsigned int j = 0; j < params.m_layerWidth; ++j) {
			rowi.push_back(model.initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
		}
		weightsHiddenHidden.push_back(rowi);
	}

	// The bias vector for the hidden vector has length params.m_layerWidth.
	bias.clear();
	bias.reserve(params.m_layerWidth);
	for (unsigned int i = 0; i < params.m_layerWidth; ++i) {
		bias.push_back(model.initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
	}

	// The softmax weight matrix has dimension layerWidth x numLabels.
	softmaxWeights.clear();
	softmaxWeights.reserve(params.m_layerWidth);
	for (unsigned int i = 0; i < params.m_layerWidth; ++i) {
		vector<PARAM_ID> rowi;
		rowi.reserve(params.m_numLabels);
		for (unsigned int label = 0; label < params.m_numLabels; ++label) {
			rowi.push_back(model.initializeNewParameter(getRandomDouble() - 0.5));
		}
		softmaxWeights.push_back(rowi);
	}

}

void initializeAllParameters(Model& model, ModelParams& params, int inputDimension, const string& rnnDirections, const string& layerWidths,
		const vector<WordSequenceDatum>& trainData, const vector<WordSequenceDatum>& devData, const vector<WordSequenceDatum>& testData,
		bool initWordVectorsRandomly) {
	params.m_inputDimension = inputDimension;
	vector<string> rnnDirectionsStrings;
	tokenize(rnnDirections, ',', rnnDirectionsStrings);
	params.m_useForwardRNN = false;
	params.m_useBackwardRNN = false;
	for (unsigned int i = 0; i < rnnDirectionsStrings.size(); ++i) {
		if (lowercase(rnnDirectionsStrings[i]) == "forward") {
			params.m_useForwardRNN = true;
		} else if (lowercase(rnnDirectionsStrings[i]) == "backward") {
			params.m_useBackwardRNN = true;
		} else {
			cerr << "Error: unexpected string in list of RNN directions to use: " << rnnDirectionsStrings[i] << ". Expected \"forward\" or \"backward\"." << endl;
		}
	}
	//vector<string> layerWidthStrings;
	//tokenize(layerWidths, ',', layerWidthStrings);
	//if (layerWidths == "0") {
	//	params.m_numLayers = 0;
	//} else {
	//	params.m_numLayers = layerWidthStrings.size();
	//}
	params.m_numLabels = 2;

	// Fill the vectors of input and output layer widths.
	//params.m_layerInputSizes.push_back(params.m_inputDimension);
	//for (unsigned int i = 0; i < params.m_numLayers; ++i) {
	//	int currentLayerWidth = atoi(layerWidthStrings[i].c_str());
	//	params.m_layerInputSizes.push_back(currentLayerWidth);
	//	params.m_layerOutputSizes.push_back(currentLayerWidth);
	//}
	params.m_layerWidth = atoi(layerWidths.c_str());
	//params.m_layerOutputSizes.push_back(params.m_numLabels);

	//params.m_finalLayerIndex = params.m_layerInputSizes.size();

	//if (params.m_layerInputSizes.size() != params.m_numLayers + 1) {
	//	cerr << "Error: number of layer input sizes (" << params.m_layerInputSizes.size()
	//		 << ") not equal to 1 + number of layers (" << (params.m_numLayers + 1) << ")" << endl;
	//}
	//if (params.m_layerOutputSizes.size() != params.m_numLayers + 1) {
	//	cerr << "Error: number of layer output sizes (" << params.m_layerOutputSizes.size()
	//		 << ") not equal to 1 + number of layers (" << (params.m_numLayers + 1) << ")" << endl;
	//}

	//for (unsigned int i = 0; i < params.m_layerInputSizes.size(); ++i) {
	//	cerr << "m_layerInputSizes[" << i << "] = " << params.m_layerInputSizes[i] << endl;
	//	cerr << "m_layerOutputSizes[" << i << "] = " << params.m_layerOutputSizes[i] << endl;
	//}

	// Below we initialize all weights to a random double between -0.01 and 0.01.
	// model.initializeNewParameter() returns the PARAM_ID (parameter index) of the new parameter,
	// which we store in the m_weights and m_bias data structures.

	double scalingFactor = 0.2;
	// The weight matrix that connects the input to the hidden vectors has dimension params.m_inputDimension x params.m_layerWidth.
	params.m_weightsObsHidden.clear();
	params.m_weightsObsHidden.reserve(params.m_inputDimension);
	for (unsigned int i = 0; i < params.m_inputDimension; ++i) {
		vector<PARAM_ID> rowi;
		rowi.reserve(params.m_layerWidth);
		for (unsigned int j = 0; j < params.m_layerWidth; ++j) {
			rowi.push_back(model.initializeNewParameter(scalingFactor * (getRandomDouble() - 0.5)));
		}
		params.m_weightsObsHidden.push_back(rowi);
	}

	if (params.m_useForwardRNN) {
		initializeRNNDirectionParameters(model, params, params.m_forwardWeightsHiddenHidden, params.m_forwardBias, params.m_forwardSoftmaxWeights, scalingFactor);
	}
	if (params.m_useBackwardRNN) {
		initializeRNNDirectionParameters(model, params, params.m_backwardWeightsHiddenHidden, params.m_backwardBias, params.m_backwardSoftmaxWeights, scalingFactor);
	}

	// The softmax bias vector has length params.m_numLabels.
	params.m_softmaxBias.clear();
	params.m_softmaxBias.reserve(params.m_numLabels);
	for (unsigned int label = 0; label < params.m_numLabels; ++label) {
		params.m_softmaxBias.push_back(model.initializeNewParameter(getRandomDouble() - 0.5));
	}

	if (params.m_updateWordVectors) {
		//params.v;//map<string, vector<PARAM_ID> > v
		// New way: only create parameters for words in training data.
		set<string> trainWords;
		for (unsigned int i = 0; i < trainData.size(); ++i) {
			// Go through all words.
			for (unsigned int j = 0; j < trainData[i].m_words.size(); ++j) {
				trainWords.insert(trainData[i].m_words[j]);
			}
			//trainWords.insert(trainData[i].m_words[1]);
		}
		for (unsigned int i = 0; i < devData.size(); ++i) {
			for (unsigned int j = 0; j < devData[i].m_words.size(); ++j) {
				trainWords.insert(devData[i].m_words[j]);
			}
			//trainWords.insert(devData[i].m_words[1]);
		}
		for (unsigned int i = 0; i < testData.size(); ++i) {
			for (unsigned int j = 0; j < testData[i].m_words.size(); ++j) {
				trainWords.insert(testData[i].m_words[j]);
			}
			//trainWords.insert(devData[i].m_words[1]);
		}
		//trainWords.insert(UNK_STRING);

		for (set<string>::const_iterator iter = trainWords.begin(); iter != trainWords.end(); ++iter) {
			const string& currWord = *iter;
			bool unk = false;
			const vector<double>* currVector = getWordVector(params.m_wordVectors, currWord, unk);
			vector<PARAM_ID> currVectorParams;
			// Go through all dimensions of the current word vector.
			for (unsigned int i = 0; i < currVector->size(); ++i) {
				// Create a new parameter for the current word vector entry.
				if (initWordVectorsRandomly) {
					currVectorParams.push_back(model.initializeNewParameter(getRandomDouble() - 0.5));
				} else {
					currVectorParams.push_back(model.initializeNewParameter(currVector->at(i)));
				}
			}
			params.v.insert(make_pair(currWord, currVectorParams));
		}
	}

	cerr << "Initialized model. # parameters = " << model.getNumParams() << endl;
}

void resetCachedNodes(Model& model, ModelParams& params, unsigned int numCachedNodes, vector<vector<Formula*> >& cachedNodes) {
	// Clear all data structures containing cached formulas from the previous datum, and set them up for the new datum.
	if (cachedNodes.size() == numCachedNodes) {
		for (unsigned int i = 0; i < cachedNodes.size(); ++i) {
			if (cachedNodes[i].size() != params.m_layerWidth) {
				cerr << "Error: params.m_cachedNodes[" << i << "].size() != " << params.m_layerWidth << endl;
			} else {
				for (unsigned int j = 0; j < cachedNodes[i].size(); ++j) {
					cachedNodes[i][j] = NULL;
				}
			}
		}
	} else {
		cachedNodes.clear();
		cachedNodes.reserve(numCachedNodes);
		for (unsigned int i = 0; i < numCachedNodes; ++i) {
			vector<Formula*> rowi(params.m_layerWidth, NULL);
			cachedNodes.push_back(rowi);
		}
	}
}

// Resets data structures for processing the next example (whether in training or decoding).
void resetForNewExample(Model& model, ModelParams& params, WordSequenceDatum& datum) {
	// We're done with all formula nodes we've previously created, so reset them and the cache.
	bool resetCachedParameterFormulasTrue = true;
	model.resetFormulaObjects(resetCachedParameterFormulasTrue);

	// The number of cached nodes is the length of the sequence times the hidden layer width.
	unsigned int numCachedNodes = datum.m_words.size();//params.m_inputDimension;

	if (params.m_useForwardRNN) {
		resetCachedNodes(model, params, numCachedNodes, params.m_cachedForwardNodes);
	}
	if (params.m_useBackwardRNN) {
		resetCachedNodes(model, params, numCachedNodes, params.m_cachedBackwardNodes);
	}
	precomputeWordVectorFormulas(&model, datum, params.m_inputDimension, params.m_updateWordVectors);
}

Formula* createHiddenNodeAtTime(Model& model, ModelParams& params, WordSequenceDatum& datum, unsigned int i, unsigned int t,
		bool doDropout, double dropoutRate, bool forward, bool scaleWeights, bool testTime) {
	if (forward) {
		if (params.m_cachedForwardNodes.at(t).at(i) != NULL) return params.m_cachedForwardNodes.at(t).at(i);
	} else {
		if (params.m_cachedBackwardNodes.at(t).at(i) != NULL) return params.m_cachedBackwardNodes.at(t).at(i);
	}

	// Sum over all dimensions of the hidden vector at the previous time step, and all dimensions of the
	// data point at the previous time step, as well as the bias weight.
	int numTermsInSum = params.m_inputDimension + 1;
	if ((forward && t > 0) || (!forward && t < (datum.m_words.size()-1))) {
		numTermsInSum += params.m_layerWidth;
	}
	//if (params.m_useForwardRNN) numTermsInSum += (params.m_layerWidth + 1);
	//if (params.m_useBackwardRNN) numTermsInSum += (params.m_layerWidth + 1);

	Formula* sumOverDimensions = model.getFormulaObject(PLUS, numTermsInSum);

	if ((forward && t > 0) || (!forward && t < (datum.m_words.size()-1))) {
		// If we're past the first time step, sum over the hidden vector in the previous time step.
		for (unsigned int j = 0; j < params.m_layerWidth; ++j) {
			bool dropout = false;
			if (doDropout && getRandomDouble() < dropoutRate) dropout = true;
			if (!dropout) {
				// Create the Formula node for the jth dimension of the final hidden vector (i.e., the time t is the number of words in the sentence).
				int prevTime;
				if (forward) {
					prevTime = t - 1;
				} else {
					prevTime = t + 1;
				}
				Formula* zj = createHiddenNodeAtTime(model, params, datum, j, prevTime, doDropout, dropoutRate, forward, scaleWeights, testTime);
				// Multiply it by the weight corresponding to the provided label and the current dimension i.
				Formula* prod = model.getFormulaObject(TIMES, 2);
				prod->addChild(zj);
				PARAM_ID currentWeightID;
				if (forward) {
					currentWeightID = params.m_forwardWeightsHiddenHidden.at(j).at(i);
				} else {
					currentWeightID = params.m_backwardWeightsHiddenHidden.at(j).at(i);
				}
				if (scaleWeights) {
					double currWeight = model.getValue(currentWeightID);
					Formula* weightConstant = model.getConstantFormulaObject(currWeight * dropoutRate);
					prod->addChild(weightConstant);
				} else {
					prod->addChild(model.getCachedParameterFormula(currentWeightID));
				}
				sumOverDimensions->addChild(prod);
			}
		}
	}
	// Always sum over the dimensions in the datum at the current time step.
	for (unsigned int j = 0; j < params.m_inputDimension; ++j) {
		bool dropout = false;
		if (doDropout && getRandomDouble() < dropoutRate) dropout = true;
		if (!dropout) {
			// Create the Formula node for the jth dimension of the datum at time step t.
			//Formula* zj = createHiddenNodeAtTime(model, params, datum, j, datum.m_words.size(), doDropout, dropoutRate, scaleWeights, testTime);
			// Multiply it by the weight corresponding to the provided label and the current dimension i.
			Formula* prod = model.getFormulaObject(TIMES, 2);
			PARAM_ID currentWeightID = params.m_weightsObsHidden.at(j).at(i);
			if (scaleWeights) {
				double currWeight = model.getValue(currentWeightID);
				Formula* weightConstant = model.getConstantFormulaObject(currWeight * dropoutRate);
				prod->addChild(weightConstant);
			} else {
				prod->addChild(model.getCachedParameterFormula(currentWeightID));
			}
			Formula* datumFormula = datum.m_vectFormulas.at(t).at(j);
			prod->addChild(datumFormula);
			sumOverDimensions->addChild(prod);
		}
	}

	if (forward) {
		sumOverDimensions->addChild(model.getCachedParameterFormula(params.m_forwardBias.at(i)));
	} else {
		sumOverDimensions->addChild(model.getCachedParameterFormula(params.m_backwardBias.at(i)));
	}
	Formula* activation = model.getFormulaObject(TANH);
	activation->addChild(sumOverDimensions);
	if (forward) {
		params.m_cachedForwardNodes[t][i] = activation;
	} else {
		params.m_cachedBackwardNodes[t][i] = activation;
	}
	//params.m_cachedNodes[layer - 1][outputDimension] = activation;
	return activation;
}

Formula* createSoftmaxScoreForLabel(Model& model, ModelParams& params, unsigned int label, WordSequenceDatum& datum,
		bool doDropout, double dropoutRate, bool scaleWeights, bool testTime) {
	// Sum over all dimensions of the final hidden vector, as well as the bias weight.
	int numTermsInSum = 1;
	if (params.m_useForwardRNN) numTermsInSum += params.m_layerWidth;
	if (params.m_useBackwardRNN) numTermsInSum += params.m_layerWidth;

	Formula* sumOverDimensions = model.getFormulaObject(PLUS, numTermsInSum);
	if (params.m_useForwardRNN) {
		bool forwardTrue = true;
		for (unsigned int j = 0; j < params.m_layerWidth; ++j) {
			bool dropout = false;
			if (doDropout && getRandomDouble() < dropoutRate) dropout = true;
			if (!dropout) {
				// Create the Formula node for the jth dimension of the final hidden vector (i.e., the time t is the number of words in the sentence).
				Formula* zj = createHiddenNodeAtTime(model, params, datum, j, ((int)(datum.m_words.size())) - 1, doDropout, dropoutRate, forwardTrue, scaleWeights, testTime);
				// Multiply it by the weight corresponding to the provided label and the current dimension i.
				Formula* prod = model.getFormulaObject(TIMES, 2);
				prod->addChild(zj);
				if (scaleWeights) {
					double currWeight = model.getValue(params.m_forwardSoftmaxWeights[j][label]);
					Formula* weightConstant = model.getConstantFormulaObject(currWeight * dropoutRate);
					prod->addChild(weightConstant);
				} else {
					prod->addChild(model.getCachedParameterFormula(params.m_forwardSoftmaxWeights.at(j).at(label)));
				}
				sumOverDimensions->addChild(prod);
			}
		}
	}
	if (params.m_useBackwardRNN) {
		bool forwardFalse = false;
		for (unsigned int j = 0; j < params.m_layerWidth; ++j) {
			bool dropout = false;
			if (doDropout && getRandomDouble() < dropoutRate) dropout = true;
			if (!dropout) {
				// Create the Formula node for the jth dimension of the final hidden vector (i.e., the time t is the number of words in the sentence).
				Formula* zj = createHiddenNodeAtTime(model, params, datum, j, 0, doDropout, dropoutRate, forwardFalse, scaleWeights, testTime);
				// Multiply it by the weight corresponding to the provided label and the current dimension i.
				Formula* prod = model.getFormulaObject(TIMES, 2);
				prod->addChild(zj);
				if (scaleWeights) {
					double currWeight = model.getValue(params.m_backwardSoftmaxWeights[j][label]);
					Formula* weightConstant = model.getConstantFormulaObject(currWeight * dropoutRate);
					prod->addChild(weightConstant);
				} else {
					prod->addChild(model.getCachedParameterFormula(params.m_backwardSoftmaxWeights.at(j).at(label)));
				}
				sumOverDimensions->addChild(prod);
			}
		}
	}
	// Add the softmax bias.
	sumOverDimensions->addChild(model.getCachedParameterFormula(params.m_softmaxBias.at(label)));
	return sumOverDimensions;
}

Formula* createLossFDAGForWordSequenceDatum(Model& model, ModelParams& params, WordSequenceDatum& datum, int exampleNumber, bool doDropout, double dropoutRate) {
	bool scaleWeightsFalse = false, testTimeFalse = false;
	// Create a term for the correct label's score, which is negated in the loss.
	Formula* negCorrect = model.getFormulaObject(NEG);
	//negCorrect->addChild(createSoftmaxScoreForLabel(dim, m, w, datum.m_label, datum));
	// Create a sum node to sum over labels.
	Formula* sum = model.getFormulaObject(PLUS, params.m_numLabels);
	// Sum over labels.
	for (unsigned int label = 0; label < params.m_numLabels; ++label) {
		Formula* score = createSoftmaxScoreForLabel(model, params, label, datum, doDropout, dropoutRate, scaleWeightsFalse, testTimeFalse);
		// If the current label is the true label, add the score to the negated correct term.
		if (label == datum.m_label) negCorrect->addChild(score);
		// Always add the exponentiated label score to the sum over labels.
		Formula* expScore = model.getFormulaObject(EXP);
		expScore->addChild(score);
		sum->addChild(expScore);
	}
	// Put a log in front of the sum of exponentiated label scores.
	Formula* logSum = model.getFormulaObject(LOG);
	logSum->addChild(sum);

	// Add the negated correct score to the log-sum over all exponentiated label scores.
	Formula* negCorrectPlusSum = model.getFormulaObject(PLUS, 2);
	negCorrectPlusSum->addChild(negCorrect);
	negCorrectPlusSum->addChild(logSum);
	// Return the sum.
	return negCorrectPlusSum;
}

bool decode(Model& model, ModelParams& params, WordSequenceDatum& datum, bool scaleWeights, double dropoutRate, int exampleNumber) {
	bool doDropoutFalse = false;
	bool testTimeTrue = true;
	double maxScore = neg_inf;
	int maxLabel = -1;
	for (unsigned int i = 0; i < params.m_numLabels; ++i) {
		Formula* score = createSoftmaxScoreForLabel(model, params, i, datum, doDropoutFalse, dropoutRate, scaleWeights, testTimeTrue);
		double currScore = score->evaluate();
		if (currScore > maxScore) {
			maxScore = currScore;
			maxLabel = i;
		}
	}
	if (maxLabel == -1) {
		cerr << "Error: no label found with score > negative infinity! on line #" << exampleNumber << endl;
	} else {
		if (maxLabel == datum.m_label) return true;
	}
	return false;
}

double doDecoding(Model& model, ModelParams& params, vector<WordSequenceDatum>& data, bool scaleWeights, double dropoutRate) {
	int numCorrect = 0;
	int total = 0;
	for (unsigned int i = 0; i < data.size(); ++i) {
		WordSequenceDatum& datum = data[i];
		// We are processing a new example, so reset Formula data structures in the model.
		resetForNewExample(model, params, datum);
		bool correct = decode(model, params, datum, scaleWeights, dropoutRate, i);
		datum.doneProcessing();
		if (correct) ++numCorrect;
		++total;
	}
	return (((double)numCorrect) / ((double)total));
}

struct CompareIntDoublePairSet {
	bool operator() (const pair<int, double>& a, const pair<int, double>& b) const {
		return a.second > b.second;
	}
};

void printRecurrentNeuralNetworkNearestNeighbors(vector<WordSequenceDatum>& data, Model& model, ModelParams& params) {
	// Go through the data and embed each sequence.
	bool doDropoutFalse = false, scaleWeightsFalse = false, testTimeFalse = false;
	bool forwardTrue = true, forwardFalse = false;

	vector<vector<double> > sequenceEmbeddings;
	int numberToPrint = 200;
	if (numberToPrint > data.size()) numberToPrint = data.size();
	sequenceEmbeddings.reserve(numberToPrint);
	for (unsigned int i = 0; i < numberToPrint; ++i) {
		WordSequenceDatum& datum = data[i];
		// We are processing a new example, so reset Formula data structures in the model.
		resetForNewExample(model, params, datum);
		// Get the RNN vector for the sequence.
		vector<double> currEmbedding;
		int sizeToReserve = 0;
		if (params.m_useForwardRNN) sizeToReserve += params.m_layerWidth;
		if (params.m_useBackwardRNN) sizeToReserve += params.m_layerWidth;

		currEmbedding.reserve(sizeToReserve);
		if (params.m_useForwardRNN) {
			for (unsigned int j = 0; j < params.m_layerWidth; ++j) {
				// Create the Formula node for the jth dimension of the final hidden vector (i.e., the time t is the number of words in the sentence).
				Formula* zj = createHiddenNodeAtTime(model, params, datum, j, ((int)(datum.m_words.size())) - 1, doDropoutFalse, -1.0, forwardTrue, scaleWeightsFalse, testTimeFalse);
				double currEntry = zj->evaluate();
				currEmbedding.push_back(currEntry);
			}
		}
		if (params.m_useBackwardRNN) {
			bool forwardFalse = false;
			for (unsigned int j = 0; j < params.m_layerWidth; ++j) {
				// Create the Formula node for the jth dimension of the final hidden vector (i.e., the time t is the number of words in the sentence).
				Formula* zj = createHiddenNodeAtTime(model, params, datum, j, 0, doDropoutFalse, -1.0, forwardFalse, scaleWeightsFalse, testTimeFalse);
				double currEntry = zj->evaluate();
				currEmbedding.push_back(currEntry);
			}
		}
		sequenceEmbeddings.push_back(currEmbedding);
	}

	// Then go back through embedded sequences and compute all pairwise cosine similarities.
	vector<vector<double> > similarities;
	similarities.reserve(numberToPrint);
	for (unsigned int i = 0; i < numberToPrint; ++i) {
		vector<double> emptySimilarities(numberToPrint, -99.0);
		similarities.push_back(emptySimilarities);
	}
	for (unsigned int i = 0; i < sequenceEmbeddings.size(); ++i) {
		for (unsigned int j = i; j < sequenceEmbeddings.size(); ++j) {
			// Compute cosine similarity between embeddings of sequences i and j.
			double currSim = cosineSimilarity(sequenceEmbeddings[i], sequenceEmbeddings[j]);
			similarities[i][j] = currSim;
			similarities[j][i] = currSim;
		}
	}

	// For each sequence, find its k-nearest neighbors.
	unsigned int numNeighbors = 10;
	cerr << "Printing k-nearest neighbors of sequence embeddings:" << endl;
	for (unsigned int i = 0; i < similarities.size(); ++i) {
		set<pair<int, double>, CompareIntDoublePairSet > pq;
		for (unsigned int j = 0; j < similarities[i].size(); ++j) {
			double currSim = similarities[i][j];
			// Insert the current similarity into the priority queue.
			pq.insert(make_pair(j, currSim));
		}
		cerr << "For sequence:" << endl;
		cerr << data[i].toString() << endl;
		for (unsigned int k = 0; k < numNeighbors && pq.size() > 0; ++k) {
			//string currWordSequence = pq.top().first;
			//double filterSum = pq.top().second;
			int datumIndex = pq.begin()->first;
			double datumSim = pq.begin()->second;
			cerr << datumSim << ":\t" << data[datumIndex].toString() << endl;
			//topMatches.push_back(currWordSequence);
			pq.erase(pq.begin());
			//pq.pop();
		}

	}
}

double trainAndTest(Model& model, const vector<string>& trainLines, const vector<string>& devLines,
		const vector<string>& testLines, const string& wordVectorFilename, const string& rnnDirections, const string& layerWidths,
		const string& algorithm, double stepsize, int minibatchSize, int numEpochs, double dropoutRate,
		bool updateWordVectors, bool initWordVectorsRandomly, double gradientNormThreshold, bool doFiniteDifferenceChecking) {
	struct timeval initialTime, endTime, time1, time2, time3, time4,
		timeBuildStart, timeBuildEnd, timeEvalStart, timeEvalEnd,
		timeDecodeStart, timeDecodeEnd;
	gettimeofday(&initialTime, 0);
	ModelParams params;
	params.m_updateWordVectors = updateWordVectors;

	set<string> vocab;
	// Build the relevant word vector vocabulary from the data lines.
	addWordsFromLinesToVocabulary(trainLines, vocab);
	addWordsFromLinesToVocabulary(devLines, vocab);
	addWordsFromLinesToVocabulary(testLines, vocab);
	// Read in word vectors corresponding to the entries in the vocabulary.
	readWordVectorFile(wordVectorFilename, vocab, params.m_wordVectors, params.m_wordVectorDimensionality, params.m_unknownWordVector);
	//readWordVectorFile(wordVectorFilename, vocab);
	// Clear the vocabulary to save memory.
	vocab.clear();

	model.resetGradientVector();
	//double stepsize = 0.001;
	map<string, int> unkWordCounts;
	vector<WordSequenceDatum> trainData, devData, testData;
	loadWordSequenceData(trainLines, trainData, params.m_wordVectors, "train", unkWordCounts);
	loadWordSequenceData(devLines, devData, params.m_wordVectors, "dev", unkWordCounts);
	loadWordSequenceData(testLines, testData, params.m_wordVectors, "test", unkWordCounts);
	printUnknownWordCounts(unkWordCounts);

	bool trainTrue = true;
	bool trainFalse = false;

	initializeAllParameters(model, params, params.m_wordVectorDimensionality, rnnDirections, layerWidths, trainData, devData, testData, initWordVectorsRandomly);

	if (params.m_updateWordVectors) {
		cerr << "Getting word vector params for training data" << endl;
		getWordVectorParamsForWordSequenceData(params.v, trainData, trainTrue);
		cerr << "Getting word vector params for dev data" << endl;
		getWordVectorParamsForWordSequenceData(params.v, devData, trainFalse);
		cerr << "Getting word vector params for test data" << endl;
		getWordVectorParamsForWordSequenceData(params.v, testData, trainFalse);
	}

	// Print nearest neighbors before any learning.
	//printNearestNeighbors(trainData, trainData, params);
	//printNearestNeighbors(trainData, devData, params);
	double maxDevAcc = -1.0;
	double testAccForMaxDevAcc = -1.0;
	//double maxTestAcc = -1.0;

	FLOAT totalBuildTime = 0.0f;
	FLOAT totalEvalTime = 0.0f;
	FLOAT totalDecodeTime = 0.0f;

	int decodeEveryK = 3000;
	if (devData.size() < 1000) {
		decodeEveryK = trainData.size() / 5;
	}

	bool doDropout = true;
	if (dropoutRate <= 0.0) doDropout = false;

	bool scaleWeightsForDecoding = doDropout;
	scaleWeightsForDecoding = false;
	if (scaleWeightsForDecoding) {
		cerr << "Scaling weights for decoding by dropout rate (" << dropoutRate << ")." << endl;
	}

	Algorithm alg = Model::getAlgorithm(algorithm);
	// Initialize vector of previous updates for SGD w/ momentum.
	vector<double> sgdMomentumUpdates;//(model.getNumParams(), 0.0);
	double momentumCoefficient = 0.95;
	// Initialize sum of squared gradients for AdaGrad and AdaDelta.
	vector<double> sumOfSquaredGradients(model.getNumParams(), 0.0);
	// Initialize sum of squared updates for AdaDelta.
	vector<double> sumOfSquaredUpdates;//(model.getNumParams(), 0.0);
	//double epsilon = 0.000001;
	//double rho = 0.95;
	if (alg != ADAGRAD) {
		cerr << "Error: only AdaGrad is currently supported!" << endl;
	}

	// Initialize sum of squared gradients for AdaGrad.
	//vector<double> sumOfSquaredGradients(getNumParams(), 0.0);
	//int maxNumFormulasUsed = getNumberOfFormulasUsedInPool();
	int maxUnaryFormulasUsed = 0;
	int maxBinaryFormulasUsed = 0;
	int maxFertileFormulasUsed = 0;
	int maxBarrenFormulasUsed = 0;
	int unaryFormulasUsed = 0;
	int binaryFormulasUsed = 0;
	int fertileFormulasUsed = 0;
	int barrenFormulasUsed = 0;
	model.getNumberOfFormulasUsedInPool(maxUnaryFormulasUsed, maxBinaryFormulasUsed, maxFertileFormulasUsed, maxBarrenFormulasUsed);

	// Train for numEpochs iterations of training.
	for (unsigned int iter = 1; iter <= numEpochs; ++iter) {
		double iterLoss = 0.0;
		double iterGradientNorm = 0.0;

		//int iterNumCorrect = 0;
		gettimeofday(&time1, 0);
		gettimeofday(&time3, 0);

		for (unsigned int exampleIndex = 0; exampleIndex < trainData.size(); ++exampleIndex) {
			WordSequenceDatum& datum = trainData[exampleIndex];
			// Sample a data point.
			//int exampleIndex = i;//rand() % (trainData.size());
			// We are processing a new example, so reset Formula data structures in the model.
			resetForNewExample(model, params, datum);

			//cerr << "Computing gradient for sentence \"" << sentenceTokens[0] << " " << sentenceTokens[1]
			//	 << "\" with true label " << trueLabel << endl;

			// Create the formula containing the loss for the current example.
			gettimeofday(&timeBuildStart, 0);
			Formula* datumLossFDAG = createLossFDAGForWordSequenceDatum(model, params, datum, exampleIndex, doDropout, dropoutRate);
			gettimeofday(&timeBuildEnd, 0);
			totalBuildTime += computeTimeDifference(timeBuildStart, timeBuildEnd);
			model.getNumberOfFormulasUsedInPool(unaryFormulasUsed, binaryFormulasUsed, fertileFormulasUsed, barrenFormulasUsed);
			if (unaryFormulasUsed > maxUnaryFormulasUsed) {
				maxUnaryFormulasUsed = unaryFormulasUsed;
				cerr << "Info: New unary formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << unaryFormulasUsed << endl;
			}
			if (binaryFormulasUsed > maxBinaryFormulasUsed) {
				maxBinaryFormulasUsed = binaryFormulasUsed;
				cerr << "Info: New binary formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << binaryFormulasUsed << endl;
			}
			if (fertileFormulasUsed > maxFertileFormulasUsed) {
				maxFertileFormulasUsed = fertileFormulasUsed;
				cerr << "Info: New fertile formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << fertileFormulasUsed << endl;
			}
			if (barrenFormulasUsed > maxBarrenFormulasUsed) {
				maxBarrenFormulasUsed = barrenFormulasUsed;
				cerr << "Info: New barren formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << barrenFormulasUsed << endl;
			}
			//int currNumFormulasUsed = getNumberOfFormulasUsedInPool();
			//if (currNumFormulasUsed > maxNumFormulasUsed) {
			//	cerr << "Info: New formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << currNumFormulasUsed << endl;
			//	maxNumFormulasUsed = currNumFormulasUsed;
			//}
			// Compute gradient and loss for current example.
			gettimeofday(&timeEvalStart, 0);
			double loss = model.evalAndComputeDerivatives(datumLossFDAG, exampleIndex, doFiniteDifferenceChecking);
			if (loss != loss) {
				//cerr << "Error: nan loss, example " << exampleIndex << " with label " << trainData[exampleIndex].m_label << endl;
				cerr << "Error: nan loss, at sentence " << exampleIndex << ":\"" << trainData[exampleIndex].toString() <<
					"\" with true label " << trainData[exampleIndex].m_label << endl;
			} else if (loss == neg_inf) {
					//cerr << "Error: nan loss, example " << exampleIndex << " with label " << trainData[exampleIndex].m_label << endl;
					cerr << "Error: -inf loss, at sentence " << exampleIndex << ":\"" << trainData[exampleIndex].toString() <<
						"\" with true label " << trainData[exampleIndex].m_label << endl;
			} else if (loss == pos_inf) {
					//cerr << "Error: nan loss, example " << exampleIndex << " with label " << trainData[exampleIndex].m_label << endl;
					cerr << "Error: inf loss, at sentence " << exampleIndex << ":\"" << trainData[exampleIndex].toString() <<
						"\" with true label " << trainData[exampleIndex].m_label << endl;
			} else {
				iterLoss += loss;
			}
			gettimeofday(&timeEvalEnd, 0);
			totalEvalTime += computeTimeDifference(timeEvalStart, timeEvalEnd);
			datum.doneProcessing();

			// If we're at the end of a mini-batch or at the end of the training data, make an update.
			bool doUpdate = false;
			if ((exampleIndex + 1) % minibatchSize == 0 || trainData.size() <= 1 || exampleIndex == trainData.size() - 1) doUpdate = true;
			if (doUpdate) {
				if (gradientNormThreshold > 0.0) {
					double gradientNorm = model.computeGradientVectorNorm();
					if (gradientNorm >= gradientNormThreshold) {
						//cerr << "Info: gradientNorm = " << gradientNorm << "; Scaling gradient vector by " << (gradientNormThreshold / gradientNorm) << endl;
						model.scaleGradientVectorByConstant(gradientNormThreshold / gradientNorm);
						//cerr << "Info: new gradientNorm = " << model.computeGradientVectorNorm() << endl;
					}
				}
				double gradientSquaredNorm = 0.0;
				if (alg == SGD_MOMENTUM) {
					gradientSquaredNorm = model.doSGDMomentumUpdate(sgdMomentumUpdates, stepsize, momentumCoefficient);
				} else if (alg == ADAGRAD) {
					gradientSquaredNorm = model.doAdaGradUpdate(sumOfSquaredGradients, stepsize);
				} else if (alg == ADADELTA){
					gradientSquaredNorm = model.doAdaDeltaUpdate(sumOfSquaredGradients, sumOfSquaredUpdates, stepsize, momentumCoefficient);
				}
				//double l2norm = sqrt(gradientSquaredNorm);
				//iterGradientNorm += l2norm;
				iterGradientNorm += model.computeGradientVectorNorm();
				//if (l2NormConstraint > 0.0) {// && l2norm > l2NormConstraint) {
				//	scaleWeightsToMatchL2Norm(params, l2NormConstraint);
				//}
				model.resetGradientVector();

				//double gradientSquaredNorm = doAdaGradUpdate(sumOfSquaredGradients, stepsize);
				//double l2norm = sqrt(gradientSquaredNorm);
				//iterGradientNorm += l2norm;
				//if (l2NormConstraint > 0.0) {// && l2norm > l2NormConstraint) {
				//	scaleWeightsToMatchL2Norm(params, l2NormConstraint);
				//}
				//resetGradientVector();
			}

			if ((exampleIndex + 1) % decodeEveryK == 0) {
				gettimeofday(&timeDecodeStart, 0);
				gettimeofday(&time4, 0);
				FLOAT timek = computeTimeDifference(time3, time4);

				// Decode dev data every decodeEveryK examples.
				if (devData.size() > 0) {
					// If we have dev data, decode the dev data.
					double devAccuracy = doDecoding(model, params, devData, scaleWeightsForDecoding, dropoutRate);
					// If it's the best dev accuracy we've seen so far, decode the test data.
					if (devAccuracy > maxDevAcc) {
						maxDevAcc = devAccuracy;
						testAccForMaxDevAcc = doDecoding(model, params, testData, scaleWeightsForDecoding, dropoutRate);
					}
					//if (testAccuracy > maxTestAcc) maxTestAcc = testAccuracy;
					cerr << "it " << iter << ", " << exampleIndex << "\tdacc\t" << devAccuracy << "\tmaxd\t" << maxDevAcc
						 << "\ttformaxd\t" << testAccForMaxDevAcc; // << "\tmaxt\t" << maxTestAcc;
				} else {
					// Here, we have no dev data, so just decode the test data.
					testAccForMaxDevAcc = doDecoding(model, params, testData, scaleWeightsForDecoding, dropoutRate);
					cerr << "it " << iter << ", " << exampleIndex << "\ttacc\t" << testAccForMaxDevAcc;
				}

				cerr << "\ttimek\t" << timek << endl;
				gettimeofday(&time3, 0);
				gettimeofday(&timeDecodeEnd, 0);
				totalDecodeTime += computeTimeDifference(timeDecodeStart, timeDecodeEnd);
				cerr << "times:\tbuild\t" << totalBuildTime << "\teval\t" << totalEvalTime
					 << "\tdec\t" << totalDecodeTime << endl;
				//printFilterNearestNeighbors(trainData, params);

				//cerr << "times:\treset\t" << totalResetTime << "\tbuild\t" << totalBuildTime << "\teval\t" << totalEvalTime
				//	 << "\tupdate\t" << totalUpdateTime << "\tdec\t" << totalDecodeTime << endl;
			}
		}
		//if (iter % 10 == 0) cerr << "it " << iter << "\tloss\t" << iterLoss <<
		//		"\tparam_norm\t" << computeParameterVectorNorm() <<
		//		"\tgrad_norm\t" << iterGradientNorm << endl;
		cerr << "it " << iter << "\tloss\t" << iterLoss << "\tparam_norm\t" << model.computeParameterVectorNorm() <<
				"\tgrad_norm\t" << iterGradientNorm;
		//double acc = ((double)iterNumCorrect) / ((double)(trainData.size()));
		//cerr << "\ttracc\t" << acc;
		//double trainAccuracy = doDecoding(trainData, m, w, dim, numLabels);
		//cerr << "\ttacc:\t" << trainAccuracy;


		if (devData.size() > 0) {
			// If we have dev data, decode the dev data.
			double devAccuracy = doDecoding(model, params, devData, scaleWeightsForDecoding, dropoutRate);
			// If it's the best dev accuracy we've seen so far, decode the test data.
			if (devAccuracy > maxDevAcc) {
				maxDevAcc = devAccuracy;
				testAccForMaxDevAcc = doDecoding(model, params, testData, scaleWeightsForDecoding, dropoutRate);
			}
			//if (testAccuracy > maxTestAcc) maxTestAcc = testAccuracy;
			cerr << "\tdacc\t" << devAccuracy << "\tmaxd\t" << maxDevAcc
			<< "\ttformaxd\t" << testAccForMaxDevAcc;// << "\tmaxt\t" << maxTestAcc;
		} else {
			// Here, we have no dev data, so just decode the test data.
			testAccForMaxDevAcc = doDecoding(model, params, testData, scaleWeightsForDecoding, dropoutRate);
			cerr << "\ttacc\t" << testAccForMaxDevAcc;// << "\tmaxt\t" << maxTestAcc;
		}
		gettimeofday(&time2, 0);
		FLOAT iterTime = computeTimeDifference(time1, time2);
		cerr << "\t\ttime\t" << iterTime;
		cerr << endl;

		if (iter % 10 == 0) printRecurrentNeuralNetworkNearestNeighbors(trainData, model, params);
	}
	// Print nearest neighbors for dev bigrams.
	//printNearestNeighbors(trainData, devData, params);

	gettimeofday(&endTime, 0);
	FLOAT cumulativeTime = computeTimeDifference(initialTime, endTime);
	cerr << "Total training time:\t" << cumulativeTime << endl;
	cerr << "Final test accuracy:\t" << testAccForMaxDevAcc << endl;
	return testAccForMaxDevAcc;
}


int main(int argc, char** argv) {
    ParamsArray params(argc, argv);
    params.setDescription("Demonstrates a (deep) recurrent neural network using fdag; uses the RNN to "
    		"build a single vector for a sequence, which is then used for classification; permits using "
    		"both forward and backward RNNs simultaneously to get a bigger feature vector for each sequence.");
    params.addConstraint("train_file",
    		"file containing training data, where each line contains the sentence followed by a tab then the "
    		"label (0 or 1)", false); // not optional
    params.addConstraint("dev_file",
    		"file containing dev data to compute accuracy for, same format as train_file", false); // not optional
    params.addConstraint("test_file",
    		"file containing test data to compute accuracy for, same format as train_file", false); // not optional
    params.addConstraint("cv",
    		"if > 0, will do CV on train_file instead of testing on dev/test; this is the number of folds to use with CV (default: 0)", true); // optional

    params.addConstraint("word_vector_file",
    		"file containing word vectors, where each line contains the word followed by a tab then the "
    		"space-separated vector entries", false); // not optional
    params.addConstraint("update_word_vectors",
    		"if true, will learn update given word vectors during learning (default: true)", true); // optional
    params.addConstraint("init_word_vectors_randomly",
    		"if true, will initialize word vectors randomly before learning (default: false)", true); // optional

    params.addConstraint("directions", "directions of RNNs to use; comma-delimited list of directions like \"forward\" and \"backward\" (default: forward)", true); // optional
    params.addConstraint("layer_widths", "widths of hidden layers, comma-delimited. e.g., \"10,10,5\", "
    		"or use \"0\" for no hidden layers (default: 10,5)", true); // optional
    params.addConstraint("algorithm", "optimization algorithm; can be \"sgd_momentum\", \"adagrad\", or \"adadelta\" (case insensitive) (default: adadelta)", true); // optional
    params.addConstraint("stepsize", "stepsize for AdaGrad (default: 0.1)", true); // optional
    params.addConstraint("minibatch_size", "mini-batch size for AdaGrad (default: 10)", true); // optional
    params.addConstraint("num_epochs", "number of epochs of learning (default: 3)", true); // optional
    params.addConstraint("dropout_rate", "probability of setting an output unit to zero; use a value <= 0 for no dropout (default: 0.5)", true); // optional
    params.addConstraint("gradient_norm_threshold", "max gradient norm (\"threshold\" from Pascanu et al paper; use a value <= 0 for no gradient rescaling (default: 1.0)", true); // optional

   	if (!params.runConstraints("run_recurrentnn")) return 0;

	// Read in command-line parameters.
	string trainFilename = params.asString("train_file");
	string devFilename = params.asString("dev_file");
	string testFilename = params.asString("test_file");
	int cvNumFolds = atoi(params.asString("cv", "0").c_str());
	string wordVectorFilename = params.asString("word_vector_file");
	bool updateWordVectors = (params.asString("update_word_vectors", "true") == "true");
	bool initWordVectorsRandomly = (params.asString("init_word_vectors_randomly", "false") == "true");
	string directions = params.asString("directions", "forward");
	string layerWidths = params.asString("layer_widths", "10,5");
	//int outputDimension = atoi(params.asString("output_dimension", "10").c_str());
	string algorithm = params.asString("algorithm", "adadelta");
	double stepsize = atof(params.asString("stepsize", "0.1").c_str());
	int minibatchSize = atoi(params.asString("minibatch_size", "10").c_str());
	int numEpochs = atoi(params.asString("num_epochs", "3").c_str());
	double dropoutRate = atof(params.asString("dropout_rate", "0.5").c_str());
	double gradientNormThreshold = atof(params.asString("gradient_norm_threshold", "1.0").c_str());

	// Change following to true to do finite difference checking of gradients.
	bool doFiniteDifferenceChecking = false;
	srand(1);

	struct timeval initialTime, endTime;
	gettimeofday(&initialTime, 0);

	if (cvNumFolds <= 0) {
		vector<string> trainLines, devLines, testLines;
		readLinesFromFile(trainFilename, trainLines);
		readLinesFromFile(devFilename, devLines);
		readLinesFromFile(testFilename, testLines);

		Model model(2000000, 10000, 10000, 60000);
		double accuracy = trainAndTest(model, trainLines, devLines, testLines, wordVectorFilename, directions, layerWidths,
			algorithm, stepsize, minibatchSize, numEpochs, dropoutRate,
			updateWordVectors, initWordVectorsRandomly, gradientNormThreshold, doFiniteDifferenceChecking);
		cerr << "Final accuracy: " << accuracy << endl;
	} else {
		// Load data.
		vector<string> data;
		readLinesFromFile(trainFilename, data);
		//int inputDimension = loadData(dataFile, data, &numLabels);
		// Do n-fold cross-validation.
		double accuracySum = 0.0;
		unsigned int dataSize = data.size();
		int foldSize = dataSize / cvNumFolds;
		vector<double> accuracies;
		for (int i = 0; i < cvNumFolds; ++i) {
			cerr << "For test fold " << i << ", creating test set with examples " << (i*foldSize) << " to " << (((i+1)*foldSize) - 1) << endl;
			// Generate train/test sets for the current fold (approximately based on discretized index cutoffs).
			vector<string> trainLines, devLines, testLines;
			for (unsigned int j = 0; j < dataSize; ++j) {
				if (((int)j) >= (i*foldSize) && ((int)j) < (i+1)*foldSize) {
					// If the current datum falls within the indices of the current test fold, add it to testData.
					testLines.push_back(data[j]);
					//cerr << "test index: " << j << endl;
				} else {
					// Otherwise, add 90% of it to the training set and 10% to the dev set
					if (j % 10 == 0) {
						devLines.push_back(data[j]);
						//cerr << "dev index: " << j << endl;
					} else {
						trainLines.push_back(data[j]);
						//cerr << "train index: " << j << endl;
					}
				}
			}
			cerr << "Size of training set: " << trainLines.size() << endl;
			cerr << "Size of dev set: " << devLines.size() << endl;
			cerr << "Size of test set: " << testLines.size() << endl;
			// Create a model for the current split.
			//ConvolutionalSentenceModel csm(2000000, 10000, 60000);
			//ConvolutionalSentenceModel csm;
			// Train and test.
			// Create a model for the current fold with given initial Formula capacity.
			Model model(2000000, 10000, 10000, 60000);
			double currAccuracy = trainAndTest(model, trainLines, devLines, testLines, wordVectorFilename, directions, layerWidths,
				algorithm, stepsize, minibatchSize, numEpochs, dropoutRate,
				updateWordVectors, initWordVectorsRandomly, gradientNormThreshold, doFiniteDifferenceChecking);
			//Model model(1000, 1000, 1000);
			// Train and test.
			//double currAccuracy = trainAndTest(model, modelParams, algorithm, trainData, testData, dropoutRate, trainEpochs, doFiniteDifferenceChecking, i+1);
			accuracies.push_back(currAccuracy);
			accuracySum += currAccuracy;

			//double currAccuracy = trainAndTest(model, trainData, devData, testData, wordVectorFilename, filters, algorithm, stepsize, minibatchSize, numEpochs, dropoutRate,
			//	updateWordVectors, initWordVectorsRandomly, doFiniteDifferenceChecking);
		}
		double accuracyMean = accuracySum / ((double) cvNumFolds);
		cerr << "CV mean accuracy: \t" << accuracyMean << endl;
		cerr << "CV accuracy stddev: \t" << standardDeviation(accuracies) << endl;
	}
    //params.addConstraint("layer_widths", "widths of hidden layers, comma-delimited. e.g., \"10,10,5\", or use \"0\" for no hidden layers (default: 10,5)", true); // optional
	//string layerWidths = params.asString("layer_widths", "10,5");

	gettimeofday(&endTime, 0);
	double cumulativeTime = computeTimeDifference(initialTime, endTime);
	cerr << "Total training time:\t" << cumulativeTime << endl;
    return 0;

	//double accuracyMean = accuracySum / 5.0;
	//cerr << "5-fold CV mean accuracy: \t" << accuracyMean << endl;
	//cerr << "5-fold CV accuracy stddev: \t" << standardDeviation(accuracies) << endl;
}
