#include <map>
#include <set>
#include <queue>
#include <algorithm>
#include <stdlib.h>
#include <sys/time.h>
#include "../classifier_utils.h"
#include "segmclassifier.h"

// public:
SegmentClassifierModel::SegmentClassifierModel(int unaryFormulaCapacity, int binaryFormulaCapacity, int fertileFormulaCapacity, int barrenFormulaCapacity) : Model(unaryFormulaCapacity, binaryFormulaCapacity, fertileFormulaCapacity, barrenFormulaCapacity) {
	m_numLabels = 2;
	m_inputDimension = 0;
	m_useStochasticPooling = false;
	m_maxSequenceLength = -1;
	m_useCoarseCost = false;
}

SegmentClassifierModel::~SegmentClassifierModel() {}

double SegmentClassifierModel::trainAndTest(const vector<string>& trainLines, const vector<string>& devLines, const vector<string>& testLines,
		const string& devPredFilenamePrefix, const string& testPredFilenamePrefix, const string& phoneSetFilename,
		const string& forwardLSTMs, const string& backwardLSTMs,
		const string& forwardGRUs, const string& backwardGRUs,
		const string& algorithm, double stepsize, int minibatchSize, int numEpochs, int decodeEveryK,
		double softmaxDropoutRate, double inputDropoutRate, double hiddenDropoutRate,
		bool stochasticPooling, double l2NormConstraint, double gradientNormThreshold, bool useCost, bool useCoarseCost, bool doFiniteDifferenceChecking) {

	struct timeval initialTime, endTime, time1, time2, time3, time4,
		timeBuildStart, timeBuildEnd, timeEvalStart, timeEvalEnd,
		timeDecodeStart, timeDecodeEnd;
	gettimeofday(&initialTime, 0);
	m_useStochasticPooling = stochasticPooling;
	m_useCoarseCost = useCoarseCost;

	resetGradientVector();
	//double stepsize = 0.001;
	map<string, int> labelMap;
	fillPhoneLabelSet(phoneSetFilename, labelMap);
	vector<MFCCSequenceDatum> trainData, devData, testData;
	cerr << "Loading training data." << endl;
	loadMFCCSequenceData(trainLines, trainData, labelMap);
	cerr << "Loading dev data." << endl;
	loadMFCCSequenceData(devLines, devData, labelMap);
	cerr << "Loading test data." << endl;
	loadMFCCSequenceData(testLines, testData, labelMap);

	m_numLabels = labelMap.size();
	fillEvaluationLabelMap(labelMap);
	m_labelNames.clear();
	m_labelNames.reserve(m_numLabels);
	for (unsigned int i = 0; i < m_numLabels; ++i) {
		m_labelNames.push_back("");
	}
	for (map<string, int>::const_iterator iter = labelMap.begin(); iter != labelMap.end(); ++iter) {
		m_labelNames[iter->second] = iter->first;
		cerr << "Storing label " << iter->first << " to labelNames[" << iter->second << "]" << endl;
	}
	for (unsigned int i = 0; i < m_numLabels; ++i) {
		cerr << "Label[" << i << "] = " << m_labelNames[i] << endl;
	}
	m_inputDimension = 39;

	// Print label set.
	cerr << "Printing label set:" << endl;
	for (map<string, int>::const_iterator iter = labelMap.begin(); iter != labelMap.end(); ++iter) {
		cerr << iter->first << endl;
	}
	cerr << "Done printing label set:" << endl;

	// Initialize parameters for each LSTM/GRU.
	initializeNetworkFromConfiguration(forwardLSTMs, true, true);
	initializeNetworkFromConfiguration(backwardLSTMs, true, false);
	initializeNetworkFromConfiguration(forwardGRUs, false, true);
	initializeNetworkFromConfiguration(backwardGRUs, false, false);

	//initializeAllParameters(trainData, devData, testData, initWordVectorsRandomly, filters);

	double minDevError = 999999.0;
	double testErrorForMinDevError = 999999.0;
	//double maxTestAcc = -1.0;

	FLOAT totalBuildTime = 0.0f;
	FLOAT totalEvalTime = 0.0f;
	FLOAT totalDecodeTime = 0.0f;

	int decodeEveryKToUse = decodeEveryK;
	if (trainData.size() < 11000) {
		decodeEveryKToUse = trainData.size() / 5;
	}

	Algorithm alg = getAlgorithm(algorithm);
	// Initialize vector of previous updates for SGD w/ momentum.
	vector<double> sgdMomentumUpdates;//(getNumParams(), 0.0);
	double momentumCoefficient = 0.95;
	// Initialize sum of squared gradients for AdaGrad and AdaDelta.
	vector<double> sumOfSquaredGradients(getNumParams(), 0.0);
	// Initialize sum of squared updates for AdaDelta.
	vector<double> sumOfSquaredUpdates(getNumParams(), 0.0);
	double epsilon = 0.000001;
	double rho = 0.95;
	if (alg != ADAGRAD) {
		cerr << "Error: only AdaGrad is currently supported!" << endl;
	}
	// Initialize sum of squared gradients for AdaGrad.
	//vector<double> sumOfSquaredGradients(getNumParams(), 0.0);
	//int maxNumFormulasUsed = getNumberOfFormulasUsedInPool();
	/*int maxUnaryFormulasUsed = 0;
	int maxBinaryFormulasUsed = 0;
	int maxFertileFormulasUsed = 0;
	int maxBarrenFormulasUsed = 0;
	int unaryFormulasUsed = 0;
	int binaryFormulasUsed = 0;
	int fertileFormulasUsed = 0;
	int barrenFormulasUsed = 0;
	getNumberOfFormulasUsedInPool(maxUnaryFormulasUsed, maxBinaryFormulasUsed, maxFertileFormulasUsed, maxBarrenFormulasUsed);
	*/
	// Train for numEpochs iterations of training.
	for (unsigned int iter = 1; iter <= numEpochs; ++iter) {
		double iterLoss = 0.0;
		double iterGradientNorm = 0.0;

		//int iterNumCorrect = 0;
		gettimeofday(&time1, 0);
		gettimeofday(&time3, 0);

		for (unsigned int exampleIndex = 0; exampleIndex < trainData.size(); ++exampleIndex) {
			MFCCSequenceDatum& datum = trainData[exampleIndex];
			// Sample a data point.
			//int exampleIndex = i;//rand() % (trainData.size());
			// We are processing a new example, so reset Formula data structures in the model.
			resetForNewExample(datum);

			//cerr << "Computing gradient for sentence \"" << sentenceTokens[0] << " " << sentenceTokens[1]
			//	 << "\" with true label " << trueLabel << endl;

			// Create the formula containing the loss for the current example.
			gettimeofday(&timeBuildStart, 0);
			Formula* datumLossFDAG = createLossFDAGForMFCCSequenceDatum(datum, softmaxDropoutRate, inputDropoutRate, hiddenDropoutRate, useCost);
			gettimeofday(&timeBuildEnd, 0);
			totalBuildTime += computeTimeDifference(timeBuildStart, timeBuildEnd);
			/*getNumberOfFormulasUsedInPool(unaryFormulasUsed, binaryFormulasUsed, fertileFormulasUsed, barrenFormulasUsed);
			if (unaryFormulasUsed > maxUnaryFormulasUsed) {
				maxUnaryFormulasUsed = unaryFormulasUsed;
				cerr << "Info: New unary formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << unaryFormulasUsed << endl;
			}
			if (binaryFormulasUsed > maxBinaryFormulasUsed) {
				maxBinaryFormulasUsed = binaryFormulasUsed;
				cerr << "Info: New binary formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << binaryFormulasUsed << endl;
			}
			if (fertileFormulasUsed > maxFertileFormulasUsed) {
				maxFertileFormulasUsed = fertileFormulasUsed;
				cerr << "Info: New fertile formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << fertileFormulasUsed << endl;
			}
			if (barrenFormulasUsed > maxBarrenFormulasUsed) {
				maxBarrenFormulasUsed = barrenFormulasUsed;
				cerr << "Info: New barren formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << barrenFormulasUsed << endl;
			}*/
			//int currNumFormulasUsed = getNumberOfFormulasUsedInPool();
			//if (currNumFormulasUsed > maxNumFormulasUsed) {
			//	cerr << "Info: New formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << currNumFormulasUsed << endl;
			//	maxNumFormulasUsed = currNumFormulasUsed;
			//}
			// Compute gradient and loss for current example.
			gettimeofday(&timeEvalStart, 0);
			double loss = evalAndComputeDerivatives(datumLossFDAG, exampleIndex, doFiniteDifferenceChecking);
			if (loss != loss) {
				//cerr << "Error: nan loss, example " << exampleIndex << " with label " << trainData[exampleIndex].m_label << endl;
				cerr << "Error: nan loss, at segment " << exampleIndex << " with true label " << trainData[exampleIndex].m_labelString << endl;
			} else {
				iterLoss += loss;
			}
			gettimeofday(&timeEvalEnd, 0);
			totalEvalTime += computeTimeDifference(timeEvalStart, timeEvalEnd);
			datum.doneProcessing();

			// If we're at the end of a mini-batch or at the end of the training data, make an update.
			bool doUpdate = false;
			if ((exampleIndex + 1) % minibatchSize == 0 || trainData.size() <= 1 || exampleIndex == trainData.size() - 1) doUpdate = true;
			if (doUpdate) {
				if (gradientNormThreshold > 0.0) {
					double gradientNorm = computeGradientVectorNorm();
					if (gradientNorm >= gradientNormThreshold) {
						//cerr << "Info: gradientNorm = " << gradientNorm << "; Scaling gradient vector by " << (gradientNormThreshold / gradientNorm) << endl;
						scaleGradientVectorByConstant(gradientNormThreshold / gradientNorm);
						//cerr << "Info: new gradientNorm = " << model.computeGradientVectorNorm() << endl;
					}
				}
				double gradientSquaredNorm = 0.0;
				if (alg == SGD_MOMENTUM) {
					gradientSquaredNorm = doSGDMomentumUpdate(sgdMomentumUpdates, stepsize, momentumCoefficient);
				} else if (alg == ADAGRAD) {
					gradientSquaredNorm = doAdaGradUpdate(sumOfSquaredGradients, stepsize);
				} else if (alg == ADADELTA){
					gradientSquaredNorm = doAdaDeltaUpdate(sumOfSquaredGradients, sumOfSquaredUpdates, epsilon, rho);
				}
				//double l2norm = sqrt(gradientSquaredNorm);
				//iterGradientNorm += l2norm;
				iterGradientNorm += computeGradientVectorNorm();
				//if (l2NormConstraint > 0.0) {// && l2norm > l2NormConstraint) {
				//	scaleWeightsToMatchL2Norm(params, l2NormConstraint);
				//}
				resetGradientVector();

				//double gradientSquaredNorm = doAdaGradUpdate(sumOfSquaredGradients, stepsize);
				//double l2norm = sqrt(gradientSquaredNorm);
				//iterGradientNorm += l2norm;
				//if (l2NormConstraint > 0.0) {// && l2norm > l2NormConstraint) {
				//	scaleWeightsToMatchL2Norm(params, l2NormConstraint);
				//}
				//resetGradientVector();
			}

			if ((exampleIndex + 1) % decodeEveryKToUse == 0) {

				gettimeofday(&timeDecodeStart, 0);
				gettimeofday(&time4, 0);
				FLOAT timek = computeTimeDifference(time3, time4);

				// Decode dev data every decodeEveryK examples.
				if (devData.size() > 0) {
					// If we have dev data, decode the dev data.
					string filenameSuffix = "." + itos(iter) + "." + itos(exampleIndex + 1) + ".txt";
					double devError = doDecoding(devData, devPredFilenamePrefix, filenameSuffix);
					//double devError = doDecoding(devData, devPredFilenamePrefix);
					// If it's the best dev error we've seen so far, decode the test data.
					if (devError < minDevError) {
						minDevError = devError;
						testErrorForMinDevError = doDecoding(testData, testPredFilenamePrefix, filenameSuffix);
					}
					//if (testAccuracy > maxTestAcc) maxTestAcc = testAccuracy;
					cerr << "it " << iter << ", " << (exampleIndex + 1) << "\tderr\t" << devError << "\tmind\t" << minDevError
						 << "\ttformind\t" << testErrorForMinDevError; // << "\tmaxt\t" << maxTestAcc;
				} else {
					// Here, we have no dev data, so just decode the test data.
					string filenameSuffix = "." + itos(iter) + "." + itos(exampleIndex + 1) + ".txt";
					testErrorForMinDevError = doDecoding(testData, testPredFilenamePrefix, filenameSuffix);
					cerr << "it " << iter << ", " << (exampleIndex + 1) << "\tterr\t" << testErrorForMinDevError;
				}

				cerr << "\ttimek\t" << timek << endl;
				gettimeofday(&time3, 0);
				gettimeofday(&timeDecodeEnd, 0);
				totalDecodeTime += computeTimeDifference(timeDecodeStart, timeDecodeEnd);
				cerr << "times:\tbuild\t" << totalBuildTime << "\teval\t" << totalEvalTime
					 << "\tdec\t" << totalDecodeTime << endl;
				//printFilterNearestNeighbors(trainData, params);

				//cerr << "times:\treset\t" << totalResetTime << "\tbuild\t" << totalBuildTime << "\teval\t" << totalEvalTime
				//	 << "\tupdate\t" << totalUpdateTime << "\tdec\t" << totalDecodeTime << endl;
			} else {
				if ((exampleIndex + 1) % 5000 == 0) {
					cerr << (exampleIndex + 1) << endl;
				}
			}
		}
		//if (iter % 10 == 0) cerr << "it " << iter << "\tloss\t" << iterLoss <<
		//		"\tparam_norm\t" << computeParameterVectorNorm() <<
		//		"\tgrad_norm\t" << iterGradientNorm << endl;
		cerr << "it " << iter << "\tloss\t" << iterLoss << "\tparam_norm\t" << computeParameterVectorNorm() <<
				"\tgrad_norm\t" << iterGradientNorm;

		if (devData.size() > 0) {
			// If we have dev data, decode the dev data.
			string filenameSuffix = "." + itos(iter) + ".txt";
			double devError = doDecoding(devData, devPredFilenamePrefix, filenameSuffix);
			// If it's the best dev error we've seen so far, decode the test data.
			if (devError < minDevError) {
				minDevError = devError;
				testErrorForMinDevError = doDecoding(testData, testPredFilenamePrefix, filenameSuffix);
			}
			//if (testAccuracy > maxTestAcc) maxTestAcc = testAccuracy;
			cerr << "\tderr\t" << devError << "\tmind\t" << minDevError
				 << "\ttformind\t" << testErrorForMinDevError; // << "\tmaxt\t" << maxTestAcc;
		} else {
			// Here, we have no dev data, so just decode the test data.
			string filenameSuffix = "." + itos(iter) + ".txt";
			testErrorForMinDevError = doDecoding(testData, testPredFilenamePrefix, filenameSuffix);
			cerr << "\tterr\t" << testErrorForMinDevError;
		}
		gettimeofday(&time2, 0);
		FLOAT iterTime = computeTimeDifference(time1, time2);
		cerr << "\t\ttime\t" << iterTime;
		cerr << endl;

		//if (iter % 5 == 0) printFilterNearestNeighbors(trainData);
	}
	// Print nearest neighbors for dev bigrams.
	//printNearestNeighbors(trainData, devData, params);

	gettimeofday(&endTime, 0);
	FLOAT cumulativeTime = computeTimeDifference(initialTime, endTime);
	cerr << "Total training time:\t" << cumulativeTime << endl;
	//cerr << "Final test accuracy:\t" << testAccForMaxDevAcc << endl;
	return minDevError;//testAccForMaxDevAcc;
}

void SegmentClassifierModel::initializeNetworkFromConfiguration(const string& cfg, bool lstm, bool forward) {
	if (cfg != "0") {
		bool initSoftmaxTrue = true;
		vector<string> networks;
		tokenize(cfg, '.', networks);
		for (unsigned int i = 0; i < networks.size(); ++i) {
			// Within each network, tokenize the number of networks from the sequence of hidden layer widths.
			vector<string> toks;
			tokenize(networks[i], ':', toks);
			if (toks.size() != 2) {
				cerr << "Error: no network parameters given in token \"" << networks[i] << "\"." << endl;
			} else {
				int numNetworks = atoi(toks[0].c_str());
				if (lstm) {
					for (int j = 0; j < numNetworks; ++j) {
						cerr << "Initializing deep LSTM " << (j+1) << " with layer widths " << toks[1] << endl;
						DeepLSTM lstmParams;
						lstmParams.initializeParameters(*this, m_inputDimension, m_numLabels, forward, toks[1], 1, initSoftmaxTrue, true, true, true, false);
						//initializeDeepLSTMParameters(model, params, lstmParams, forward, toks[1]);
						m_lstmParams.push_back(lstmParams);
					}
				} else {
					for (int j = 0; j < numNetworks; ++j) {
						cerr << "Initializing deep GRU " << (j+1) << " with layer widths " << toks[1] << endl;
						DeepGRU gruParams;
						gruParams.initializeParameters(*this, m_inputDimension, m_numLabels, forward, toks[1], 1, initSoftmaxTrue);
						m_gruParams.push_back(gruParams);
					}
				}
			}
		}
	}
}


void SegmentClassifierModel::fillPhoneLabelSet(const string& phoneSetFilename, map<string, int>& labelMap) {
	vector<string> phoneSetLines;
	readLinesFromFile(phoneSetFilename, phoneSetLines);
	for (unsigned int i = 0; i < phoneSetLines.size(); ++i) {
		labelMap.insert(make_pair(phoneSetLines[i], i));
	}
}

void SegmentClassifierModel::addLabelMappingToEvaluationLabelMap(const map<string, int>& labelMap, const string& s1, const string& s2) {
	//map<int, int> m_evaluationLabelMap;
	// Look up the two labels in the label map.
	map<string, int>::const_iterator iter1 = labelMap.find(s1);
	map<string, int>::const_iterator iter2 = labelMap.find(s2);
	// If the labels were found, add an integer mapping.
	if (iter1 != labelMap.end() && iter2 != labelMap.end()) {
		cerr << "For evaluation, mapping label " << s1 << " (" << iter1->second << ") to coarse label " << s2 << " (" << iter2->second << ")" << endl;
		m_evaluationLabelMap.insert(make_pair(iter1->second, iter2->second));
	} else {
		cerr << "Error: labels missing from label map: " << s1 << " or " << s2 << endl;
	}
}

void SegmentClassifierModel::fillEvaluationLabelMap(const map<string, int>& labelMap) {
	string silenceString = "pau";
	addLabelMappingToEvaluationLabelMap(labelMap, "ao",   "aa");
	addLabelMappingToEvaluationLabelMap(labelMap, "ax",   "ah");
	addLabelMappingToEvaluationLabelMap(labelMap, "axh",  "ah");
	addLabelMappingToEvaluationLabelMap(labelMap, "axr",  "er");
	addLabelMappingToEvaluationLabelMap(labelMap, "hv",   "hh");
	addLabelMappingToEvaluationLabelMap(labelMap, "ix",   "ih");
	addLabelMappingToEvaluationLabelMap(labelMap, "el",   "l");
	addLabelMappingToEvaluationLabelMap(labelMap, "em",   "m");
	addLabelMappingToEvaluationLabelMap(labelMap, "en",   "n");
	addLabelMappingToEvaluationLabelMap(labelMap, "nx",   "n");
	addLabelMappingToEvaluationLabelMap(labelMap, "eng",  "ng");
	addLabelMappingToEvaluationLabelMap(labelMap, "zh",   "sh");
	addLabelMappingToEvaluationLabelMap(labelMap, "ux",   "uw");
	addLabelMappingToEvaluationLabelMap(labelMap, "pcl",  silenceString);
	addLabelMappingToEvaluationLabelMap(labelMap, "tcl",  silenceString);
	addLabelMappingToEvaluationLabelMap(labelMap, "kcl",  silenceString);
	addLabelMappingToEvaluationLabelMap(labelMap, "bcl",  silenceString);
	addLabelMappingToEvaluationLabelMap(labelMap, "dcl",  silenceString);
	addLabelMappingToEvaluationLabelMap(labelMap, "gcl",  silenceString);
	//addLabelMappingToEvaluationLabelMap(labelMap, "h#",   silenceString);
	addLabelMappingToEvaluationLabelMap(labelMap, "pau",  silenceString);
	addLabelMappingToEvaluationLabelMap(labelMap, "epi",  silenceString);
	addLabelMappingToEvaluationLabelMap(labelMap, "<s>",  silenceString);
	addLabelMappingToEvaluationLabelMap(labelMap, "</s>", silenceString);
}

// Resets data structures for processing the next example (whether in training or decoding).
void SegmentClassifierModel::resetForNewExample(MFCCSequenceDatum& datum) {
	// We're done with all formula nodes we've previously created, so reset them and the cache.
	bool resetCachedParameterFormulasTrue = true;
	resetFormulaObjects(resetCachedParameterFormulasTrue);
	// The number of cached nodes is the length of the sequence times the hidden layer width.
	unsigned int numCachedNodes = datum.m_sequenceLength;
	for (unsigned int i = 0; i < m_lstmParams.size(); ++i) {
		m_lstmParams[i].resetForNewExample(0, numCachedNodes);
	}
	for (unsigned int i = 0; i < m_gruParams.size(); ++i) {
		m_gruParams[i].resetForNewExample(0, numCachedNodes);
	}
	precomputeMFCCFormulas(this, datum, m_inputDimension);
}

Formula* SegmentClassifierModel::createLossFDAGForMFCCSequenceDatum(MFCCSequenceDatum& datum, double softmaxDropoutRate, double inputDropoutRate, double hiddenDropoutRate, bool useCost) {
	// Create a term for the correct label's score, which is negated in the loss.
	Formula* negCorrect = getFormulaObject(NEG, 1);
	// Create a sum node to sum over labels.
	Formula* sum = getFormulaObject(PLUS, m_numLabels);
	// Sum over labels.
	for (unsigned int label = 0; label < m_numLabels; ++label) {
		Formula* score = createSoftmaxScoreForLabel(label, datum, softmaxDropoutRate, inputDropoutRate, hiddenDropoutRate, useCost);
		// If the current label is the true label, add the score to the negated correct term.
		if (label == datum.m_label) {
			negCorrect->addChild(score);
		} else if (useCost) {
			// Here, the label is not correct, so we want to add a cost constant to the score.
			double costConstant = 1.0;
			if (m_useCoarseCost) {
				int coarseLabel = label;
				map<int, int>::const_iterator iter1 = m_evaluationLabelMap.find(label);
				if (iter1 != m_evaluationLabelMap.end()) coarseLabel = iter1->second;
				int coarseGoldLabel = datum.m_label;
				map<int, int>::const_iterator iter2 = m_evaluationLabelMap.find(datum.m_label);
				if (iter2 != m_evaluationLabelMap.end()) coarseGoldLabel = iter2->second;
				if (coarseLabel == coarseGoldLabel) costConstant = 0.0;
			}
			score->addChild(getConstantFormulaObject(costConstant));
			// Check if the current label and the true label map to the same coarse label.
		}
		// Always add the exponentiated label score to the sum over labels.
		Formula* expScore = getFormulaObject(EXP, 1);
		expScore->addChild(score);
		sum->addChild(expScore);
	}
	// Put a log in front of the sum of exponentiated label scores.
	Formula* logSum = getFormulaObject(LOG, 1);
	logSum->addChild(sum);
	// Add the negated correct score to the log-sum over all exponentiated label scores.
	Formula* negCorrectPlusSum = getFormulaObject(PLUS, 2);
	negCorrectPlusSum->addChild(negCorrect);
	negCorrectPlusSum->addChild(logSum);
	// Return the sum.
	return negCorrectPlusSum;
}

Formula* SegmentClassifierModel::createSoftmaxScoreForLabel(unsigned int label, MFCCSequenceDatum& datum,
		double softmaxDropoutRate, double inputDropoutRate, double hiddenDropoutRate, bool usingCost) {
	// Sum over all dimensions of the final hidden vectors for all LSTMs, as well as the bias weight.
	// We only use the bias weight from the first LSTM.
	int numTermsInSum = 1;
	if (usingCost) ++numTermsInSum;
	for (unsigned int lstmIndex = 0; lstmIndex < m_lstmParams.size(); ++lstmIndex) {
		numTermsInSum += m_lstmParams[lstmIndex].getTotalHiddenVectorSize();
	}
	for (unsigned int gruIndex = 0; gruIndex < m_gruParams.size(); ++gruIndex) {
		numTermsInSum += m_gruParams[gruIndex].getTotalHiddenVectorSize();
	}

	Formula* sumOverDimensions = getFormulaObject(PLUS, numTermsInSum);

	for (unsigned int lstmIndex = 0; lstmIndex < m_lstmParams.size(); ++lstmIndex) {
		DeepLSTM& lstmParams = m_lstmParams[lstmIndex];
		// Go through all hidden layers for the current deep LSTM.
		for (unsigned int j = 0; j < lstmParams.m_layerWidths.size(); ++j) {
			// Go through all entries in the current hidden layer for the current deep LSTM.
			for (unsigned int k = 0; k < lstmParams.m_layerWidths[j]; ++k) {
				bool dropout = false;
				if (softmaxDropoutRate > 0.0 && getRandomDouble() < softmaxDropoutRate) dropout = true;
				if (!dropout) {
					// Create the Formula node for the kth dimension of the final hidden vector at layer j (i.e., the time t is the number of words in the sentence).
					int startTime = 0;
					if (lstmParams.m_forward) startTime = datum.m_sequenceLength - 1;
					int startDepth = j;
					Formula* zj = lstmParams.createHiddenNodeAtTimeAtDepth(*this, datum, 0, k, startTime, startDepth, inputDropoutRate, hiddenDropoutRate);
					// Multiply it by the weight corresponding to the provided label and the current dimension.
					Formula* prod = getFormulaObject(TIMES, 2);
					prod->addChild(zj);
					prod->addChild(getCachedParameterFormula(lstmParams.m_softmaxWeights.at(j).at(k).at(label)));
					sumOverDimensions->addChild(prod);
				}
			}
		}
	}
	for (unsigned int gruIndex = 0; gruIndex < m_gruParams.size(); ++gruIndex) {
		DeepGRU& gruParams = m_gruParams[gruIndex];
		// Go through all hidden layers for the current deep GRU.
		for (unsigned int j = 0; j < gruParams.m_layerWidths.size(); ++j) {
			// Go through all entries in the current hidden layer for the current deep GRU.
			for (unsigned int k = 0; k < gruParams.m_layerWidths[j]; ++k) {
				bool dropout = false;
				if (softmaxDropoutRate > 0.0 && getRandomDouble() < softmaxDropoutRate) dropout = true;
				if (!dropout) {
					// Create the Formula node for the jth dimension of the final hidden vector (i.e., the time t is the number of words in the sentence).
					int startTime = 0;
					if (gruParams.m_forward) startTime = datum.m_sequenceLength - 1;
					int startDepth = j;
					Formula* zj = gruParams.createHiddenNodeAtTimeAtDepth(*this, datum, 0, k, startTime, startDepth, inputDropoutRate, hiddenDropoutRate);
					// Multiply it by the weight corresponding to the provided label and the current dimension.
					Formula* prod = getFormulaObject(TIMES, 2);
					prod->addChild(zj);
					prod->addChild(getCachedParameterFormula(gruParams.m_softmaxWeights.at(j).at(k).at(label)));
					sumOverDimensions->addChild(prod);
				}
			}
		}
	}
	// Add the softmax bias.
	PARAM_ID singleBias;
	if (m_lstmParams.size() > 0) {
		singleBias = m_lstmParams[0].m_softmaxBias.at(0).at(label);
	} else if (m_gruParams.size() > 0) {
		singleBias = m_gruParams[0].m_softmaxBias.at(0).at(label);
	} else {
		cerr << "Error: no LSTM or GRU parameters initialized!" << endl;
		singleBias = 0;
	}
	sumOverDimensions->addChild(getCachedParameterFormula(singleBias));
	return sumOverDimensions;
}

double SegmentClassifierModel::doDecoding(vector<MFCCSequenceDatum>& data, const string& predFilenamePrefix, const string& filenameSuffix) {
	int numErrors = 0;
	int total = 0;
	bool writeToFile = false;
	ofstream outputFile;
	if (predFilenamePrefix != "" && predFilenamePrefix != "NONE") {
		string outputFilename = predFilenamePrefix + filenameSuffix;
		outputFile.open(outputFilename.c_str(), ofstream::out);
		writeToFile = true;
	}
	for (unsigned int i = 0; i < data.size(); ++i) {
		MFCCSequenceDatum& datum = data[i];
		// We are processing a new example, so reset Formula data structures in the model.
		resetForNewExample(datum);
		int predictedLabel = -1;
		string posteriorsToWrite;
		bool correct = decode(datum, i, predictedLabel, writeToFile, posteriorsToWrite);
		if (writeToFile) {
			//string predictedLabelString = m_labelNames[predictedLabel];
			outputFile << posteriorsToWrite << endl;
		}
		datum.doneProcessing();
		if (!correct) ++numErrors;
		++total;
	}
	if (writeToFile) outputFile.close();
	return (((double)numErrors) / ((double)total));
}

bool SegmentClassifierModel::decode(MFCCSequenceDatum& datum, int exampleNumber, int& predictedLabel, bool writeToFile, string& posteriorsToWrite) {
	double maxScore = neg_inf;
	int maxLabel = -1;
	vector<double> labelScores;
	double sumExpScores = 0.0;
	posteriorsToWrite = "";
	for (unsigned int i = 0; i < m_numLabels; ++i) {
		Formula* score = createSoftmaxScoreForLabel(i, datum, -1.0, -1.0, -1.0, false);
		double currScore = score->evaluate();
		if (currScore > maxScore) {
			maxScore = currScore;
			maxLabel = i;
		}
		if (writeToFile) {
			double expCurrScore = exp(currScore);
			sumExpScores += expCurrScore;
			labelScores.push_back(expCurrScore);
		}
	}

	if (maxLabel == -1) {
		cerr << "Error: no label found with score > negative infinity! on line #" << exampleNumber << endl;
	} else {
		// Map the gold and predicted labels to their coarse labels.
		predictedLabel = maxLabel;

		if (writeToFile) {
			posteriorsToWrite += m_labelNames[predictedLabel];
			for (unsigned int i = 0; i < labelScores.size(); ++i) {
				posteriorsToWrite += " ";
				double currentPosterior = labelScores[i] / sumExpScores;
				posteriorsToWrite += ftos(currentPosterior);
			}
		}

		int coarsePredictedLabel = maxLabel;
		map<int, int>::const_iterator iter1 = m_evaluationLabelMap.find(maxLabel);
		if (iter1 != m_evaluationLabelMap.end()) coarsePredictedLabel = iter1->second;
		int coarseGoldLabel = datum.m_label;
		map<int, int>::const_iterator iter2 = m_evaluationLabelMap.find(datum.m_label);
		if (iter2 != m_evaluationLabelMap.end()) coarseGoldLabel = iter2->second;
		if (coarsePredictedLabel == coarseGoldLabel) return true;
		//if (maxLabel == datum.m_label) return true;

	}
	return false;
}

