#include <iostream>
#include <stdlib.h>
#include <algorithm>
#include <sys/time.h>
#include "../../utils/cmdparams.h"
#include "../../fdag/model.h"
#include "../classifier_utils.h"
#include "../rnn/lstm.h"
#include "../rnn/gru.h"
#include "../rnn/sru.h"
#include "../cnn/cnn.h"

struct ModelParams {
	// The model can contain multiple deep LSTM RNNs, multiple deep GRU (gated recurrent unit) RNNs,
	// multiple deep SRU (shift-reduce unit) RNNs, and a CNN.
	vector<DeepLSTM> m_lstmParams;
	vector<DeepGRU> m_gruParams;
	vector<DeepSRU> m_sruParams;
	ConvolutionalNeuralNetwork m_cnnParams;
	bool m_usingCNN;

	// Parameters and flags for word vectors:
	bool m_updateWordVectors;				// If true, will include word vector entries as parameters to be learned.
	map<string, vector<double> > m_wordVectors;	// Word vector entries from word vector file.
	map<string, vector<PARAM_ID> > v;		// Word vector parameters, initialized to word vector values from file.
	int m_wordVectorDimensionality;			// Dimensionality of word vectors.
	vector<double>* m_unknownWordVector;	// Pointer to unknown word vector.

	vector<vector<PARAM_ID> > m_wordVectorMapping;

	// The input dimension is the number of features.
	unsigned int m_inputDimension;
	int m_mappedInputDimension;
	// The number of labels.
	int m_numLabels;
};

void initializeSharedParameters(Model& model, ModelParams& params, int inputDimension,
		const vector<WordSequenceDatum>& trainData, const vector<WordSequenceDatum>& devData, const vector<WordSequenceDatum>& testData,
		bool updateWordVectors, bool initWordVectorsRandomly, int mapWordVectorsToDim, double scalingFactor, bool demoMode) {
	params.m_inputDimension = inputDimension;
	params.m_numLabels = 2;
	params.m_updateWordVectors = updateWordVectors;
	params.m_mappedInputDimension = mapWordVectorsToDim;

	if (params.m_updateWordVectors) {

		// Only create parameters for words in training/dev/test data. Words that only appear in dev/test data
		// will not have their parameters updated during training; it's just useful to have parameters for them
		// so that we can do the same procedure during decoding as we do during training.
		set<string> paramWords;
		insertWordSequenceDataWordsIntoSet(trainData, paramWords);
		insertWordSequenceDataWordsIntoSet(devData, paramWords);
		insertWordSequenceDataWordsIntoSet(testData, paramWords);

		//if (demoMode) {
		//	// If we'll be doing demo mode, create parameters for all words.
		//	for (map<string, vector<double> >::const_iterator iter = params.m_wordVectors.begin(); iter != params.m_wordVectors.end(); ++iter) {
		//		const string& currWord = iter->first;
		//		paramWords.insert(currWord);
		//	}
		//}

		for (set<string>::const_iterator iter = paramWords.begin(); iter != paramWords.end(); ++iter) {
			const string& currWord = *iter;
			bool unk = false;
			const vector<double>* currVector = getWordVector(params.m_wordVectors, currWord, unk);
			vector<PARAM_ID> currVectorParams;
			// Go through all dimensions of the current word vector.
			for (unsigned int i = 0; i < currVector->size(); ++i) {
				// Create a new parameter for the current word vector entry.
				if (initWordVectorsRandomly) {
					currVectorParams.push_back(model.initializeNewParameter(getRandomDouble() - 0.5));
				} else {
					currVectorParams.push_back(model.initializeNewParameter(currVector->at(i)));
				}
			}
			params.v.insert(make_pair(currWord, currVectorParams));
		}
	} else {
		if (mapWordVectorsToDim > 0) {
			cerr << "Mapping word vectors to dimensionality " << params.m_mappedInputDimension << ". Creating " << params.m_inputDimension << "-by-"
				 << params.m_mappedInputDimension << " matrix." << endl;
			model.initializeParameterMatrix(params.m_wordVectorMapping, params.m_inputDimension, params.m_mappedInputDimension, scalingFactor);

		}
	}
	cerr << "Initialized model. # parameters = " << model.getNumParams() << endl;
}

// Resets data structures for processing the next example (whether in training or decoding).
void resetForNewExample(Model& model, ModelParams& params, WordSequenceDatum& datum) {
	// We're done with all formula nodes we've previously created, so reset them and the cache.
	bool resetCachedParameterFormulasTrue = true;
	model.resetFormulaObjects(resetCachedParameterFormulasTrue);
	// The number of cached nodes is the length of the sequence times the hidden layer width.
	unsigned int numCachedNodes = datum.m_sequenceLength;
	for (unsigned int i = 0; i < params.m_lstmParams.size(); ++i) {
		params.m_lstmParams[i].resetForNewExample(0, numCachedNodes);
	}
	for (unsigned int i = 0; i < params.m_gruParams.size(); ++i) {
		params.m_gruParams[i].resetForNewExample(0, numCachedNodes);
	}
	for (unsigned int i = 0; i < params.m_sruParams.size(); ++i) {
		params.m_sruParams[i].resetForNewExample(0, numCachedNodes);
	}
	if (params.m_usingCNN) {
		params.m_cnnParams.resetForNewExample();
	}
	if (params.m_mappedInputDimension > 0) {
		precomputeWordVectorFormulasWithMapping(&model, datum, params.m_inputDimension, params.m_updateWordVectors, params.m_mappedInputDimension, params.m_wordVectorMapping);
	} else {
		precomputeWordVectorFormulas(&model, datum, params.m_inputDimension, params.m_updateWordVectors);
	}
}

Formula* createSoftmaxScoreForLabel(Model& model, ModelParams& params, unsigned int label, WordSequenceDatum& datum,
		double softmaxDropoutRate, double inputDropoutRate, double hiddenDropoutRate) {
	// Sum over all dimensions of the final hidden vectors for all RNNs and the CNN, as well as the bias weight.
	// We only use the bias weight from the first LSTM.
	int numTermsInSum = 1;
	for (unsigned int lstmIndex = 0; lstmIndex < params.m_lstmParams.size(); ++lstmIndex) {
		numTermsInSum += params.m_lstmParams[lstmIndex].getTotalHiddenVectorSize();
	}
	for (unsigned int gruIndex = 0; gruIndex < params.m_gruParams.size(); ++gruIndex) {
		numTermsInSum += params.m_gruParams[gruIndex].getTotalHiddenVectorSize();
	}
	for (unsigned int sruIndex = 0; sruIndex < params.m_sruParams.size(); ++sruIndex) {
		numTermsInSum += params.m_sruParams[sruIndex].getTotalHiddenVectorSize();
	}
	if (params.m_usingCNN) numTermsInSum += params.m_cnnParams.getNumberOfFilters();

	Formula* sumOverDimensions = model.getFormulaObject(PLUS, numTermsInSum);

	for (unsigned int lstmIndex = 0; lstmIndex < params.m_lstmParams.size(); ++lstmIndex) {
		DeepLSTM& lstmParams = params.m_lstmParams[lstmIndex];
		// Go through all hidden layers for the current deep LSTM.
		for (unsigned int j = 0; j < lstmParams.m_layerWidths.size(); ++j) {
			// Go through all entries in the current hidden layer for the current deep LSTM.
			for (unsigned int k = 0; k < lstmParams.m_layerWidths[j]; ++k) {
				bool dropout = false;
				if (softmaxDropoutRate > 0.0 && getRandomDouble() < softmaxDropoutRate) dropout = true;
				if (!dropout) {
					// Create the Formula node for the kth dimension of the final hidden vector at layer j (i.e., the time t is the number of words in the sentence).
					int startTime = 0;
					if (lstmParams.m_forward) startTime = datum.m_sequenceLength - 1;
					int startDepth = j;
					Formula* zj = lstmParams.createHiddenNodeAtTimeAtDepth(model, datum, 0, k, startTime, startDepth, inputDropoutRate, hiddenDropoutRate);
					// Multiply it by the weight corresponding to the provided label and the current dimension.
					Formula* prod = model.getFormulaObject(TIMES, 2);
					prod->addChild(zj);
					prod->addChild(model.getCachedParameterFormula(lstmParams.m_softmaxWeights.at(j).at(k).at(label)));
					sumOverDimensions->addChild(prod);
				}
			}
		}
	}
	for (unsigned int gruIndex = 0; gruIndex < params.m_gruParams.size(); ++gruIndex) {
		DeepGRU& gruParams = params.m_gruParams[gruIndex];
		// Go through all hidden layers for the current deep GRU.
		for (unsigned int j = 0; j < gruParams.m_layerWidths.size(); ++j) {
			// Go through all entries in the current hidden layer for the current deep GRU.
			for (unsigned int k = 0; k < gruParams.m_layerWidths[j]; ++k) {
				bool dropout = false;
				if (softmaxDropoutRate > 0.0 && getRandomDouble() < softmaxDropoutRate) dropout = true;
				if (!dropout) {
					// Create the Formula node for the jth dimension of the final hidden vector (i.e., the time t is the number of words in the sentence).
					int startTime = 0;
					if (gruParams.m_forward) startTime = datum.m_sequenceLength - 1;
					int startDepth = j;
					Formula* zj = gruParams.createHiddenNodeAtTimeAtDepth(model, datum, 0, k, startTime, startDepth, inputDropoutRate, hiddenDropoutRate);
					// Multiply it by the weight corresponding to the provided label and the current dimension.
					Formula* prod = model.getFormulaObject(TIMES, 2);
					prod->addChild(zj);
					prod->addChild(model.getCachedParameterFormula(gruParams.m_softmaxWeights.at(j).at(k).at(label)));
					sumOverDimensions->addChild(prod);
				}
			}
		}
	}
	for (unsigned int sruIndex = 0; sruIndex < params.m_sruParams.size(); ++sruIndex) {
		DeepSRU& sruParams = params.m_sruParams[sruIndex];
		// Go through all hidden layers for the current deep SRU.
		for (unsigned int j = 0; j < sruParams.m_layerWidths.size(); ++j) {
			// Go through all entries in the current hidden layer for the current deep SRU.
			for (unsigned int k = 0; k < sruParams.m_layerWidths[j]; ++k) {
				bool dropout = false;
				if (softmaxDropoutRate > 0.0 && getRandomDouble() < softmaxDropoutRate) dropout = true;
				if (!dropout) {
					// Create the Formula node for the jth dimension of the final hidden vector (i.e., the time t is the number of words in the sentence).
					int startTime = 0;
					if (sruParams.m_forward) startTime = datum.m_sequenceLength - 1;
					int startDepth = j;
					Formula* zj = sruParams.createHiddenNodeAtTimeAtDepth(model, datum, 0, k, startTime, startDepth, inputDropoutRate, hiddenDropoutRate);
					// Multiply it by the weight corresponding to the provided label and the current dimension.
					Formula* prod = model.getFormulaObject(TIMES, 2);
					prod->addChild(zj);
					prod->addChild(model.getCachedParameterFormula(sruParams.m_softmaxWeights.at(j).at(k).at(label)));
					sumOverDimensions->addChild(prod);
				}
			}
		}
	}
	if (params.m_usingCNN) {
		for (int i = 0; i < params.m_cnnParams.getNumberOfFilters(); ++i) {
			bool dropout = false;
			if (softmaxDropoutRate > 0.0 && getRandomDouble() < softmaxDropoutRate) dropout = true;
			if (!dropout) {
				Formula* ci = NULL;
				// Create the Formula node for c_i using the ith convolutional filter.
				ci = params.m_cnnParams.createFilterPoolingOutputNode(model, datum, i);
				// Multiply it by the weight corresponding to the provided label and the current dimension i.
				Formula* prod = model.getFormulaObject(TIMES, 2);
				prod->addChild(ci);
				prod->addChild(model.getCachedParameterFormula(params.m_cnnParams.m_softmaxWeights.at(label).at(i)));
				sumOverDimensions->addChild(prod);
			}
		}
	}
	// Add the softmax bias.
	PARAM_ID singleBias;
	if (params.m_lstmParams.size() > 0) {
		singleBias = params.m_lstmParams[0].m_softmaxBias.at(0).at(label);
	} else if (params.m_gruParams.size() > 0) {
		singleBias = params.m_gruParams[0].m_softmaxBias.at(0).at(label);
	} else if (params.m_sruParams.size() > 0) {
		singleBias = params.m_sruParams[0].m_softmaxBias.at(0).at(label);
	} else if (params.m_usingCNN) {
		singleBias = params.m_cnnParams.m_softmaxBias.at(label);
	} else {
		cerr << "Error: no LSTM, GRU, SRU, or CNN parameters initialized!" << endl;
		singleBias = 0;
	}
	sumOverDimensions->addChild(model.getCachedParameterFormula(singleBias));
	return sumOverDimensions;
}

Formula* createLossFDAGForWordSequenceDatum(Model& model, ModelParams& params, WordSequenceDatum& datum, int exampleNumber,
		double softmaxDropoutRate, double inputDropoutRate, double hiddenDropoutRate) {
	// Create a term for the correct label's score, which is negated in the loss.
	Formula* negCorrect = model.getFormulaObject(NEG, 1);
	// Create a sum node to sum over labels.
	Formula* sum = model.getFormulaObject(PLUS, params.m_numLabels);
	// Sum over labels.
	for (unsigned int label = 0; label < params.m_numLabels; ++label) {
		Formula* score = createSoftmaxScoreForLabel(model, params, label, datum, softmaxDropoutRate, inputDropoutRate, hiddenDropoutRate);
		// If the current label is the true label, add the score to the negated correct term.
		if (label == datum.m_label) negCorrect->addChild(score);
		// Always add the exponentiated label score to the sum over labels.
		Formula* expScore = model.getFormulaObject(EXP, 1);
		expScore->addChild(score);
		sum->addChild(expScore);
	}
	// Put a log in front of the sum of exponentiated label scores.
	Formula* logSum = model.getFormulaObject(LOG, 1);
	logSum->addChild(sum);

	// Add the negated correct score to the log-sum over all exponentiated label scores.
	Formula* negCorrectPlusSum = model.getFormulaObject(PLUS, 2);
	negCorrectPlusSum->addChild(negCorrect);
	negCorrectPlusSum->addChild(logSum);
	// Return the sum.
	return negCorrectPlusSum;
}

bool decode(Model& model, ModelParams& params, WordSequenceDatum& datum, int exampleNumber, int& prediction) {
	double maxScore = neg_inf;
	int maxLabel = -1;
	for (int i = 0; i < params.m_numLabels; ++i) {
		Formula* score = createSoftmaxScoreForLabel(model, params, i, datum, -1.0, -1.0, -1.0);
		double currScore = score->evaluate();
		if (currScore > maxScore) {
			maxScore = currScore;
			maxLabel = i;
		}
	}
	prediction = maxLabel;
	if (maxLabel == -1) {
		cerr << "Error: no label found with score > negative infinity! on line #" << exampleNumber << endl;
	} else {
		if (maxLabel == datum.m_label) return true;
	}
	return false;
}

double doDecoding(Model& model, ModelParams& params, vector<WordSequenceDatum>& data, bool usingAccuracy) {
	int numCorrect = 0;
	int total = 0;
	int numOnesCorrect = 0;
	int numOnesPredicted = 0;
	int numTrueOnes = 0;
	for (unsigned int i = 0; i < data.size(); ++i) {
		WordSequenceDatum& datum = data[i];
		// We are processing a new example, so reset Formula data structures in the model.
		resetForNewExample(model, params, datum);
		int predictedLabel = -1;
		bool correct = decode(model, params, datum, i, predictedLabel);
		if (predictedLabel == 1) {
			++numOnesPredicted;
			if (correct) {
				++numOnesCorrect;
			}
		}
		if (datum.m_label == 1) {
			++numTrueOnes;
		}
		datum.doneProcessing();
		if (correct) ++numCorrect;
		++total;
	}
	if (usingAccuracy) {
		return (((double)numCorrect) / ((double)total));
	} else {
		if (numTrueOnes == 0) {
			cerr << "Error: no true labels provided!" << endl;
		}
		// Compute precision and recall.
		if (numOnesPredicted == 0 || numTrueOnes == 0) return 0.0;
		double precision = (((double)numOnesCorrect) / ((double)numOnesPredicted));
		double recall = (((double)numOnesCorrect) / ((double)numTrueOnes));
		// Return the harmonic mean of precision and recall.
		return (2.0 * precision * recall) / (precision + recall);
	}
}

struct CompareIntDoublePairSet {
	bool operator() (const pair<int, double>& a, const pair<int, double>& b) const {
		return a.second > b.second;
	}
};

struct GateScore {
	int m_seq;	// Index of the sequence.
	int m_pos;	// Index of the position within the sequence.
	double m_score;	// Value of the gate.
};

struct GateScoreLess {
    bool operator ()(const GateScore& a, const GateScore& b) const {
    		return (a.m_score < b.m_score);
    }
};

void printGateScore(const GateScore& gs, const WordSequenceDatum& d, double numDimensions = 1.0) {
	cerr << gs.m_seq << ", 0-" << gs.m_pos << " (" << (gs.m_score / numDimensions) << "):" << endl
		 << d.toStringInsertStringAfterWordIndex("|||", gs.m_pos) << endl;
}

void printGateScores(vector<vector<GateScore> >& gateScores, const vector<WordSequenceDatum>& data, const string& gateType, int numToPrint) {
	// First, sum over dimensions to produce a new vector. We do this first because below we will sort the vectors, losing the order.
	vector<GateScore> gateScoresSumDimensions;
	if (gateScores.size() == 0) return;
	int numSequencePositions = gateScores[0].size();
	double numDimensions = (double)(gateScores.size());
	// Loop over sequence positions.
	for (unsigned int i = 0; i < numSequencePositions; ++i) {
		GateScore gs;
		gs.m_seq = gateScores[0][i].m_seq;
		gs.m_pos = gateScores[0][i].m_pos;
		gs.m_score = 0.0;
		// Loop over dimensions.
		for (unsigned int j = 0; j < gateScores.size(); ++j) {
			GateScore& currGS = gateScores[j][i];
			if (gs.m_seq != currGS.m_seq) cerr << "Error: i = " << i << ", j = " << j << ", gs.m_seq = " << gs.m_seq << ", currGS.m_seq = " << currGS.m_seq << endl;
			if (gs.m_pos != currGS.m_pos) cerr << "Error: i = " << i << ", j = " << j << ", gs.m_pos = " << gs.m_pos << ", currGS.m_pos = " << currGS.m_pos << endl;
			gs.m_score += currGS.m_score;
		}
		gateScoresSumDimensions.push_back(gs);
	}
	// Now sort the single vector that has been summed across all dimensions.
	sort(gateScoresSumDimensions.begin(), gateScoresSumDimensions.end(), GateScoreLess());
	cerr << "For " << gateType << ", average across dimensions, smallest matches:" << endl;
	// Print the first numToPrint and last numToPrint sequence prefixes that activate the current dimension of the current gate.
	for (unsigned int j = 0; j < numToPrint; ++j) {
		printGateScore(gateScoresSumDimensions[j], data[gateScoresSumDimensions[j].m_seq], numDimensions);
	}
	cerr << endl;
	cerr << "For " << gateType << ", average across dimensions, largest matches:" << endl;
	int startIndex = gateScoresSumDimensions.size() - numToPrint;
	int endIndex = gateScoresSumDimensions.size();
	for (unsigned int j = endIndex - 1; j >= startIndex; --j) {
		printGateScore(gateScoresSumDimensions[j], data[gateScoresSumDimensions[j].m_seq], numDimensions);
	}
	cerr << endl << endl;

	// Go through the dimensions of the gate vector.
	for (unsigned int i = 0; i < gateScores.size(); ++i) {
		sort(gateScores[i].begin(), gateScores[i].end(), GateScoreLess());
		cerr << "For " << gateType << ", dimension " << i << ", smallest matches:" << endl;
		// Print the first numToPrint and last numToPrint sequence prefixes that activate the current dimension of the current gate.
		for (unsigned int j = 0; j < numToPrint; ++j) {
			printGateScore(gateScores[i][j], data[gateScores[i][j].m_seq]);
		}
		cerr << endl;
		cerr << "For " << gateType << ", dimension " << i << ", largest matches:" << endl;
		int startIndex = gateScores[i].size() - numToPrint;
		int endIndex = gateScores[i].size();
		for (unsigned int j = endIndex - 1; j >= startIndex; --j) {
			printGateScore(gateScores[i][j], data[gateScores[i][j].m_seq]);
		}
		cerr << endl << endl;
	}

}

void printRecurrentNeuralNetworkNearestNeighbors(vector<WordSequenceDatum>& data, Model& model, ModelParams& params) {
	// Go through the data and embed each sequence.
	bool doDropoutFalse = false;//, scaleWeightsFalse = false;
	//bool forwardTrue = true, forwardFalse = false;

	vector<vector<double> > sequenceEmbeddings;
	int numberToPrint = 3000;
	if (numberToPrint > data.size()) numberToPrint = data.size();
	sequenceEmbeddings.reserve(numberToPrint);
	// For GRU reset and update gates, store the positions in the sequences that maximize and minimize the gates, for each dimension.
	// Outer vector ranges over dimensions, next vector ranges over tuples containing sequence index, position within sequence,
	// and score of the gate at that position in that sequence.
	vector<vector<GateScore> > resetGateScores;
	vector<vector<GateScore> > updateGateScores;
	vector<vector<GateScore> > updatePlusResetGateScores;
	vector<vector<GateScore> > updateMinusResetGateScores;
	// Initialize these two vectors with the dimension of the gate vectors.
	bool computeGateScores = false;
	if (params.m_gruParams.size() > 0) {
		computeGateScores = true;
		int dimensionToUse = params.m_gruParams[0].m_layerWidths[0];
		resetGateScores.reserve(dimensionToUse);
		updateGateScores.reserve(dimensionToUse);
		updatePlusResetGateScores.reserve(dimensionToUse);
		updateMinusResetGateScores.reserve(dimensionToUse);
		for (unsigned int i = 0; i < dimensionToUse; ++i) {
			vector<GateScore> emptyVector;
			resetGateScores.push_back(emptyVector);
			updateGateScores.push_back(emptyVector);
			updatePlusResetGateScores.push_back(emptyVector);
			updateMinusResetGateScores.push_back(emptyVector);
		}
	}

	for (unsigned int i = 0; i < numberToPrint; ++i) {
		WordSequenceDatum& datum = data[i];
		// We are processing a new example, so reset Formula data structures in the model.
		resetForNewExample(model, params, datum);
		// Get the RNN vector for the sequence.
		vector<double> currEmbedding;
		int sizeToReserve = 0;
		for (unsigned int lstmIndex = 0; lstmIndex < params.m_lstmParams.size(); ++lstmIndex) {
			sizeToReserve += params.m_lstmParams[lstmIndex].getTotalHiddenVectorSize();
		}
		for (unsigned int gruIndex = 0; gruIndex < params.m_gruParams.size(); ++gruIndex) {
			sizeToReserve += params.m_gruParams[gruIndex].getTotalHiddenVectorSize();
		}
		currEmbedding.reserve(sizeToReserve);

		for (unsigned int lstmIndex = 0; lstmIndex < params.m_lstmParams.size(); ++lstmIndex) {
			DeepLSTM& lstmParams = params.m_lstmParams[lstmIndex];
			// Go through all hidden layers for the current deep LSTM.
			for (unsigned int j = 0; j < lstmParams.m_layerWidths.size(); ++j) {
				// Go through all entries in the current hidden layer for the current deep LSTM.
				for (unsigned int k = 0; k < lstmParams.m_layerWidths[j]; ++k) {
					// Create the Formula node for the kth dimension of the final hidden vector at layer j (i.e., the time t is the number of words in the sentence).
					int startTime = 0;
					if (lstmParams.m_forward) startTime = datum.m_sequenceLength - 1;
					Formula* zj = lstmParams.createHiddenNodeAtTimeAtDepth(model, datum, 0, k, startTime, j, doDropoutFalse, -1.0);
					double currEntry = zj->evaluate();
					currEmbedding.push_back(currEntry);
				}
			}
		}
		for (unsigned int gruIndex = 0; gruIndex < params.m_gruParams.size(); ++gruIndex) {
			DeepGRU& gruParams = params.m_gruParams[gruIndex];
			// Go through all hidden layers for the current deep LSTM.
			for (unsigned int j = 0; j < gruParams.m_layerWidths.size(); ++j) {
				// Go through all entries in the current hidden layer for the current deep LSTM.
				for (unsigned int k = 0; k < gruParams.m_layerWidths[j]; ++k) {
					// Create the Formula node for the kth dimension of the final hidden vector at layer j (i.e., the time t is the number of words in the sentence).
					int startTime = 0;
					if (gruParams.m_forward) startTime = datum.m_sequenceLength - 1;
					Formula* zj = gruParams.createHiddenNodeAtTimeAtDepth(model, datum, 0, k, startTime, j, doDropoutFalse, -1.0);
					double currEntry = zj->evaluate();
					currEmbedding.push_back(currEntry);
					// Also compute GRU reset and update gates at every position in the sequence.
					if (computeGateScores) {
						if (gruIndex == 0 && j == 0) {
							for (unsigned int position = 0; position < datum.m_sequenceLength; ++position) {
								Formula* resetGateFormula = gruParams.createResetGateAtTimeAtDepth(model, datum, 0, k, position, j, doDropoutFalse, -1.0);
								GateScore gsr;
								gsr.m_seq = i;
								gsr.m_pos = position;
								gsr.m_score = resetGateFormula->evaluate();
								resetGateScores[k].push_back(gsr);

								Formula* updateGateFormula = gruParams.createUpdateGateAtTimeAtDepth(model, datum, 0, k, position, j, doDropoutFalse, -1.0);
								GateScore gsu;
								gsu.m_seq = i;
								gsu.m_pos = position;
								gsu.m_score = updateGateFormula->evaluate();
								updateGateScores[k].push_back(gsu);

								GateScore gsupr;
								gsupr.m_seq = i;
								gsupr.m_pos = position;
								gsupr.m_score = gsr.m_score + gsu.m_score;
								updatePlusResetGateScores[k].push_back(gsupr);

								GateScore gsumr;
								gsumr.m_seq = i;
								gsumr.m_pos = position;
								gsumr.m_score = gsu.m_score - gsr.m_score;
								updateMinusResetGateScores[k].push_back(gsumr);
							}
						}
					}
				}
			}
		}
		sequenceEmbeddings.push_back(currEmbedding);
	}

	// Then go back through embedded sequences and compute all pairwise cosine similarities.
	vector<vector<double> > similarities;
	similarities.reserve(numberToPrint);
	for (unsigned int i = 0; i < numberToPrint; ++i) {
		vector<double> emptySimilarities(numberToPrint, -99.0);
		similarities.push_back(emptySimilarities);
	}
	for (unsigned int i = 0; i < sequenceEmbeddings.size(); ++i) {
		for (unsigned int j = i; j < sequenceEmbeddings.size(); ++j) {
			// Compute cosine similarity between embeddings of sequences i and j.
			double currSim = cosineSimilarity(sequenceEmbeddings[i], sequenceEmbeddings[j]);
			similarities[i][j] = currSim;
			similarities[j][i] = currSim;
		}
	}

	// For each sequence, find its k-nearest neighbors.
	unsigned int numNeighbors = 10;
	cerr << "Printing k-nearest neighbors of sequence embeddings:" << endl;
	for (unsigned int i = 0; i < similarities.size(); ++i) {
		set<pair<int, double>, CompareIntDoublePairSet > pq;
		for (unsigned int j = 0; j < similarities[i].size(); ++j) {
			double currSim = similarities[i][j];
			// Insert the current similarity into the priority queue.
			pq.insert(make_pair(j, currSim));
		}
		cerr << "For sequence:" << endl;
		cerr << data[i].toString() << endl;
		for (unsigned int k = 0; k < numNeighbors && pq.size() > 0; ++k) {
			//string currWordSequence = pq.top().first;
			//double filterSum = pq.top().second;
			int datumIndex = pq.begin()->first;
			double datumSim = pq.begin()->second;
			cerr << datumSim << ":\t" << data[datumIndex].toString() << endl;
			//topMatches.push_back(currWordSequence);
			pq.erase(pq.begin());
			//pq.pop();
		}
	}
	if (computeGateScores) {
		int gateScoresToPrint = 20;
		printGateScores(resetGateScores, data, "reset", gateScoresToPrint);
		printGateScores(updateGateScores, data, "update", gateScoresToPrint);
		printGateScores(updatePlusResetGateScores, data, "update+reset", gateScoresToPrint);
		printGateScores(updateMinusResetGateScores, data, "update-reset", gateScoresToPrint);
	}
}

void loadWordVectors(ModelParams& params, const vector<string>& trainLines, const vector<string>& devLines,
		const vector<string>& testLines, const string& wordVectorFilename, bool demoMode) {
	// Build the relevant word vector vocabulary from the data lines.
	set<string> vocab;
	if (!demoMode) {
		addWordsFromLinesToVocabulary(trainLines, vocab);
		addWordsFromLinesToVocabulary(devLines, vocab);
		addWordsFromLinesToVocabulary(testLines, vocab);
	}
	// Read in word vectors corresponding to the entries in the vocabulary.
	readWordVectorFile(wordVectorFilename, vocab, params.m_wordVectors, params.m_wordVectorDimensionality, params.m_unknownWordVector);
}

void initializeNetworkFromConfiguration(Model& model, ModelParams& params, const string& cfg,
		const string& networkType, bool forward) {
	if (cfg != "0") {
		bool initSoftmaxTrue = true;
		int inputDimensionToUse = params.m_inputDimension;
		if (params.m_mappedInputDimension > 0) inputDimensionToUse = params.m_mappedInputDimension;
		vector<string> networks;
		tokenize(cfg, '.', networks);
		for (unsigned int i = 0; i < networks.size(); ++i) {
			// Within each network, tokenize the number of networks from the sequence of hidden layer widths.
			vector<string> toks;
			tokenize(networks[i], ':', toks);
			if (toks.size() < 2) {
				cerr << "Error: no network parameters given in token \"" << networks[i] << "\"." << endl;
			} else {
				int numNetworks = atoi(toks[0].c_str());
				if (networkType == "lstm") {
					bool useInputGates = true;
					bool useForgetGates = true;
					bool useOutputGates = true;
					bool useLookahead = false;
					if (toks.size() >= 3) {
						useInputGates = false;
						useForgetGates = false;
						useOutputGates = false;
						for (unsigned int configIndex = 0; configIndex < toks[2].length(); ++configIndex) {
							if (toks[2].at(configIndex) == 'l') {
								useLookahead = true;
							} else if (toks[2].at(configIndex) == 'i') {
								useInputGates = true;
							} else if (toks[2].at(configIndex) == 'f') {
								useForgetGates = true;
							} else if (toks[2].at(configIndex) == 'o') {
								useOutputGates = true;
							}
						}
					}
					for (int j = 0; j < numNetworks; ++j) {
						cerr << "Initializing deep LSTM " << (j+1) << " with layer widths " << toks[1];
						if (useLookahead) cerr << ", with lookahead";
						if (!useInputGates) cerr << ", without input gates";
						if (!useForgetGates) cerr << ", without forget gates";
						if (!useOutputGates) cerr << ", without output gates";
						cerr << endl;
						DeepLSTM lstmParams;
						lstmParams.initializeParameters(model, inputDimensionToUse, params.m_numLabels,
							forward, toks[1], 1, initSoftmaxTrue, useInputGates, useForgetGates, useOutputGates, useLookahead);
						params.m_lstmParams.push_back(lstmParams);
					}
				} else if (networkType == "gru") {
					for (int j = 0; j < numNetworks; ++j) {
						cerr << "Initializing deep GRU " << (j+1) << " with layer widths " << toks[1] << endl;
						DeepGRU gruParams;
						gruParams.initializeParameters(model, inputDimensionToUse, params.m_numLabels, forward, toks[1], 1, initSoftmaxTrue);
						params.m_gruParams.push_back(gruParams);
					}
				} else if (networkType == "sru") {
					bool useLookahead = false;
					if (toks.size() >= 3 && toks[2] == "1") useLookahead = true;
					bool useOneDimGate = false;

					for (int j = 0; j < numNetworks; ++j) {
						cerr << "Initializing deep SRU " << (j+1) << " with layer widths " << toks[1] << endl;
						DeepSRU sruParams;
						sruParams.initializeParameters(model, inputDimensionToUse, params.m_numLabels, forward, toks[1], 1, initSoftmaxTrue, useLookahead, useOneDimGate);
						params.m_sruParams.push_back(sruParams);
					}
				}
			}
		}
	}
}

void initializeCNNFromConfiguration(Model& model, ModelParams& params, const string& filters) {
	params.m_usingCNN = false;
	if (filters != "0" && filters != "none" && filters != "NONE" && filters != "") {
		bool initSoftmaxTrue = true;
		int inputDimensionToUse = params.m_inputDimension;
		if (params.m_mappedInputDimension > 0) inputDimensionToUse = params.m_mappedInputDimension;
		params.m_cnnParams.initializeParameters(model, inputDimensionToUse, params.m_numLabels, filters, initSoftmaxTrue);
		params.m_usingCNN = true;
	} else {
		cerr << "Not using any CNN filters." << endl;
	}
}

void runDemo(Model& model, ModelParams& params);

double trainAndTest(Model& model, const vector<string>& trainLines, const vector<string>& devLines,
		const vector<string>& testLines, const string& wordVectorFilename,
		const string& forwardLSTMs, const string& backwardLSTMs,
		const string& forwardGRUs, const string& backwardGRUs,
		const string& forwardSRUs, const string& backwardSRUs,
		const string& filters,
		const string& algorithm, double stepsize, int minibatchSize, int numEpochs, int decodeEveryK,
		double softmaxDropoutRate, double inputDropoutRate, double hiddenDropoutRate,
		bool updateWordVectors, bool initWordVectorsRandomly, int mapWordVectorsToDim,
		double gradientNormThreshold, const string& evalMetric, bool doFiniteDifferenceChecking, bool demoMode) {

	struct timeval initialTime, endTime, time1, time2, time3, time4,
		timeBuildStart, timeBuildEnd, timeEvalStart, timeEvalEnd, timeDecodeStart, timeDecodeEnd;
	gettimeofday(&initialTime, 0);
	bool usingAccuracy = true;
	if (evalMetric == "F1") {
		usingAccuracy = false;
		cerr << "Using F1 instead of accuracy." << endl;
	}

	ModelParams params;
	// First, load word vectors corresponding to words in train/dev/test data.
	loadWordVectors(params, trainLines, devLines, testLines, wordVectorFilename, demoMode);

	model.resetGradientVector();
	//double stepsize = 0.001;
	map<string, int> unkWordCounts;
	vector<WordSequenceDatum> trainData, devData, testData;
	loadWordSequenceData(trainLines, trainData, params.m_wordVectors, "train", unkWordCounts);
	loadWordSequenceData(devLines, devData, params.m_wordVectors, "dev", unkWordCounts);
	loadWordSequenceData(testLines, testData, params.m_wordVectors, "test", unkWordCounts);
	printUnknownWordCounts(unkWordCounts);

	bool trainTrue = true;
	bool trainFalse = false;

	// Initialize all parameters shared across LSTMs/GRUs.
	double scalingFactor = 0.2;
	initializeSharedParameters(model, params, params.m_wordVectorDimensionality, trainData, devData,
		testData, updateWordVectors, initWordVectorsRandomly, mapWordVectorsToDim, scalingFactor, demoMode);
	// Initialize parameters for each LSTM/GRU/SRU/CNN.
	initializeNetworkFromConfiguration(model, params, forwardLSTMs, "lstm", true);
	initializeNetworkFromConfiguration(model, params, backwardLSTMs, "lstm", false);
	initializeNetworkFromConfiguration(model, params, forwardGRUs, "gru", true);
	initializeNetworkFromConfiguration(model, params, backwardGRUs, "gru", false);
	initializeNetworkFromConfiguration(model, params, forwardSRUs, "sru", true);
	initializeNetworkFromConfiguration(model, params, backwardSRUs, "sru", false);
	initializeCNNFromConfiguration(model, params, filters);

	if (params.m_updateWordVectors) {
		cerr << "Getting word vector params for training data" << endl;
		getWordVectorParamsForWordSequenceData(params.v, trainData, trainTrue);
		cerr << "Getting word vector params for dev data" << endl;
		getWordVectorParamsForWordSequenceData(params.v, devData, trainFalse);
		cerr << "Getting word vector params for test data" << endl;
		getWordVectorParamsForWordSequenceData(params.v, testData, trainFalse);
	}

	// Print nearest neighbors before any learning.
	//printNearestNeighbors(trainData, trainData, params);
	//printNearestNeighbors(trainData, devData, params);
	double maxDevAcc = -1.0;
	double testAccForMaxDevAcc = -1.0;
	//double maxTestAcc = -1.0;

	FLOAT totalBuildTime = 0.0f;
	FLOAT totalEvalTime = 0.0f;
	FLOAT totalDecodeTime = 0.0f;

	int decodeEveryKToUse = decodeEveryK;
	if (decodeEveryKToUse <= 0) {
		decodeEveryKToUse = trainData.size() / 5;
	}

	//bool doDropout = true;
	//if (dropoutRate <= 0.0) doDropout = false;

	bool scaleWeightsForDecoding = false;//doDropout;
	//scaleWeightsForDecoding = false;
	//if (scaleWeightsForDecoding) {
	//	cerr << "Scaling weights for decoding by dropout rate (" << dropoutRate << ")." << endl;
	//}

	Algorithm alg = Model::getAlgorithm(algorithm);
	// Initialize vector of previous updates for SGD w/ momentum.
	vector<double> sgdMomentumUpdates;//(model.getNumParams(), 0.0);
	double momentumCoefficient = 0.95;
	// Initialize sum of squared gradients for AdaGrad and AdaDelta.
	vector<double> sumOfSquaredGradients(model.getNumParams(), 0.0);
	// Initialize sum of squared updates for AdaDelta.
	vector<double> sumOfSquaredUpdates;//(model.getNumParams(), 0.0);
	//double epsilon = 0.000001;
	//double rho = 0.95;
	if (alg != ADAGRAD) {
		cerr << "Error: only AdaGrad is currently supported!" << endl;
	}

	// Initialize sum of squared gradients for AdaGrad.
	//vector<double> sumOfSquaredGradients(getNumParams(), 0.0);
	//int maxNumFormulasUsed = getNumberOfFormulasUsedInPool();
	int maxUnaryFormulasUsed = 0;
	int maxBinaryFormulasUsed = 0;
	int maxFertileFormulasUsed = 0;
	int maxBarrenFormulasUsed = 0;
	int unaryFormulasUsed = 0;
	int binaryFormulasUsed = 0;
	int fertileFormulasUsed = 0;
	int barrenFormulasUsed = 0;
	model.getNumberOfFormulasUsedInPool(maxUnaryFormulasUsed, maxBinaryFormulasUsed, maxFertileFormulasUsed, maxBarrenFormulasUsed);

	// Train for numEpochs iterations of training.
	for (unsigned int iter = 1; iter <= numEpochs; ++iter) {
		double iterLoss = 0.0;
		double iterGradientNorm = 0.0;

		//int iterNumCorrect = 0;
		gettimeofday(&time1, 0);
		gettimeofday(&time3, 0);

		for (unsigned int exampleIndex = 0; exampleIndex < trainData.size(); ++exampleIndex) {
			WordSequenceDatum& datum = trainData[exampleIndex];
			// Sample a data point.
			//int exampleIndex = i;//rand() % (trainData.size());
			// We are processing a new example, so reset Formula data structures in the model.
			resetForNewExample(model, params, datum);

			//cerr << "Computing gradient for sentence \"" << sentenceTokens[0] << " " << sentenceTokens[1]
			//	 << "\" with true label " << trueLabel << endl;

			// Create the formula containing the loss for the current example.
			gettimeofday(&timeBuildStart, 0);
			Formula* datumLossFDAG = createLossFDAGForWordSequenceDatum(model, params, datum, exampleIndex, softmaxDropoutRate, inputDropoutRate, hiddenDropoutRate);
			gettimeofday(&timeBuildEnd, 0);
			totalBuildTime += computeTimeDifference(timeBuildStart, timeBuildEnd);
			model.getNumberOfFormulasUsedInPool(unaryFormulasUsed, binaryFormulasUsed, fertileFormulasUsed, barrenFormulasUsed);
			if (unaryFormulasUsed > maxUnaryFormulasUsed) {
				maxUnaryFormulasUsed = unaryFormulasUsed;
				cerr << "Info: New unary formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << unaryFormulasUsed << endl;
			}
			if (binaryFormulasUsed > maxBinaryFormulasUsed) {
				maxBinaryFormulasUsed = binaryFormulasUsed;
				cerr << "Info: New binary formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << binaryFormulasUsed << endl;
			}
			if (fertileFormulasUsed > maxFertileFormulasUsed) {
				maxFertileFormulasUsed = fertileFormulasUsed;
				cerr << "Info: New fertile formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << fertileFormulasUsed << endl;
			}
			if (barrenFormulasUsed > maxBarrenFormulasUsed) {
				maxBarrenFormulasUsed = barrenFormulasUsed;
				cerr << "Info: New barren formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << barrenFormulasUsed << endl;
			}
			//int currNumFormulasUsed = getNumberOfFormulasUsedInPool();
			//if (currNumFormulasUsed > maxNumFormulasUsed) {
			//	cerr << "Info: New formula watermark reached on example " << exampleIndex << ", iter " << iter << ": " << currNumFormulasUsed << endl;
			//	maxNumFormulasUsed = currNumFormulasUsed;
			//}
			// Compute gradient and loss for current example.
			gettimeofday(&timeEvalStart, 0);
			double loss = model.evalAndComputeDerivatives(datumLossFDAG, exampleIndex, doFiniteDifferenceChecking);
			if (loss != loss) {
				//cerr << "Error: nan loss, example " << exampleIndex << " with label " << trainData[exampleIndex].m_label << endl;
				cerr << "Error: nan loss, at sentence " << exampleIndex << ":\"" << trainData[exampleIndex].toString() <<
					"\" with true label " << trainData[exampleIndex].m_label << endl;
			} else if (loss == neg_inf) {
					//cerr << "Error: nan loss, example " << exampleIndex << " with label " << trainData[exampleIndex].m_label << endl;
					cerr << "Error: -inf loss, at sentence " << exampleIndex << ":\"" << trainData[exampleIndex].toString() <<
						"\" with true label " << trainData[exampleIndex].m_label << endl;
			} else if (loss == pos_inf) {
					//cerr << "Error: nan loss, example " << exampleIndex << " with label " << trainData[exampleIndex].m_label << endl;
					cerr << "Error: inf loss, at sentence " << exampleIndex << ":\"" << trainData[exampleIndex].toString() <<
						"\" with true label " << trainData[exampleIndex].m_label << endl;
			} else {
				iterLoss += loss;
			}
			gettimeofday(&timeEvalEnd, 0);
			totalEvalTime += computeTimeDifference(timeEvalStart, timeEvalEnd);
			datum.doneProcessing();

			// If we're at the end of a mini-batch or at the end of the training data, make an update.
			bool doUpdate = false;
			if ((exampleIndex + 1) % minibatchSize == 0 || trainData.size() <= 1 || exampleIndex == trainData.size() - 1) doUpdate = true;
			if (doUpdate) {
				if (gradientNormThreshold > 0.0) {
					double gradientNorm = model.computeGradientVectorNorm();
					if (gradientNorm >= gradientNormThreshold) {
						//cerr << "Info: gradientNorm = " << gradientNorm << "; Scaling gradient vector by " << (gradientNormThreshold / gradientNorm) << endl;
						model.scaleGradientVectorByConstant(gradientNormThreshold / gradientNorm);
						//cerr << "Info: new gradientNorm = " << model.computeGradientVectorNorm() << endl;
					}
				}
				double gradientSquaredNorm = 0.0;
				if (alg == SGD_MOMENTUM) {
					gradientSquaredNorm = model.doSGDMomentumUpdate(sgdMomentumUpdates, stepsize, momentumCoefficient);
				} else if (alg == ADAGRAD) {
					gradientSquaredNorm = model.doAdaGradUpdate(sumOfSquaredGradients, stepsize);
				} else if (alg == ADADELTA){
					gradientSquaredNorm = model.doAdaDeltaUpdate(sumOfSquaredGradients, sumOfSquaredUpdates, stepsize, momentumCoefficient);
				}
				//double l2norm = sqrt(gradientSquaredNorm);
				//iterGradientNorm += l2norm;
				iterGradientNorm += model.computeGradientVectorNorm();
				//if (l2NormConstraint > 0.0) {// && l2norm > l2NormConstraint) {
				//	scaleWeightsToMatchL2Norm(params, l2NormConstraint);
				//}
				model.resetGradientVector();

				//double gradientSquaredNorm = doAdaGradUpdate(sumOfSquaredGradients, stepsize);
				//double l2norm = sqrt(gradientSquaredNorm);
				//iterGradientNorm += l2norm;
				//if (l2NormConstraint > 0.0) {// && l2norm > l2NormConstraint) {
				//	scaleWeightsToMatchL2Norm(params, l2NormConstraint);
				//}
				//resetGradientVector();
			}

			if ((exampleIndex + 1) % decodeEveryKToUse == 0) {
				gettimeofday(&timeDecodeStart, 0);
				gettimeofday(&time4, 0);
				FLOAT timek = computeTimeDifference(time3, time4);

				// Decode dev data every decodeEveryK examples.
				if (devData.size() > 0) {
					// If we have dev data, decode the dev data.
					double devAccuracy = doDecoding(model, params, devData, usingAccuracy);
					// If it's the best dev accuracy we've seen so far, decode the test data.
					if (devAccuracy > maxDevAcc) {
						maxDevAcc = devAccuracy;
						testAccForMaxDevAcc = doDecoding(model, params, testData, usingAccuracy);
					}
					//if (testAccuracy > maxTestAcc) maxTestAcc = testAccuracy;
					cerr << "it " << iter << ", " << exampleIndex << "\tdacc\t" << devAccuracy << "\tmaxd\t" << maxDevAcc
						 << "\ttformaxd\t" << testAccForMaxDevAcc; // << "\tmaxt\t" << maxTestAcc;
				} else {
					// Here, we have no dev data, so just decode the test data.
					testAccForMaxDevAcc = doDecoding(model, params, testData, usingAccuracy);
					cerr << "it " << iter << ", " << exampleIndex << "\ttacc\t" << testAccForMaxDevAcc;
				}

				cerr << "\ttimek\t" << timek << endl;
				gettimeofday(&time3, 0);
				gettimeofday(&timeDecodeEnd, 0);
				totalDecodeTime += computeTimeDifference(timeDecodeStart, timeDecodeEnd);
				cerr << "times:\tbuild\t" << totalBuildTime << "\teval\t" << totalEvalTime
					 << "\tdec\t" << totalDecodeTime << endl;
				//printFilterNearestNeighbors(trainData, params);

				//cerr << "times:\treset\t" << totalResetTime << "\tbuild\t" << totalBuildTime << "\teval\t" << totalEvalTime
				//	 << "\tupdate\t" << totalUpdateTime << "\tdec\t" << totalDecodeTime << endl;
			}
		}
		//if (iter % 10 == 0) cerr << "it " << iter << "\tloss\t" << iterLoss <<
		//		"\tparam_norm\t" << computeParameterVectorNorm() <<
		//		"\tgrad_norm\t" << iterGradientNorm << endl;
		cerr << "it " << iter << "\tloss\t" << iterLoss << "\tparam_norm\t" << model.computeParameterVectorNorm() <<
				"\tgrad_norm\t" << iterGradientNorm;
		//double acc = ((double)iterNumCorrect) / ((double)(trainData.size()));
		//cerr << "\ttracc\t" << acc;
		//double trainAccuracy = doDecoding(trainData, m, w, dim, numLabels);
		//cerr << "\ttacc:\t" << trainAccuracy;


		if (devData.size() > 0) {
			// If we have dev data, decode the dev data.
			double devAccuracy = doDecoding(model, params, devData, usingAccuracy);
			// If it's the best dev accuracy we've seen so far, decode the test data.
			if (devAccuracy > maxDevAcc) {
				maxDevAcc = devAccuracy;
				testAccForMaxDevAcc = doDecoding(model, params, testData, usingAccuracy);
			}
			//if (testAccuracy > maxTestAcc) maxTestAcc = testAccuracy;
			cerr << "\tdacc\t" << devAccuracy << "\tmaxd\t" << maxDevAcc
			<< "\ttformaxd\t" << testAccForMaxDevAcc;// << "\tmaxt\t" << maxTestAcc;
		} else {
			// Here, we have no dev data, so just decode the test data.
			testAccForMaxDevAcc = doDecoding(model, params, testData, usingAccuracy);
			cerr << "\ttacc\t" << testAccForMaxDevAcc;// << "\tmaxt\t" << maxTestAcc;
		}
		gettimeofday(&time2, 0);
		FLOAT iterTime = computeTimeDifference(time1, time2);
		cerr << "\t\ttime\t" << iterTime;
		cerr << endl;

		if (iter % 10 == 0)
			printRecurrentNeuralNetworkNearestNeighbors(trainData, model, params);
	}
	// Print nearest neighbors for dev bigrams.
	//printNearestNeighbors(trainData, devData, params);

	gettimeofday(&endTime, 0);
	FLOAT cumulativeTime = computeTimeDifference(initialTime, endTime);
	cerr << "Total training time:\t" << cumulativeTime << endl;
	cerr << "Final test accuracy:\t" << testAccForMaxDevAcc << endl;
	if (demoMode) runDemo(model, params);
	return testAccForMaxDevAcc;
}

void displayHelp() {
	cout << "Enter the next word in the sentence and press enter, or enter one of the following:" << endl;
	cout << "\t--r\treset (start new sentence)" << endl;
	cout << "\t--h\tdisplay this help" << endl;
	cout << "\t--q\tquit" << endl;
}

void runDemo(Model& model, ModelParams& params) {

	cout << "Starting demo mode (--h for help). " << endl;
	bool setUWVToFalse = false;
	if (params.m_updateWordVectors) {
		copyTrainedWordVectorParamValuesToWordVectorMap(&model, params.v, params.m_wordVectors);
		params.m_updateWordVectors = false;
		setUWVToFalse = true;
	}

	WordSequenceDatum d;
	string currentSentence = "";
	double positivity = 0.0;
	while (1) {
		string responseFull;
		cout << "(positivity: ";
		printf("%1.3f", positivity);
		cout << ") " << currentSentence << " : ";
		cin >> responseFull;
		cout << endl;
		string response = trimString(responseFull);
		if (response == "--q" || response == "--Q") {
			cout << "Quitting..." << endl;
			if (setUWVToFalse) params.m_updateWordVectors = true;
			return;
		} else if (response == "--R" || response == "--r") {
			// Start a new sentence.
			currentSentence = "";
			d.clear();
			positivity = 0.0;
		} else if (response == "--H" || response == "--h") {
			displayHelp();
		} else {
			// Process the given word.
			if (currentSentence != "") currentSentence += " ";
			currentSentence += response;
			// Tokenize the response into words.
			//vector<string> responseTokens;
			//tokenize(response, ' ', responseTokens);
			d.m_words.push_back(response);
			d.m_sequenceLength = (int)(d.m_words.size());
			//tokenize(lineTokens[0], ' ', d.m_words);

			if (params.m_updateWordVectors) {
				for (unsigned int j = d.m_sequenceLength - 1; j < d.m_sequenceLength; ++j) {
					bool unk = false;
					const vector<PARAM_ID>* vect = getWordVectorParams(params.v, d.m_words[j], unk);
					if (unk) {
						// If the word was unknown, this may just mean that there was no parameter created for it.
						cout << "unknown word: " << d.m_words[j] << endl;
					}
					d.m_vectParams.push_back(vect);
				}

			} else {
				for (unsigned int j = d.m_sequenceLength - 1; j < d.m_sequenceLength; ++j) {
					bool unk = false;
					const vector<double>* vect = getWordVector(params.m_wordVectors, d.m_words[j], unk);
					if (unk) cout << "unknown word: " << d.m_words[j] << endl;
					d.m_vects.push_back(vect);
				}
			}
			// Resets data structures for processing the next example (whether in training or decoding).
			resetForNewExample(model, params, d);
			// Compute the label scores of the sentence.
			vector<double> labelScores;
			double normalizer = 0.0;
			for (unsigned int label = 0; label < params.m_numLabels; ++label) {
				Formula* score = createSoftmaxScoreForLabel(model, params, label, d, -1.0, -1.0, -1.0);
				double expScore = exp(score->evaluate());
				labelScores.push_back(expScore);
				normalizer += expScore;
			}
			if (params.m_lstmParams.size() > 0) {
				double inputGateSum = 0.0, forgetGateSum = 0.0, outputGateSum = 0.0;
				//vector<double> inputGateLayerSums, forgetGateLayerSums, outputGateLayerSums;
				double numGates = 0.0;
				bool anyUsedInputGates = false;
				bool anyUsedForgetGates = false;
				bool anyUsedOutputGates = false;
				for (unsigned int lstmIndex = 0; lstmIndex < params.m_lstmParams.size(); ++lstmIndex) {
					DeepLSTM& lstmParams = params.m_lstmParams[lstmIndex];
					if (lstmParams.m_useInputGates) anyUsedInputGates = true;
					if (lstmParams.m_useForgetGates) anyUsedForgetGates = true;
					if (lstmParams.m_useOutputGates) anyUsedOutputGates = true;
					// Go through all hidden layers for the current deep LSTM.
					for (unsigned int j = 0; j < lstmParams.m_layerWidths.size(); ++j) {
						// Go through all entries in the current hidden layer for the current deep LSTM.
						// Create the Formula node for the kth dimension of the final hidden vector at layer j (i.e., the time t is the number of words in the sentence).
						int startTime = 0;
						if (lstmParams.m_forward) startTime = d.m_sequenceLength - 1;
						int startDepth = j;
						lstmParams.getGateReport(model, d, 0, startTime, startDepth, inputGateSum, forgetGateSum, outputGateSum);
						numGates += ((double)(lstmParams.m_layerWidths[j]));
					}
				}
				double inputGateAverage = inputGateSum / numGates;
				double forgetGateAverage = forgetGateSum / numGates;
				double outputGateAverage = outputGateSum / numGates;
				cout << "LSTM gates for \"" << response << "\":" << endl;
				if (anyUsedInputGates) {
					cout << "input: ";
					printf("%1.3f", inputGateAverage);
				}
				if (anyUsedForgetGates) {
					cout << "\t\tremember: ";
					printf("%1.3f", forgetGateAverage);
				}
				if (anyUsedOutputGates) {
					cout << "\t\toutput: ";
					printf("%1.3f", outputGateAverage);
				}
				cout << endl;
			}
			if (params.m_gruParams.size() > 0) {
				double updateGateSum = 0.0, resetGateSum = 0.0;
				double numGates = 0.0;
				for (unsigned int gruIndex = 0; gruIndex < params.m_gruParams.size(); ++gruIndex) {
					DeepGRU& gruParams = params.m_gruParams[gruIndex];
					// Go through all hidden layers for the current deep GRU.
					for (unsigned int j = 0; j < gruParams.m_layerWidths.size(); ++j) {
						// Go through all entries in the current hidden layer for the current deep GRU.
						// Create the Formula node for the kth dimension of the final hidden vector at layer j (i.e., the time t is the number of words in the sentence).
						int startTime = 0;
						if (gruParams.m_forward) startTime = d.m_sequenceLength - 1;
						int startDepth = j;
						gruParams.getGateReport(model, d, 0, startTime, startDepth, updateGateSum, resetGateSum);
						numGates += ((double)(gruParams.m_layerWidths[j]));
					}
				}
				double updateGateAverage = updateGateSum / numGates;
				double resetGateAverage = resetGateSum / numGates;
				cout << "GRU gates for \"" << response << "\": \tuse word: ";
				printf("%1.3f", updateGateAverage);
				cout << ", \tremember prev memory: ";
				printf("%1.3f", resetGateAverage);
				cout << ", \tuse word and remember: ";
				printf("%1.3f", (updateGateAverage + resetGateAverage) / 2.0);
				cout << ", \tuse word and forget: ";
				printf("%1.3f", updateGateAverage - resetGateAverage);
				cout << endl;
			}
			if (params.m_sruParams.size() > 0) {
				for (unsigned int sruIndex = 0; sruIndex < params.m_sruParams.size(); ++sruIndex) {
					DeepSRU& sruParams = params.m_sruParams[sruIndex];
					if (sruParams.m_forward) {
						// Go through all hidden layers for the current deep SRU.
						for (unsigned int j = 0; j < sruParams.m_layerWidths.size(); ++j) {
							double reduceGateSum = 0.0;
							double numGates = 0.0;
							vector<double> reduceGates;
							// Go through all entries in the current hidden layer for the current deep SRU.
							// Create the Formula node for the kth dimension of the final hidden vector at layer j (i.e., the time t is the number of words in the sentence).
							int startTime = 0;
							if (sruParams.m_forward) startTime = d.m_sequenceLength - 1;
							int startDepth = j;
							sruParams.getGateReport(model, d, 0, startTime, startDepth, reduceGates, reduceGateSum);
							numGates += ((double)(sruParams.m_layerWidths[j]));
							cout << "SRU reduce gate, layer " << j << ": ";
							for (unsigned int k = 0; k < reduceGates.size(); ++k) {
								cout << reduceGates[k];
								cout << ", ";
							}
							cout << "avg: " << (reduceGateSum / numGates) << endl;
						}
					}
				}
			}
			//cout << "neg: " << (labelScores[0] / normalizer) << endl;
			//cout << "pos: " << (labelScores[1] / normalizer) << endl;
			positivity = (labelScores[1] / normalizer);
		}
	}
}

int main(int argc, char** argv) {
    ParamsArray params(argc, argv);
    params.setDescription("Sequence classification using recurrent neural networks (long short-term memory, gated recurrent unit) "
    		"and convolutional neural networks (with n-gram filters).");
    params.addConstraint("train_file",
    		"file containing training data, where each line contains the sentence followed by a tab then the "
    		"label (0 or 1)", false); // not optional
    params.addConstraint("dev_file",
    		"file containing dev data to compute accuracy for, same format as train_file", false); // not optional
    params.addConstraint("test_file",
    		"file containing test data to compute accuracy for, same format as train_file", false); // not optional
    params.addConstraint("cv",
    		"if > 0, will do CV on train_file instead of testing on dev/test; this is the number of folds to use with CV (default: 0)", true); // optional

    params.addConstraint("word_vector_file",
    		"file containing word vectors, where each line contains the word followed by a tab then the "
    		"space-separated vector entries", false); // not optional
    params.addConstraint("update_word_vectors",
    		"if true, will learn update given word vectors during learning (default: true)", true); // optional
    params.addConstraint("init_word_vectors_randomly",
    		"if true, will initialize word vectors randomly before learning (default: false)", true); // optional
    params.addConstraint("map_word_vectors_to_dim", "reduced dimensionality of learned word vector mapping; use a value <= 0 if no mapping is desired (default: 0)", true); // optional

    params.addConstraint("forward_lstms", "configuration of forward LSTMs to train; in "
    		"format \"number:layer_widths:config.number:layer_widths:config...\", where \"number\" "
    		"is the number of forward LSTMs to train that have the following sequence of hidden layer widths"
    		"and configuration, where config is an optional character sequence containing \"l\" for lookahead, "
    		"\"i\" to use input gates, \"f\" to use forget gates, "
    		"and \"o\" to use output gates. If config is not provided, it defaults to \"ifo\" (i.e., standard LSTM without lookahead). "
    		"E.g., to train one forward LSTM with "
    		"one hidden layer of width 10 with lookahead and 2 forward LSTMs with one hidden layer of width 5 and "
    		"a second hidden layer of width 7, pass \"1:10:lifo.2:5,7\". "
    		"Pass \"0\" if no forward LSTMs are desired. (default: 1:10)", true); // optional
    params.addConstraint("backward_lstms", "configuration of backward LSTMs to train, same format as forward_lstms (default: 0)", true); // optional
    params.addConstraint("forward_grus", "configuration of forward GRUs to train, same format as forward_lstms (default: 0)", true); // optional
    params.addConstraint("backward_grus", "configuration of backward GRUs to train, same format as forward_lstms (default: 0)", true); // optional
    params.addConstraint("forward_srus", "configuration of forward SRUs to train, same format as forward_lstms (default: 0)", true); // optional
    params.addConstraint("backward_srus", "configuration of backward SRUs to train, same format as forward_lstms (default: 0)", true); // optional
    params.addConstraint("cnn_filters", "configuration of CNN filters to use, in format \"width:number,width:number,...\".\n"
    		"For example, \"3:20,4:50,5:100\" will use 20 trigram filters, 50 4-gram filters, and 100 5-gram filters.\n"
    		"Leave blank or pass \"0\" if no filters are desired (default: 0)", true); // optional

    params.addConstraint("algorithm", "optimization algorithm; can be \"sgd_momentum\", \"adagrad\", or \"adadelta\" (case insensitive) (default: adadelta)", true); // optional
    params.addConstraint("stepsize", "stepsize for AdaGrad (default: 0.1)", true); // optional
    params.addConstraint("minibatch_size", "mini-batch size for AdaGrad (default: 10)", true); // optional
    params.addConstraint("num_epochs", "number of epochs of learning (default: 3)", true); // optional
    params.addConstraint("decode_every_k", "frequency of dev/test data decoding during training, will decode after this number of examples; if 0, will do after each fifth of the training data (default: 0)", true); // optional
    params.addConstraint("softmax_dropout_rate", "probability of setting a unit to zero in the final vector used in the softmax classifier; use a value <= 0 for no dropout (default: 0.5)", true); // optional
    params.addConstraint("input_dropout_rate", "probability of setting a unit to zero in the transition from the input word vector to the first hidden vector; use a value <= 0 for no dropout (default: 0.5)", true); // optional
    params.addConstraint("hidden_dropout_rate", "probability of setting a unit to zero in subsequent transitions from hidden vectors to hidden vectors; use a value <= 0 for no dropout (default: 0.5)", true); // optional
    params.addConstraint("gradient_norm_threshold", "max gradient norm (\"threshold\" from Pascanu et al paper; use a value <= 0 for no gradient rescaling (default: 1.0)", true); // optional
    params.addConstraint("eval_metric", "evaluation metric, can be \"accuracy\" or \"F1\" (default: accuracy)", true); // optional
    params.addConstraint("demo_mode", "if true, will launch interactive demo mode after training/loading model (default: false)", true); // optional

   	if (!params.runConstraints("run_lstmrnn")) return 0;

	// Read in command-line parameters.
	string trainFilename = params.asString("train_file");
	string devFilename = params.asString("dev_file");
	string testFilename = params.asString("test_file");
	int cvNumFolds = atoi(params.asString("cv", "0").c_str());
	string wordVectorFilename = params.asString("word_vector_file");
	bool updateWordVectors = (params.asString("update_word_vectors", "true") == "true");
	bool initWordVectorsRandomly = (params.asString("init_word_vectors_randomly", "false") == "true");
	int mapWordVectorsToDim = atoi(params.asString("map_word_vectors_to_dim", "0").c_str());
	string forwardLSTMs = params.asString("forward_lstms", "1:10");
	string backwardLSTMs = params.asString("backward_lstms", "0");
	string forwardGRUs = params.asString("forward_grus", "0");
	string backwardGRUs = params.asString("backward_grus", "0");
	string forwardSRUs = params.asString("forward_srus", "0");
	string backwardSRUs = params.asString("backward_srus", "0");
	string cnnFilters = params.asString("cnn_filters", "0");
	string algorithm = params.asString("algorithm", "adadelta");
	double stepsize = atof(params.asString("stepsize", "0.1").c_str());
	int minibatchSize = atoi(params.asString("minibatch_size", "10").c_str());
	int numEpochs = atoi(params.asString("num_epochs", "3").c_str());
	int decodeEveryK = atoi(params.asString("decode_every_k", "0").c_str());
	double softmaxDropoutRate = atof(params.asString("softmax_dropout_rate", "0.5").c_str());
	double inputDropoutRate = atof(params.asString("input_dropout_rate", "0.5").c_str());
	double hiddenDropoutRate = atof(params.asString("hidden_dropout_rate", "0.5").c_str());
	double gradientNormThreshold = atof(params.asString("gradient_norm_threshold", "1.0").c_str());
	string evalMetric = params.asString("eval_metric", "accuracy");
	bool demoMode = (params.asString("demo_mode", "false") == "true");

	// Change following to true to do finite difference checking of gradients.
	bool doFiniteDifferenceChecking = false;
	srand(1);

	struct timeval initialTime, endTime;
	gettimeofday(&initialTime, 0);

	if (cvNumFolds <= 0) {
		vector<string> trainLines, devLines, testLines;
		readLinesFromFile(trainFilename, trainLines);
		readLinesFromFile(devFilename, devLines);
		readLinesFromFile(testFilename, testLines);

		Model model(10000, 2000000, 10000, 60000);
		double accuracy = trainAndTest(model, trainLines, devLines, testLines, wordVectorFilename, forwardLSTMs, backwardLSTMs,
			forwardGRUs, backwardGRUs, forwardSRUs, backwardSRUs, cnnFilters, algorithm, stepsize, minibatchSize, numEpochs, decodeEveryK,
			softmaxDropoutRate, inputDropoutRate, hiddenDropoutRate,
			updateWordVectors, initWordVectorsRandomly, mapWordVectorsToDim, gradientNormThreshold, evalMetric, doFiniteDifferenceChecking, demoMode);
		cerr << "Final accuracy: " << accuracy << endl;
	} else {
		// Load data.
		vector<string> data;
		readLinesFromFile(trainFilename, data);
		//int inputDimension = loadData(dataFile, data, &numLabels);
		// Do n-fold cross-validation.
		double accuracySum = 0.0;
		unsigned int dataSize = data.size();
		int foldSize = dataSize / cvNumFolds;
		vector<double> accuracies;
		for (int i = 0; i < cvNumFolds; ++i) {
			cerr << "For test fold " << i << ", creating test set with examples " << (i*foldSize) << " to " << (((i+1)*foldSize) - 1) << endl;
			// Generate train/test sets for the current fold (approximately based on discretized index cutoffs).
			vector<string> trainLines, devLines, testLines;
			for (unsigned int j = 0; j < dataSize; ++j) {
				if (((int)j) >= (i*foldSize) && ((int)j) < (i+1)*foldSize) {
					// If the current datum falls within the indices of the current test fold, add it to testData.
					testLines.push_back(data[j]);
					//cerr << "test index: " << j << endl;
				} else {
					// Otherwise, add 90% of it to the training set and 10% to the dev set
					if (j % 10 == 0) {
						devLines.push_back(data[j]);
						//cerr << "dev index: " << j << endl;
					} else {
						trainLines.push_back(data[j]);
						//cerr << "train index: " << j << endl;
					}
				}
			}
			cerr << "Size of training set: " << trainLines.size() << endl;
			cerr << "Size of dev set: " << devLines.size() << endl;
			cerr << "Size of test set: " << testLines.size() << endl;
			// Create a model for the current split.
			Model model(10000, 2000000, 10000, 60000);
			// Train and test.
			double currAccuracy = trainAndTest(model, trainLines, devLines, testLines, wordVectorFilename,
				forwardLSTMs, backwardLSTMs, forwardGRUs, backwardGRUs, forwardSRUs, backwardSRUs, cnnFilters,
				algorithm, stepsize, minibatchSize, numEpochs, decodeEveryK,
				softmaxDropoutRate, inputDropoutRate, hiddenDropoutRate,
				updateWordVectors, initWordVectorsRandomly, mapWordVectorsToDim, gradientNormThreshold, evalMetric, doFiniteDifferenceChecking, demoMode);
			accuracies.push_back(currAccuracy);
			accuracySum += currAccuracy;
		}
		double accuracyMean = accuracySum / ((double) cvNumFolds);
		cerr << "CV mean accuracy: \t" << accuracyMean << endl;
		cerr << "CV accuracy stddev: \t" << standardDeviation(accuracies) << endl;
	}
	gettimeofday(&endTime, 0);
	double cumulativeTime = computeTimeDifference(initialTime, endTime);
	cerr << "Total training time:\t" << cumulativeTime << endl;
    return 0;
}
